#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <syslog.h>
#include <match.h>


/*
 * Processing of all kind of filters to drop articles before
 * actually downloading them. We need a drop file for each
 * consumer. So drop files must be part of consumer plug-ins.
 */

typedef struct _DROPLIST {
    int     drop;           /* 0 = get, 1 = drop */
    time_t  expires;        /* expiration time of this entry */
    char    *ngs;           /* list of newsgroups to check */
    char    *itm;           /* name of header line to check */
    char    *pat;           /* pattern to match with header */
    int     maxlines;       /* max number of lines */
} DROPLIST;

static int ndrops;
static char *droplines;
static DROPLIST *droplist;
static char dflt[] = "*";

/************************************************************************
*
*
************************************************************************/
static void report_error(char *name, int lnum, char *errtext)
{
    syslog(LOG_ERR, "%s(%u): %s, line ignored", name, lnum, errtext);
}

/************************************************************************
*
*
************************************************************************/
int load_dropfile(char *name)
{
    char **lpp;
    char *lp;

    if((ndrops = load_textfile(name, &droplines)) <= 0)
        return(ndrops);

    droplist = calloc(ndrops, sizeof(DROPLIST));
    ndrops = 0;
    for(lpp = groups; (cp = lp = *lpp++) != NULL; ) {

        /*
         * Interpret the type of this line, either Drop or Get.
         */
        switch(*lp) {
        case 'd':
        case 'D':
            drop = 1;
            break;
        case 'g':
        case 'G':
            drop = 0;
            break;
        default:
            drop = -1;
            break;
        }
        if(drop == -1) {
            report_error(name, lnum, "Syntax error, bad 1st char");
            continue;
        }

        /*
         * Need colon
         */
        if(*++cp != ':') {
            report_error(name, lnum, "2nd char must be colon");
            continue;
        }

        /*
         * Get optional expire time
         */
        if(*++cp != '/') {
            expires = atol(cp);
            while(*cp && *cp != '/')
                cp++;
            if(*cp != '/') {
                report_error(name, lnum, "Missing slash");
                continue;
            }
        }
        else
            expires = 0;

        /*
         * Get optional newsgroup list
         */
        if(*++cp != '/') {
            ngs = cp;
            while(*cp && *cp != '/')
                cp++;
            if(*cp != '/') {
                report_error(name, lnum, "Missing slash");
                continue;
            }
            *cp = '\0';
        }
        else
            ngs = dflt;

        /*
         * Get optional item
         */
        if(*++cp != ':') {
            itm = cp;
            while(*cp && *cp != ':')
                cp++;
            if(*cp != ':') {
                report_error(name, lnum, "Missing colon");
                continue;
            }
            *cp = '\0';
        }
        else
            itm = dflt;

        droplist[ndrops].drop = dropit;
        droplist[ndrops].expires = expires;
        droplist[ndrops].ngs = ngs;
        droplist[ndrops].itm = itm;
        droplist[ndrops].pat = cp + 1;
        if(stricmp(itm, "Lines") == 0)
            droplist[ndrops].maxlines = atoi(cp + 1);

        if(DOLOG(LOG_DROP)) {
            syslog(LOG_DEBUG, "%s(%u) %c:%ld/%s/%s:%s",
                                              cfg.filterfile,
                                              lnum,
                                              droplist[ndrops].drop ? 'D' : 'G',
                                              droplist[ndrops].expires,
                                              droplist[ndrops].ngs,
                                              droplist[ndrops].itm,
                                              droplist[ndrops].pat);
        ndrops++;
    }
    return(ndrops);
}

/************************************************************************
*
* drop_it
*
* In the final version we need an array of overview items, where the
* first seven items are fixed and the rest will be prepended by the
* name of the line followed by a colon.
*
************************************************************************/
int drop_it(char **headlines, char *newsgroups)
{
    int dropit = 0;
    DROPLIST *dp;

    for(idrop = 0, dp = droplist; idrop < ndrops; idrop++, dp++) {
        if(cmp_groups(newsgroups, dp->ngs) == 0)
            continue;

        for(i = 0; headline[i]; i++) {
            if((cp = strchr(headline[i], ':')) != NULL) {
                *cp = '\0';
                if(wildmat(headline[i], dp->itm)) {
                    *cp++ = ':';
                    while(*cp == ' ')
                        cp++;
                    if(dp->maxlines) {
                        if(dp->maxlines < atoi(cp)) {
                            dropit = dp->drop;
                            if(DOLOG(LOG_DROP))
                                syslog(LOG_DEBUG, "%s: %d < %d",
                                        dropit ? "Drop" : "Get",
                                        dp->maxlines, atoi(cp));
                            break;
                        }
                    }
                    else if(wildmat(cp, dp->pat)) {
                        dropit = dp->drop;
                        if(DOLOG(LOG_DROP))
                            syslog(LOG_DEBUG, "%s: '%s' = '%s'",
                                    dropit ? "Drop" : "Get",
                                    dp->pat, cp);
                        break;
                    }
                }
                else
                    *cp = ':';
            }
        }
    }
    return(dropit);
}

static char *ovfmt[] = {
    "Subject"   ,
    "From"      ,
    "Date"      ,
    "Message-ID",
    "References",
    "Bytes"     ,
    "Lines"
};

static int x_ovfmt = sizeof(ovfmt) / sizeof(ovfmt[0]);

#define MAX_OVERITEMS   16

/************************************************************************
*
*  drop_over
*
*  Compares an overview line against all entries of our filter file.
*
*  Returns 1 if the article should be dropped, 0 otherwise.
*
************************************************************************/
int drop_over(char *group, char *overline)
{
    int dropit = 0;
    register char *cp = overline;
    register char *cp2;
    char *ovp[MAX_OVERITEMS];
    char **headline;
    int n_ovp = 0;
    int i, j;
    int got_ngs = 0;

    for(;;) {
        while(*cp && *cp != '\t')
            cp++;
        if(*cp == '\t')
            *cp++ = '\0';
        else
            break;
        if(n_ovp >= MAX_OVERITEMS)
            break;
        ovp[n_ovp++] = cp;
    }

    if(n_ovp) {
        headline = malloc((n_ovp + 2) * sizeof(char *));

        for(i = j = 0; i < n_ovp; i++) {
            if(i < x_ovfmt) {
                if(ovp[i][0]) {
                    headline[j] = malloc(strlen(ovfmt[i]) + strlen(ovp[i]) + 3);
                    strcpy(headline[j], ovfmt[i]);
                    strcat(headline[j], ": ");
                    strcat(headline[j], ovp[i]);
                    j++;
                }
            }
            else {
                if((cp = strchr(ovp[i], ':')) != NULL && *(cp + 1)) {
                    *cp++ = '\0';
                    headline[j] = malloc(strlen(ovp[i]) + strlen(cp) + 3);
                    strcpy(headline[j], ovp[i]);
                    strcat(headline[j], ": ");
                    strcat(headline[j], cp);
                    j++;
                    if(strcmp(ovp[i], "Xref") == 0) {
                        headline[j] = malloc(sizeof("Newsgroups: ") + strlen(cp));
                        strcpy(headline[j], "Newsgroups: ");
                        for(cp2 = headline[j] + strlen(headline[j]); *cp; ) {
                            while(*cp && *cp != ':' && *cp != ' ')
                                *cp2++ = *cp++;
                            *cp2++ = ',';
                            while(*cp && *cp != ' ')
                                cp++;
                            while(*cp == ' ')
                                cp++;
                        }
                        *--cp2 = '\0';
                        j++;
                        got_ngs = 1;
                    }
                }
            }
        }
        if(!got_ngs) {
            headline[j] = malloc(sizeof("Newsgroups: ") + strlen(group));
            strcpy(headline[j], "Newsgroups: ");
            strcat(headline[j], group);
            j++;
        }
        headline[j] = NULL;
        dropit = new_drop(headline);
        for(j = 0; headline[j]; j++)
            free(headline[j]);
        free(headline);
    }

    return(dropit);
}

/************************************************************************/
/*  cmp_groups                                                          */
/*                                                                      */
/*  Compares a list of comma separated newsgroups against a pattern,    */
/*  which may contain wildcards.                                        */
/*                                                                      */
/*  Returns 1 if the pattern matches any of the newsgroups in the list  */
/*  or 0 otherwise.                                                     */
/*                                                                      */
/************************************************************************/
static int cmp_groups(char *nglist, char *ngpat)
{
    int result = 0;
    char *cp;

    while(*nglist) {
        if((cp = strchr(nglist, ',')) != NULL)
            *cp = '\0';
        result = ngmatch(nglist, ngpat);
        if(cp) {
            *cp++ = ',';
            if(result)
                break;
        }
        else
            break;
        nglist = cp;
    }
    return(result);
}
