/************************************************************************
*
* read_header
*
* Collects all header lines from a file stream into one large buffer
* and fills an array of character pointers with the starting position
* of these lines.
*
* Parameter    Description
* ------------ -------------------------------------------------------
* rbv          Points to a buffered tcp stream.
*
* headlines    Dynamic line buffer storage.
* ------------ -------------------------------------------------------
* Return value The number of lines or -1 on errors.
*
************************************************************************/
int read_header(FILE *fp, LINBUF *headlines)
{
    char *linebuf = malloc(MAX_BUFSIZ);
    char *line;
    int  nlines = 0;

    /*
     * Read all lines upto the first empty line and
     * interpret them as header lines.
     */
    linbuf_clear(headlines);
    for(;;) {
        if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.sockwait) <= 0) {
            nlines = -1;
            break;
        }
        line = linebuf;
        if(linebuf[0] == '.') {
            if(linebuf[1] == '\0')
                break;
            line++;
        }

        if(*line == ' ' || *line == '\t') {
            /*
             * Continuation lines start with space or tab. Ignore
             * them until we found our first real header line.
             */
            if(!nlines)
                continue;
            /*
             * Flag this continuation line and skip leading
             * spaces.
             */
            while(*line == ' ' || *line == '\t')
                line++;
            linbuf_append(headlines, line, ' ');
        }

        /*
         * The header ends with an empty line
         */
        else if(*line == '\0')
            break;

        /*
         * Ignore 'From ' and 'Received: ' lines which may have been
         * left over by some mailers.
         */
        else if(strnicmp(line, "From ", 5) == 0)
            continue;
        else if(strnicmp(line, ">From ", 6) == 0)
            continue;
        else if(!strnicmp(line, "Received: ", 10))
            continue;
        else {
            linbuf_add(headlines, line);
            nlines++;
        }
    }
    free(linebuf);

    return(nlines);
}

/************************************************************************/
/*  receive_block                                                       */
/*                                                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* rbv          Points to a buffered tcp stream.                        */
/*                                                                      */
/* bodylines    Dynamic line buffer storage.                            */
/* ------------ ------------------------------------------------------- */
/* Return value The number of lines or -1 on errors.                    */
/*                                                                      */
/************************************************************************/
int read_article(int *rbv, LINBUF *bodylines, char *signfile)
{
    char *linebuf = malloc(MAX_BUFSIZ);
    char *line;
    int  nlines = 0;
    FILE *fp;

    linbuf_clear(bodylines);
    for(;;) {
        if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.sockwait) <= 0) {
            nlines = -1;
            break;
        }
        if(linebuf[0] == '.') {
            if(linebuf[1] == '\0')
                break;
            line = linebuf + 1;
        }
        else
            line = linebuf;
        linbuf_add(bodylines, line);
        nlines++;
    }

    /*
     * Add an optional signature file.
     */
    if(signfile && *signfile) {
        if((fp = xfopen(signfile, "r")) != NULL) {
            while(fgets(linebuf, MAX_BUFSIZ, fp)) {
                linbuf_add(bodylines, linebuf);
                nlines++;
            }
            fclose(fp);
        }
        else
            syslog(LOG_ERR, "%s", _strerror(signfile));
    }

    free(linebuf);

    return(nlines);
}
