/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software has been inspired by several UNIX programs and their   */
/* OS/2 ports:                                                          */
/*                                                                      */
/* slurp - written 1992/93 by Stephen Hebditch and ported to OS/2 by    */
/* Jochen Friedrich.                                                    */
/*                                                                      */
/* nntpxfer - written by Brian Kantor                                   */
/*                                                                      */
/* nntpget - author unknown                                             */
/*                                                                      */
/* nntpclnt - Copyright 1992 by Stan Barber                             */
/*                                                                      */
/* souper - written by Chin Huang                                       */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * This module contains routines that handle article filtering.
 */

/*
 * $Log$
 */

const char drop_rcsid[] = "$Id$";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <syslog.h>
#include <xfile.h>

#include "config.h"
#include "chanx.h"


typedef struct _DROPLIST {
    struct _DROPLIST *nxt;  /* Link pointer */
    int     drop;           /* 0 = get, 1 = drop */
    time_t  expires;        /* expiration time of this entry */
    char    *ngs;           /* list of newsgroups to check */
    char    *itm;           /* name of header line to check */
    char    *pat;           /* pattern to match with header */
    int     maxlines;       /* max number of lines */
} DROPLIST;

DROPLIST *drop = NULL;

extern int ngmatch(char *group, char *list);
extern int wildmat(char *text, char *p);

/************************************************************************/
/*                                                                      */
/*  This routine reads and interprets a file named 'filter.cf'.         */
/*                                                                      */
/*  The file contains one entry per line.                               */
/*                                                                      */
/*  Lines starting with...                                              */
/*  ...space or '#' are ignored.                                        */
/*  ...'G' are interpreted as GET commands.                             */
/*  ...'D' are interpreted as DROP commands.                            */
/*                                                                      */
/*  Lines with GET and DROP commands must follow this syntax:           */
/*                                                                      */
/*  <D|G>:[<expire>]/[<newsgroups>]/[<header>]:<pattern>                */
/*                                                                      */
/*  <expire>                                                            */
/*  contains an optional expiration date of this entry. This            */
/*  part is not currently supported but reserved for future             */
/*  use. If the entry expires, the line will be commented.              */
/*                                                                      */
/*  <newsgroups>                                                        */
/*  defines an optional list of newsgroups, for which this              */
/*  line is valid. Wildcards may be used, several newsgroups            */
/*  must be separated by commas. If no newsgroup is specified,          */
/*  '*' is assumed.                                                     */
/*                                                                      */
/*  <header>                                                            */
/*  is optional and specifies the name of a headerline. If not          */
/*  given, all headerlines will be searched.                            */
/*                                                                      */
/*  <pattern>                                                           */
/*  This string may contain wildcards and specifies the                 */
/*  pattern to be searched in the given headerlines of the              */
/*  given newsgroups.                                                   */
/*                                                                      */
/*  After Chanx retrieved an article header, it will                    */
/*  sequentially walk through filter.cf to find out wether to           */
/*  drop or to get the body of that article. Because of the             */
/*  sequential processing, later entries in filter.cf may               */
/*  overwrite previous ones.                                            */
/*                                                                      */
/*  To give an example let's assume, we want to retrieve only           */
/*  those articles of all groups, which contain the word changi         */
/*  in the subject line.                                                */
/*                                                                      */
/*  D://:*                                                              */
/*  G://Subject:*Changi*                                                */
/*                                                                      */
/*  The first line will drop all articles. This is needed               */
/*  because initially Chanx would retrieve any article.                 */
/*                                                                      */
/*  The second line overwrites the drop for those articles which        */
/*  contain the word Changi in the subject. Note that all strings       */
/*  are not case-sensitive.                                             */
/*                                                                      */
/*  Returns 0 if a filter file could be read, -1 if not.                */
/*                                                                      */
/************************************************************************/
int read_drop(void)
{
    FILE *fp;
    char *cp;
    char *buf;
    int lnum = 0;
    int dropit;
    time_t expires;
    char *ngs;
    char *itm;
    DROPLIST *dp = drop;

    if((fp = xfopen(cfg.filterfile, "r")) == NULL) {
        if(DOLOG(LOG_DROP))
            syslog(LOG_DEBUG, "%s", _strerror(cfg.filterfile));
        return(-1);
    }

    buf = malloc(BUFSIZ);
    while(fgets(buf, BUFSIZ, fp)) {
        lnum++;
        if((cp = strchr(buf, '\n')) == NULL) {
            syslog(LOG_ERR, "%s (%u): Line too long", cfg.filterfile, lnum);
            fgets(buf, BUFSIZ, fp);
            continue;
        }
        *cp = '\0';
        cp = buf;
        switch(*cp) {
        case 'D':
        case 'd':
            dropit = 1;
            break;
        case 'G':
        case 'g':
            dropit = 0;
            break;
        case '\0':
        case ' ':
        case '#':
            dropit = -1;
            break;
        default:
            syslog(LOG_ERR, "%s (%u): Syntax error", cfg.filterfile, lnum);
            dropit = -1;
            break;
        }
        if(dropit == -1)
            continue;

        /*
         * Need colon
         */
        if(*++cp != ':') {
            syslog(LOG_ERR, "%s (%u): Missing colon", cfg.filterfile, lnum);
            continue;
        }

        /*
         * Get optional expire time
         */
        if(*++cp != '/') {
            expires = atol(cp);
            while(*cp && *cp != '/')
                cp++;
            if(*cp != '/') {
                syslog(LOG_ERR, "%s (%u): Missing slash", cfg.filterfile, lnum);
                continue;
            }
        }
        else
            expires = 0;

        /*
         * Get optional newsgroup list
         */
        if(*++cp != '/') {
            ngs = cp;
            while(*cp && *cp != '/')
                cp++;
            if(*cp != '/') {
                syslog(LOG_ERR, "%s (%u): Missing slash", cfg.filterfile, lnum);
                continue;
            }
            *cp = '\0';
        }
        else
            ngs = "*";

        /*
         * Get optional item
         */
        if(*++cp != ':') {
            itm = cp;
            while(*cp && *cp != ':')
                cp++;
            if(*cp != ':') {
                syslog(LOG_ERR, "%s (%u): Missing colon", cfg.filterfile, lnum);
                continue;
            }
            *cp = '\0';
        }
        else
            itm = "*";

        if(drop == NULL)
            dp = drop = memset(malloc(sizeof(DROPLIST)), 0, sizeof(DROPLIST));
        else {
            dp->nxt = memset(malloc(sizeof(DROPLIST)), 0, sizeof(DROPLIST));
            dp = dp->nxt;
        }
        dp->drop = dropit;
        dp->expires = expires;
        dp->ngs = strdup(ngs);
        dp->itm = strdup(itm);
        dp->pat = strdup(cp + 1);
        if(stricmp(itm, "Lines"))
            dp->maxlines = 0;
        else
            dp->maxlines = atoi(cp + 1);
        if(DOLOG(LOG_DROP)) {
            syslog(LOG_DEBUG, "%s(%u) %c:%ld/%s/%s:%s",
                                              cfg.filterfile,
                                              lnum,
                                              dp->drop ? 'D' : 'G',
                                              dp->expires,
                                              dp->ngs,
                                              dp->itm,
                                              dp->pat);
        }
    }
    free(buf);

    fclose(fp);

    return(0);
}

/************************************************************************/
/*  cmp_groups                                                          */
/*                                                                      */
/*  Compares a list of comma separated newsgroups against a pattern,    */
/*  which may contain wildcards.                                        */
/*                                                                      */
/*  Returns 1 if the pattern matches any of the newsgroups in the list  */
/*  or 0 otherwise.                                                     */
/*                                                                      */
/************************************************************************/
static int cmp_groups(char *nglist, char *ngpat)
{
    int result = 0;
    char *cp;

    while(*nglist) {
        if((cp = strchr(nglist, ',')) != NULL)
            *cp = '\0';
        result = ngmatch(nglist, ngpat);
        if(cp) {
            *cp++ = ',';
            if(result)
                break;
        }
        else
            break;
        nglist = cp;
    }
    return(result);
}

/************************************************************************/
/*  new_drop                                                            */
/*                                                                      */
/*  Compares an article header against all entries of our filter file.  */
/*                                                                      */
/*  Returns 1 if the article should be dropped, 0 otherwise.            */
/*                                                                      */
/************************************************************************/
int new_drop(char *headline[])
{
    int dropit = 0;
    DROPLIST *dp = drop;
    char *ngline = NULL;
    char *cp;
    int i;

    /*
     * Locate newsgroups line in linked list of header lines
     */
    for(i = 0; headline[i]; i++) {
        if(strnicmp(headline[i], "Newsgroups:", strlen("Newsgroups:")) == 0) {
            ngline = headline[i] + strlen("Newsgroups:");
            while(*ngline == ' ' || *ngline == '\t')
                ngline++;
            break;
        }
    }
    if(ngline) {
        /*
         * Loop for all filter entries
         */
        while(dp) {
            if(cmp_groups(ngline, dp->ngs)) {
                /*
                 * Loop for all header lines
                 */
                for(i = 0; headline[i]; i++) {
                    if((cp = strchr(headline[i], ':')) != NULL) {
                        *cp = '\0';
                        if(wildmat(headline[i], dp->itm)) {
                            *cp++ = ':';
                            while(*cp == ' ')
                                cp++;
                            if(dp->maxlines) {
                                if(dp->maxlines < atoi(cp)) {
                                    dropit = dp->drop;
                                    if(DOLOG(LOG_DROP))
                                        syslog(LOG_DEBUG, "%s: %d < %d",
                                                dropit ? "Drop" : "Get",
                                                dp->maxlines, atoi(cp));
                                    break;
                                }
                            }
                            else if(wildmat(cp, dp->pat)) {
                                dropit = dp->drop;
                                if(DOLOG(LOG_DROP))
                                    syslog(LOG_DEBUG, "%s: '%s' = '%s'",
                                            dropit ? "Drop" : "Get",
                                            dp->pat, cp);
                                break;
                            }
                        }
                        else
                            *cp = ':';
                    }
                }
            }
            dp = dp->nxt;
        }
    }
    return(dropit);
}

static char *ovfmt[] = {
    "Subject"   ,
    "From"      ,
    "Date"      ,
    "Message-ID",
    "References",
    "Bytes"     ,
    "Lines"
};

static int x_ovfmt = sizeof(ovfmt) / sizeof(ovfmt[0]);

#define MAX_OVERITEMS   16

/************************************************************************/
/*  over_drop                                                           */
/*                                                                      */
/*  Compares an overview line against all entries of our filter file.   */
/*                                                                      */
/*  Returns 1 if the article should be dropped, 0 otherwise.            */
/*                                                                      */
/************************************************************************/
int over_drop(char *group, char *overline)
{
    int dropit = 0;
    register char *cp = overline;
    register char *cp2;
    char *ovp[MAX_OVERITEMS];
    char **headline;
    int n_ovp = 0;
    int i, j;
    int got_ngs = 0;

    for(;;) {
        while(*cp && *cp != '\t')
            cp++;
        if(*cp == '\t')
            *cp++ = '\0';
        else
            break;
        if(n_ovp >= MAX_OVERITEMS)
            break;
        ovp[n_ovp++] = cp;
    }

    if(n_ovp) {
        headline = malloc((n_ovp + 2) * sizeof(char *));

        for(i = j = 0; i < n_ovp; i++) {
            if(i < x_ovfmt) {
                if(ovp[i][0]) {
                    headline[j] = malloc(strlen(ovfmt[i]) + strlen(ovp[i]) + 3);
                    strcpy(headline[j], ovfmt[i]);
                    strcat(headline[j], ": ");
                    strcat(headline[j], ovp[i]);
                    j++;
                }
            }
            else {
                if((cp = strchr(ovp[i], ':')) != NULL && *(cp + 1)) {
                    *cp++ = '\0';
                    headline[j] = malloc(strlen(ovp[i]) + strlen(cp) + 3);
                    strcpy(headline[j], ovp[i]);
                    strcat(headline[j], ": ");
                    strcat(headline[j], cp);
                    j++;
                    if(strcmp(ovp[i], "Xref") == 0) {
                        headline[j] = malloc(sizeof("Newsgroups: ") + strlen(cp));
                        strcpy(headline[j], "Newsgroups: ");
                        for(cp2 = headline[j] + strlen(headline[j]); *cp; ) {
                            while(*cp && *cp != ':' && *cp != ' ')
                                *cp2++ = *cp++;
                            *cp2++ = ',';
                            while(*cp && *cp != ' ')
                                cp++;
                            while(*cp == ' ')
                                cp++;
                        }
                        *--cp2 = '\0';
                        j++;
                        got_ngs = 1;
                    }
                }
            }
        }
        if(!got_ngs) {
            headline[j] = malloc(sizeof("Newsgroups: ") + strlen(group));
            strcpy(headline[j], "Newsgroups: ");
            strcat(headline[j], group);
            j++;
        }
        headline[j] = NULL;
        dropit = new_drop(headline);
        for(j = 0; headline[j]; j++)
            free(headline[j]);
        free(headline);
    }

    return(dropit);
}
