/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software has been inspired by several UNIX programs and their   */
/* OS/2 ports:                                                          */
/*                                                                      */
/* slurp - written 1992/93 by Stephen Hebditch and ported to OS/2 by    */
/* Jochen Friedrich.                                                    */
/*                                                                      */
/* nntpxfer - written by Brian Kantor                                   */
/*                                                                      */
/* nntpget - author unknown                                             */
/*                                                                      */
/* nntpclnt - Copyright 1992 by Stan Barber                             */
/*                                                                      */
/* souper - written by Chin Huang                                       */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * This module contains routines to retrieve message ids with the
 * NEWNEWS command.
 */
/*
 * $Log$
 */

const char getids_rcsid[] = "$Id$";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>

#include <chlimits.h>
#include <sostream.h>
#include <syslog.h>
#include <xfile.h>

#include "config.h"
#include "account.h"
#include "chanx.h"

/************************************************************************/
/*  add_oldids                                                          */
/*                                                                      */
/*  This routine tries to open a textfile containing message ids at the */
/*  beginning of each line. All ids found are added to the id pool with */
/*  filtering switched off.                                             */
/*                                                                      */
/************************************************************************/
static void add_oldids(char *fname)
{
    FILE *fp;
    char *line = malloc(MAX_BUFSIZ);

    if((fp = xfopen(fname, "r")) != NULL) {
        char *cp;

        while(fgets(line, MAX_BUFSIZ, fp)) {
            if(*line != '<')
                continue;
            for(cp = line; ; cp++)
                if(*cp == '>' || *cp == ' ' || *cp == '\t' || *cp == '\0')
                    break;
            if(*cp != '>')
                continue;
            *cp = '\0';
            add_id(line, 0);
        }
        fclose(fp);
    }
    free(line);
}

/************************************************************************/
/*  add_newids                                                          */
/*                                                                      */
/*  This routine sends a NEWNEWS command to the server and adds all     */
/*  ids received from the server to the id pool with filtering being    */
/*  switched on.                                                        */
/*                                                                      */
/*  Returns 1 on success, 0 if limit reached and -1 on broken           */
/*  connection.                                                         */
/*                                                                      */
/************************************************************************/
static int add_newids(int *rbv, char *nglist)
{
    int result = 0;
    int nbytes;
    char *line = malloc(MAX_BUFSIZ);

    /*
     * Send the NEWNEWS command to the remote server and process
     * the answer.
     */
    for(;;) {
        sprintf(line, "NEWNEWS %s %06lu %06lu GMT", nglist, cfg.lastdate,
                cfg.lasttime);
        if(cfg.distribution[0]) {
            strcat(line, " <");
            strcat(line, cfg.distribution);
            strcat(line, ">");
        }
        syslog(LOG_NOTICE, "%s", line);
        if(SoStreamTransact(rbv, line, line, cfg.remotewait * 2)) {
            free(line);
            return(-1);
        }
        if(atoi(line) != 480 && (DOMOD(MOD_AUTHSIMPLE) || atoi(line) != 450))
            break;
        if(authenticate(rbv, line)) {
            free(line);
            return(-1);
        }
    }
    if(*line != '2') {
        syslog(LOG_ERR, "Server error: %s", line);
        free(line);
        return(0);
    }

    /*
     * The remote sends us new ids line by line and terminates
     * the transmission by a line containing a single dot.
     */
    for(;;) {
        if((nbytes = SoStreamBufRecv(line, MAX_BUFSIZ, rbv, cfg.remotewait)) <= 0) {
            result = -1;
            break;
        }
        if(*line == '.') {
            result = 1;
            break;
        }
        if(watch_limits(nbytes)) {
            add_id(NULL, 0);
            break;
        }
        add_id(line, cfg.filtering);
    }

    free(line);
    return(result);
}

/************************************************************************/
/*  read_newsrc                                                         */
/*                                                                      */
/*  Read a standard newsrc file and create a comma separated list of    */
/*  newsgroups.                                                         */
/*                                                                      */
/*  Returns a pointer to the created list of NULL on failures. The      */
/*  list is allocated from heap space and has to be freed by the        */
/*  caller.                                                             */
/*                                                                      */
/************************************************************************/
static char *read_newsrc(char *fname)
{
    register char *cp;
    char *nglist;
    char *line;
    FILE *fp;
    int  nglen;
    int  needed = 0;

    if((fp = xfopen(fname, "r")) == NULL) {
        syslog(LOG_ERR, "%s", _strerror(fname));
        return(NULL);
    }

    line = malloc(MAX_BUFSIZ);
    while(fgets(line, MAX_BUFSIZ, fp)) {
        for(cp = line, nglen = 0; *cp && *cp != ':' && *cp > ' '; cp++)
            nglen++;
        if(*cp == ':')
            needed += nglen + 1;
    }
    free(line);

    if(needed == 0)
        return(NULL);

    rewind(fp);
    nglist = malloc((size_t)(needed + 1));
    memset(nglist, 0, (size_t)(needed + 1));
    line = nglist;
    while(needed && fgets(line, needed + 1, fp)) {
        for(cp = line, nglen = 0; *cp && *cp != ':' && *cp > ' '; cp++)
            nglen++;
        if(*cp == ':') {
            *cp++ = ',';
            line = cp;
            needed -= nglen + 1;
        }
    }
    *--line = '\0';

    return(nglist);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static void _Optlink id_thread(void *pv)
{
    int *rbv;
    int rc = 0;

    /*
     * Add previously dropped articles.
     */
    if(cfg.iwantfile[0])
        add_oldids(cfg.iwantfile);

    /*
     * Connect to remote
     */
    if((rbv = remote_connect()) != NULL) {
        char *nglist;
        char *ngnext;
        char *ngend = NULL;

        acci.id_state = STA_CONNECTED;

        /*
         * Create a newsgroup list in heap space.
         */
        if(cfg.newsrc[0])
            nglist = read_newsrc(cfg.newsrc);
        else
            nglist = strdup(cfg.grouplist);

        /*
         * Break list in NNTP transmitable parts and query them.
         */
        if(nglist) {
            for(ngnext = nglist; *ngnext; ngnext = ngend) {
                if(strlen(ngnext) > 384) {
                    ngend = ngnext + 384;
                    while(*ngend != ',')
                        ngend--;
                    *ngend++ = '\0';
                }
                else
                    ngend = "";
                if((rc = add_newids(rbv, ngnext)) != 1)
                    break;
            }
            free(nglist);
        }
        /*
         * Add id end marker, close connection and flag
         * the main thread that we've finished.
         */
        add_id(NULL, 0);
        SoStreamClose(*rbv);
        SoStreamBufDestroy(rbv);
        if(rc == -1)
            acci.id_state = STA_BROKEN;
        else
            acci.id_state = STA_DONE;
    }
    else
        acci.id_state = STA_BROKEN;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void collect_ids(void)
{
    _beginthread(id_thread, NULL, 32736, NULL);
}
