/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software has been inspired by several UNIX programs and their   */
/* OS/2 ports:                                                          */
/*                                                                      */
/* slurp - written 1992/93 by Stephen Hebditch and ported to OS/2 by    */
/* Jochen Friedrich.                                                    */
/*                                                                      */
/* nntpxfer - written by Brian Kantor                                   */
/*                                                                      */
/* nntpget - author unknown                                             */
/*                                                                      */
/* nntpclnt - Copyright 1992 by Stan Barber                             */
/*                                                                      */
/* souper - written by Chin Huang                                       */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * This module contains routines to add and get message ids to and from
 * the message id pool.
 */
/*
 * $Log$
 */

const char idpool_rcsid[] = "$Id$";

#include <stdlib.h>
#include <string.h>

#include <chlimits.h>
#include <syslog.h>
#include <sysdep.h>
#include <mutex.h>
#include <sostream.h>

#include "config.h"
#include "account.h"
#include "chanx.h"

#define MUTEXWAIT   (123000 + __LINE__)

typedef struct {
    void *next;
    int  filter;
    char id[1];
} MSGID_LINK;

static MSGID_LINK *idbot = NULL;
static MSGID_LINK *idtop = NULL;

static long lck_idpool = -1;
static void *rbv;
static int sock;

/************************************************************************/
/*                                                                      */
/************************************************************************/
int init_idpool(void)
{
    char *intro;

    lck_idpool = create_mutex();

    if ((sock = SoStreamConnect(cfg.localaddr, cfg.localport, 119)) < 0)
        return(-1);

    rbv = SoStreamBufCreate(sock);
    intro = malloc(MAX_BUFSIZ);
    if(SoStreamBufRecv(intro, MAX_BUFSIZ, rbv, cfg.remotewait) <= 0) {
        syslog(LOG_ERR, "Failed to get server intro");
        free(intro);
        exit_idpool();
        return(-1);
    }

    if (intro[0] != '2') {
        syslog(LOG_ERR, "Error %s", intro);
        free(intro);
        exit_idpool();
        return(-1);
    }
    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void exit_idpool(void)
{
    SoStreamBufDestroy(rbv);
    SoStreamClose(sock);
    destroy_mutex(lck_idpool);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void add_id(char *id, int filter)
{
    if(request_mutex(&lck_idpool, MUTEXWAIT))
        return;

    if(id == NULL) {
        /*
         * No more ids available.
         */
        if(idtop) {
            idtop->next = malloc(sizeof(MSGID_LINK));
            idtop = idtop->next;
        }
        else
            idbot = idtop = malloc(sizeof(MSGID_LINK));
        idtop->next = NULL;
        idtop->filter = -1;
        idtop->id[0] = '\0';
    }
    else if(!DOMOD(MOD_TEST) && offer_id(rbv, id) == 0)
        acci.num_dupes++;
    else {
        size_t idlen = strlen(id);

        if((idlen >= MAX_MSGID) || (strchr(id, ' ') != NULL) ||
           (*id != '<') || (*(id + idlen - 1) != '>'))
            syslog(LOG_ERR, "Got illegal id: '%s'", id);
        else {
            if(idtop) {
                idtop->next = malloc(sizeof(MSGID_LINK) + idlen);
                idtop = idtop->next;
            }
            else
                idbot = idtop = malloc(sizeof(MSGID_LINK) + idlen);
            idtop->next = NULL;
            idtop->filter = filter;
            memcpy(idtop->id, id, idlen + 1);
        }
        acci.num_new++;
    }
    release_mutex(lck_idpool);
}

/************************************************************************/
/*                                                                      */
/* get_id                                                               */
/*                                                                      */
/* Retrieves the next available id from the pool. If there is no id     */
/* currently available, then the routine loops with a sleep.            */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* id           Points to a buffer receiving the message id. This       */
/*              buffer should be at least MAX_MSGID in size.            */
/* ------------ ------------------------------------------------------- */
/* Return value -1 if no more ids are available                         */
/*              0 if the article needs no filtering                     */
/*              1 if the article needs filtering                        */
/*                                                                      */
/************************************************************************/
int get_id(char *id)
{
    int filter;
    int waiting = 1;

    while(waiting) {
        if(request_mutex(&lck_idpool, MUTEXWAIT) == 0) {
            if(idbot) {
                filter = idbot->filter;
                if(filter != -1) {
                    void *lnk = idbot;
                    strcpy(id, idbot->id);
                    if((idbot = idbot->next) == NULL)
                        idtop = NULL;
                    free(lnk);
                }
                waiting = 0;
            }
            release_mutex(lck_idpool);
        }
        if(waiting)
            ms_sleep(100);
    }
    return(filter);
}
