/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software has been inspired by several UNIX programs and their   */
/* OS/2 ports:                                                          */
/*                                                                      */
/* slurp - written 1992/93 by Stephen Hebditch and ported to OS/2 by    */
/* Jochen Friedrich.                                                    */
/*                                                                      */
/* nntpxfer - written by Brian Kantor                                   */
/*                                                                      */
/* nntpget - author unknown                                             */
/*                                                                      */
/* nntpclnt - Copyright 1992 by Stan Barber                             */
/*                                                                      */
/* souper - written by Chin Huang                                       */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/
/*
 * This module contains routines to connect to a remote server
 * and receive articles from it.
 */
/*
 * $Log$
 */

const char remconn_rcsid[] = "$Id$";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <chlimits.h>
#include <syslog.h>
#include <sostream.h>

#include "config.h"
#include "chanx.h"

/************************************************************************/
/*  remote_connect                                                      */
/*                                                                      */
/*  Connects to a remote. Hostname and port are given in the config     */
/*  structure.                                                          */
/*                                                                      */
/*  If configured to do so, this routine will also send SLAVE and       */
/*  MODE READER command to the remote.                                  */
/*                                                                      */
/*  Returns a pointer to a buffered socket stream (see sostream.dll)    */
/*  if successful, NULL otherwise.                                      */
/*                                                                      */
/************************************************************************/
int *remote_connect(void)
{
    void *rbv;
    int sock;
    char *buf = malloc(MAX_BUFSIZ);

    /*
     * Connect to remote
     */
    if ((sock = SoStreamConnect(cfg.remoteaddr, cfg.remoteport, 119)) < 0) {
        syslog(LOG_ERR, "Could not connect to %s", cfg.remoteaddr);
        free(buf);
        return(NULL);
    }

    rbv = SoStreamBufCreate(sock);
    if(SoStreamBufRecv(buf, MAX_BUFSIZ, rbv, cfg.remotewait) <= 0) {
        SoStreamClose(sock);
        SoStreamBufDestroy(rbv);
        free(buf);
        return(NULL);
    }

    if (buf[0] != '2') {
        syslog(LOG_ERR, "sock error: %s", buf);
        SoStreamClose(sock);
        SoStreamBufDestroy(rbv);
        free(buf);
        return(NULL);
    }

    /*
     * Set posting ability flag
     */
    cfg.canpost = (atoi(buf) != 201);

    /*
     * Tell them we're a slave process
     */
    if(DOMOD(MOD_SLAVE)) {
        for(;;) {
            if(SoStreamTransact(rbv, "SLAVE", buf, cfg.remotewait)) {
                SoStreamClose(sock);
                SoStreamBufDestroy(rbv);
                free(buf);
                return(NULL);
            }
            if(atoi(buf) != 480 && (DOMOD(MOD_AUTHSIMPLE) || atoi(buf) != 450))
                break;
            if(authenticate(rbv, buf)) {
                SoStreamClose(sock);
                SoStreamBufDestroy(rbv);
                free(buf);
                return(NULL);
            }
        }
        if (buf[0] != '2')
            syslog(LOG_NOTICE, "Warning: %s", buf);
    }

    /*
     * Mode reader may be used on INN servers
     * to invoke NNRPD, the newsreader server.
     */
    if(DOMOD(MOD_READER)) {
        for(;;) {
            if(SoStreamTransact(rbv, "MODE READER", buf, cfg.remotewait)) {
                SoStreamClose(sock);
                SoStreamBufDestroy(rbv);
                free(buf);
                return(NULL);
            }
            if(atoi(buf) != 480 && (DOMOD(MOD_AUTHSIMPLE) || atoi(buf) != 450))
                break;
            if(authenticate(rbv, buf)) {
                SoStreamClose(sock);
                SoStreamBufDestroy(rbv);
                free(buf);
                return(NULL);
            }
        }
        if (buf[0] != '2')
            syslog(LOG_NOTICE, "Warning: %s", buf);
    }

    /*
     * If username starts with an exclamation mark, we
     * actively authenticate ourself.
     */
    if(cfg.auth_user[0] == '!') {
        if(authenticate(rbv, buf)) {
            SoStreamClose(sock);
            SoStreamBufDestroy(rbv);
            free(buf);
            return(NULL);
        }
    }

    free(buf);
    return(rbv);
}

/************************************************************************/
/*  receive_header                                                      */
/*                                                                      */
/*  This routine assumes that a HEADER command has been already sent.   */
/*                                                                      */
/************************************************************************/
long receive_header(int *rbv, char **headbuf, size_t *sizep, char *hdrline[])
{
    long result = 0;
    size_t hdr_size = 0;
    size_t left = *sizep;
    char *cp = *headbuf;
    size_t line_size;
    char *linebuf = malloc(MAX_BUFSIZ);
    char *line;
    int  cline = 0;

    if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.remotewait) <= 0)
        result = -1;
    else if(linebuf[0] == '4')
        syslog(LOG_NOTICE, "Article missing: %s", linebuf);
    else if(linebuf[0] != '2')
        syslog(LOG_NOTICE, "Server error: %s", linebuf);
    else {
        for(;;) {
            if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.remotewait) <= 0)
                break;
            if(linebuf[0] == '.') {
                if(linebuf[1] == '\0')
                    break;
                line = linebuf + 1;
            }
            else
                line = linebuf;
            line_size = strlen(line) + 1;
            if(line_size > left) {
                left += 32736;
                *sizep += 32736;
                *headbuf = realloc(*headbuf, *sizep);
                if(_heapmin()) {
                    syslog(LOG_ERR, "Heap memory failure");
                    break;
                }
                cp = *headbuf + hdr_size;
            }
            memcpy(cp, line, line_size);
            if(cline < MAX_HEADLINES)
                hdrline[cline++] = cp;
            else
                syslog(LOG_ERR, "Too many header lines");
            cp += line_size;
            left -= line_size;
            hdr_size += line_size;
            if(DOMOD(MOD_DOSFILE))
                hdr_size++;
        }
        result = (long)hdr_size;
    }
    free(linebuf);
    hdrline[cline] = NULL;

    return(result);
}

/************************************************************************/
/* receive_block                                                        */
/*                                                                      */
/* This routine is used to receive either an articles body or a         */
/* complete article from the remote server. A BODY or ARTICLE           */
/* command must have been send before calling this one.                 */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* rbv          Handle to a stream buffer.                              */
/*                                                                      */
/* bodypp       Pointer to a pointer to memory block.                   */
/*                                                                      */
/* sizep        Pointer to a variable containing the number of bytes    */
/*              available in the given memory block.                    */
/* ------------ ------------------------------------------------------- */
/* Return value Returns the size of the received block, 0 if the        */
/*              remote site reports an error or sends an empty block,   */
/*              or -1 on connection errors.                             */
/*                                                                      */
/************************************************************************/
long receive_block(int *rbv, char **blkpp, size_t *sizep)
{
    long result = 0;
    size_t art_size = 0;
    size_t left = *sizep;
    char *cp = *blkpp;
    size_t line_size;
    char *linebuf = malloc(MAX_BUFSIZ);
    char *line;

    /*
     * Get the remote respose of a command sent
     * before calling this routine.
     */

    if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.remotewait) <= 0)
        result = -1;
    else if(linebuf[0] == '4')
        syslog(LOG_NOTICE, "Article missing: %s", linebuf);
    else if(linebuf[0] != '2')
        syslog(LOG_NOTICE, "Server error: %s", linebuf);
    else {

        /*
         * Receive the block line by line until any
         * of them contains just a single dot.
         */
        for(;;) {
            if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.remotewait) <= 0)
                break;
            if(linebuf[0] == '.') {
                if(linebuf[1] == '\0')
                    break;
                line = linebuf + 1;
            }
            else
                line = linebuf;
            line_size = strlen(line);

            /*
             * If the new line plus linefeed plus string termination
             * will not fit into the rest of the buffer, then
             * we reallocate more memory.
             */
            if(line_size + 2 > left) {
                if(*blkpp == NULL) {
                    left = 32736;
                    *sizep = 32736;
                    *blkpp = malloc(*sizep);
                }
                else {
                    left += 32736;
                    *sizep += 32736;
                    *blkpp = realloc(*blkpp, *sizep);
                }
                cp = *blkpp + art_size;
            }
            memcpy(cp, line, line_size);
            cp += line_size;
            *cp++ = '\n';
            *cp = '\0';
            line_size++;
            left -= line_size;
            art_size += line_size;
            if(DOMOD(MOD_DOSFILE))
                art_size++;
        }
        result = (long)art_size;
    }
    free(linebuf);

    return(result);
}
