/*
 *  simple history database manager for UUPC news
 *
 *  Author:  Kai Uwe Rommel <rommel@ars.muc.de>
 *  Created: Sun Aug 15 1993
 *
 *  Bug reports related to THIS modified version should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

/*--------------------------------------------------------------------*/
/*       Changes Copyright (c) 1989-1994 by Kendra Electronic         */
/*       Wonderworks.                                                 */
/*                                                                    */
/*       All rights reserved except those explicitly granted by       */
/*       the UUPC/extended license agreement.                         */
/*--------------------------------------------------------------------*/

const char dbm_rcsid[] = "$Id: dbm.c,v 1.1 1997/02/13 11:21:43 oliver Exp $";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

#include <syslog.h>
#include <sysdep.h>

#include "dbm.h"
#include "idx.h"

datum nullitem = {NULL, 0};

DBM *dbm_open(const char *name, int flags, int mode)
{
    DBM *db = malloc(sizeof(DBM));
    char *filename = malloc(_MAX_PATH);
    int retries = 20;

    strcpy(filename, name);
    strcat(filename, DBM_EXT_DBF);

    while(retries--) {
        if ((db->dbffile = open(filename, flags | O_BINARY, mode)) != -1)
            break;
        if(errno != EACCESS)
            break;
        ms_sleep(100);
    }

    if (db->dbffile == -1) {
        syslog(LOG_ERR, "%s", _strerror(filename));
        free(db);
        free(filename);
        return NULL;
    }

    strcpy(filename, name);
    strcat(filename, DBM_EXT_IDX);

    if ((db->idxfile = open(filename, flags | O_BINARY, mode)) == -1) {
        syslog(LOG_ERR, "%s", _strerror(filename));
        close(db->dbffile);
        free(db);
        free(filename);
        return NULL;
    }

    if ((db->idx = idx_init(db->idxfile)) == NULL) {
        syslog(LOG_ERR, "Unable to initialize history index");
        close(db->dbffile);
        close(db->idxfile);
        free(db);
        free(filename);
        return NULL;
    }

    db->magic = DBM_MAGIC;

    free(filename);
    return db;
}

void dbm_close(DBM * db)
{
    if (db == NULL || db->magic != DBM_MAGIC) {
        syslog(LOG_ERR, "Invalid dbm handle");
        return;
    }

    idx_exit(db->idx);

    if(close(db->idxfile))
        syslog(LOG_ERR, "%s", _strerror("Close dbm index"));
    if(close(db->dbffile))
        syslog(LOG_ERR, "%s", _strerror("Close dbm"));

    free(db);
}

int dbm_store(DBM * db, datum key, datum val, int flag)
{
    char *buffer;
    size_t sob;
    int  sobdiff;
    long offset;
    int size;

    if (db == NULL || db->magic != DBM_MAGIC) {
        syslog(LOG_ERR, "Invalid dbm handle");
        return -1;
    }

    if ((offset = lseek(db->dbffile, 0, SEEK_END)) == -1L) {
        syslog(LOG_ERR, "%s", _strerror("Seek history"));
        return -1;
    }

    sob = key.dsize + val.dsize + 3;
    if((sobdiff = sob - sizeof(db->buffer)) > 0)
        val.dptr -= sobdiff;
    buffer = memset(malloc(sob), ' ', sob);
    memcpy(buffer, key.dptr, key.dsize);
    size = key.dsize;
    buffer[size - 1] = ' ';  /* replace zero */
    memcpy(buffer + size, val.dptr, val.dsize);
    size += val.dsize;
    buffer[size - 1] = '\n';

    if (idx_addkey(db->idx, key.dptr, offset, size) == -1) {
        free(buffer);
        return -1;
    }

    if (write(db->dbffile, buffer, size) != size) {
        syslog(LOG_ERR, "%s", _strerror("dbm store"));
        free(buffer);
        return -1;
    }
    if(flag) ;

    free(buffer);
    return 0;
}

/*
 * Changed some code to keep the ids of expired/cancelled articles.
 * As this is a bad hack done to these holy routines, I changed the
 * name from dmb_delete to dbm_cancel.
 *
 * 25.05.95 harald
 */
int dbm_cancel(DBM * db, datum key)
{
    register char *cp;
    char *buffer;
    long offset;
    size_t size;

    if (db == NULL || db->magic != DBM_MAGIC)
        return -1;

    if (idx_getkey(db->idx, key.dptr, &offset, (int *)(&size)) != -1) {
        if (lseek(db->dbffile, offset, SEEK_SET) == -1L) {
            syslog(LOG_ERR, "%s", _strerror("seek history"));
            return -1;
        }

        buffer = malloc(size + 1);
        buffer[size] = '\0';
        if (read(db->dbffile, buffer, size) != size) {
            free(buffer);
            return -1;
        }
        cp = buffer;
        while(*cp && *cp != ' ' && *cp != '\t')
            cp++;
        while(*cp == ' ' || *cp == '\t')
            cp++;
        while(*cp)
            *cp++ = ' ';
        buffer[size - 1] = '\n';

        if (lseek(db->dbffile, offset, SEEK_SET) == -1L) {
            syslog(LOG_ERR, "%s", _strerror("seek history"));
            free(buffer);
            return -1;
        }

        if (write(db->dbffile, buffer, size) != size) {
            free(buffer);
            return -1;
        }
        free(buffer);
    }

    return 0;
}

datum dbm_fetch(DBM * db, datum key)
{
    datum val = nullitem;
    long offset;
    int size;

    if (db == NULL || db->magic != DBM_MAGIC)
        return nullitem;

    if (db->stream && strcmp(key.dptr, db->buffer) == 0) {
        val.dptr = db->value;
        val.dsize = strlen(val.dptr) + 1;
    }
    else if (idx_getkey(db->idx, key.dptr, &offset, &size) != -1) {
        if ((offset = lseek(db->dbffile, offset, SEEK_SET)) == -1L) {
            syslog(LOG_ERR, "%s", _strerror("seek history"));
            return nullitem;
        }

        if (read(db->dbffile, db->buffer, size) != size)
            return nullitem;

        db->buffer[size - 1] = 0;  /* delete \n */

        val.dptr = strchr(db->buffer, ' ') + 1;
        val.dsize = strlen(val.dptr) + 1;
    }

    return val;
}

/* Accessing the database sequentially is not as easy as the records
 * are of variable length to save space and make it look like a text
 * file. So we just put a stream on top of it and read it this way. */

datum dbm_firstkey(DBM * db)
{
    datum val = nullitem;
    char *ptr;
    int handle;

    if (db == NULL || db->magic != DBM_MAGIC)
        return nullitem;

    if (lseek(db->dbffile, 0, SEEK_SET) == -1L) {
        syslog(LOG_ERR, "%s", _strerror("seek history"));
        return nullitem;
    }

    if ((handle = dup(db->dbffile)) == -1)
        return nullitem;

    if ((db->stream = fdopen(handle, "rb")) == NULL) {
        syslog(LOG_ERR, "%s", _strerror("dbm firstkey"));
        return nullitem;
    }

    do {                        /* skip blanked out records */
        if (fgets(db->buffer, sizeof(db->buffer), db->stream) == NULL)
        return nullitem;
    } while (db->buffer[0] == ' ');

    if ((ptr = strchr(db->buffer, ' ')) == NULL)
        return nullitem;

    db->buffer[strlen(db->buffer) - 1] = 0;        /* delete \n */

    *ptr = 0;
    db->value = ptr + 1;

    val.dptr = db->buffer;
    val.dsize = strlen(db->buffer) + 1;

    return val;
}

datum dbm_nextkey(DBM * db)
{
    datum val = nullitem;
    char *ptr;

    if (db == NULL || db->magic != DBM_MAGIC || db->stream == NULL)
        return nullitem;

    do {                          /* skip blanked out records */
        if (fgets(db->buffer, sizeof(db->buffer), db->stream) == NULL)
            return fclose(db->stream), (db->stream = NULL), nullitem;
    } while (db->buffer[0] == ' ');

    if ((ptr = strchr(db->buffer, ' ')) == NULL)
        return nullitem;

    db->buffer[strlen(db->buffer) - 1] = 0;         /* delete \n */

    *ptr = 0;
    db->value = ptr + 1;

    val.dptr = db->buffer;
    val.dsize = strlen(db->buffer) + 1;

    return val;
}

/*
 * $Log: dbm.c,v $
 * Revision 1.1  1997/02/13 11:21:43  oliver
 * initial revision
 *
 */
