/*
 * This program was written by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

const char expire_rcsid[]="$Id: expire.c,v 1.1 1997/02/13 11:21:45 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <time.h>
#include <signal.h>
#include <process.h>

#include <version.h>
#include <chlimits.h>
#include <syslog.h>
#include <getopt.h>
#include <sostream.h>
#include <sodtgram.h>
#include <xfile.h>
#include <sysdep.h>
#include <ndbm.h>

#include <dircache.h>
#include <nactive.h>

#include "config.h"
#include "rebuild.h"
#include "expire.h"

long ids_processed; /* Ids of processed active articles excl. remembered */
long ids_kept;      /* Ids of active articles not expired */
long ids_expired;   /* Ids of expired articles */
long files_deleted; /* Article files deleted */
long dirs_removed;  /* Group directories removed */
long old_dropped;   /* Remembered ids dropped */
long old_kept;      /* Remembered ids retained */
long bytes_removed; /* Size of removed files */
long now;
FILE *zapfile;
int syslog_opt = LOG_CONS | LOG_NDELAY | LOG_NOWAIT;

static void parsedbgopt(char *opt);
static void parsexclopt(char *opt);
static void usage(void);
static void cleanup(int sig);

char server_cmd[MAX_BUFSIZ];

static FILE *fppid = NULL;

/************************************************************************/
/*                                                                      */
/*  EXPIRE                                                              */
/*                                                                      */
/*  Return codes                                                        */
/*                                                                      */
/*      0  OK, articles expired                                         */
/*      1  Wrong parameters, usage help displayed                       */
/*      2  Nothing to expire                                            */
/*      3  Couldn't read active file, nothing done                      */
/*      4  Error while expiring                                         */
/*      5  Failed to update active file, expiring done anyway           */
/*      6  Error reading config file                                    */
/*      8  Already running                                              */
/*                                                                      */
/************************************************************************/
int main(int argc, char **argv)
{
    int result = 0;
    int option;
    char *cp;

    setvbuf(stderr, NULL, _IONBF, 0);
    _tzset();
    time(&now);

    /*
     * Initialize socket interface.
     */
    if(SoInitialize()) {
        fprintf(stderr, "FATAL ERROR: TCP/IP initialization failed\n");
        return(128);
    }

    fprintf(stderr, "Expire %s - %s %s\n", version, __DATE__, __TIME__);

    init_cfg();

    /*
     * Create pid file and keep it open for lock.
     */
    if(cfg.pidfile[0] != '\0') {
        if((fppid = fopen(cfg.pidfile, "w")) == NULL) {
            perror(cfg.pidfile);
            return(8);
        }
        fprintf(fppid, "%u\n", _getpid());
        fflush(fppid);
    }

    /*
     * Process command line options
     */
    while((option = getopt(argc, argv, "?aAbCc:d:e:E:hIiLMpq:rR:s:S:t:TUu:vVw:x:y:z:")) != EOF) {
        switch(option) {
        case 'A':
            cfg.createactive = 1;
        case 'a':
            cfg.rebuildactive = 1;
            break;
        case 'b':
            cfg.removebad = 1;
            break;
        case 'C':   /* Cancel expired articles, don't rebuild history */
            cfg.cancel = 1;
            break;
        case 'c':   /* Set name of control file */
            strcpy(cfg.expireconf, optarg);
            break;
        case 'd':   /* Debugging */
            parsedbgopt(optarg);
            break;
        case 'e':
            cfg.expiresecs = convtime(optarg, 'd');
            break;
        case 'E':
            cfg.forgetsecs = convtime(optarg, 'd');
            break;
        case 'h':       /* ignore history */
            cfg.nohistory = 1;
            break;
        case 'I':
            cfg.ignorexp = 2;
            break;
        case 'i':       /* Ignore any existing expiration date */
            cfg.ignorexp = 1;
            break;
        case 'L':       /* Low priority */
            set_priority(-1);
            break;
        case 'M':       /* Use last write instead of creation date */
            cfg.uselastwrite = 1;
            break;
        case 'p':       /* Use posting instead of arrival date */
            cfg.usepost = 1;
            break;
        case 'q':       /* Quit after specified number of errors */
            cfg.quitbad = atoi(optarg);
            break;
        case 'r':       /* rebuild history */
            cfg.nohistory = 1;
            cfg.dorebuild = 1;
            break;
        case 'R':
            strcpy(cfg.expreason, optarg);
            break;
        case 's':       /* Sorting program to be called with rebuilding */
            strcpy(cfg.sortcall, optarg);
            break;
        case 'S':       /* Name of a summary file */
            strcpy(cfg.summary, optarg);
            break;
        case 't':       /* Define temporary directory */
            strcpy(cfg.tempdir, optarg);
            break;
        case 'T':       /* Sort active file */
            cfg.sortactive = 1;
            break;
        case 'U':
            cfg.oldindex = 1;
            break;
        case 'u':
            strcpy(cfg.logname, optarg);
            break;
        case 'V':
            cfg.verify++;
            break;
        case 'v':
            cfg.verbose++;
            break;
        case 'w':
            if(*optarg == '-')
                cfg.timewarp = -convtime(optarg + 1, 'd');
            else
                cfg.timewarp = convtime(optarg, 'd');
            break;
        case 'x':
            parsexclopt(optarg);
            break;
        case 'y':
            strcpy(cfg.copyfrom, optarg);
            break;
        case 'z':
            if(zapfile == NULL) {
                if(*optarg == '+')
                    zapfile = fopen(optarg + 1, "a");
                else
                    zapfile = fopen(optarg, "w");
                if(zapfile == NULL)
                    syslog(LOG_ERR, "%s", _strerror(optarg));
            }
            break;
        default:
            fprintf(stderr, "Unknown option -%c\n", option);
        case '?':
            usage();
            return(1);
        }
    }

    /*
     * All command line options are processed, check if one
     * argument is left to give us a list of newsgroups.
     */
    argc -= optind;
    argv += optind;
    if(argc)
        strcpy(cfg.groups, argv[0]);

    /*
     * Store name of a temporary directory with trailing backspace
     */
    for(cp = cfg.tempdir; *cp; cp++) {
        if(*cp == '/')
            *cp = '\\';
    }
    if(*(cp - 1) != '\\')
        *cp++ = '\\';

    /*
     * Open syslog channel
     */
    openlog("expire", syslog_opt, LOG_NEWS);
    syslog(LOG_INFO, "Version %s-%s up and running...", version, __DATE__);

    /*
     * Setup signal handling
     */
    signal(SIGABRT, cleanup);
    signal(SIGBREAK, cleanup);
    signal(SIGINT, cleanup);
    signal(SIGSEGV, cleanup);
    signal(SIGTERM, cleanup);

    /*
     * Send flush cmd to server
     */
    if(!cfg.noserver) {
        sprintf(server_cmd, "%s\001F\001%s", cfg.logname, cfg.expreason);
        syslog(LOG_ERR, "%s, %s\n", cfg.controlhost, cfg.controlport);
        if(SoDtgramTransact(cfg.controlhost, cfg.controlport, 119,
                            server_cmd, server_cmd, 30) == -1)
            syslog(LOG_ERR, "Can't flush server");
    }

    /*
     * Load, rebuild or create active file
     */
    if(!cfg.verify) {
        if(cfg.verbose) {
            if(cfg.createactive)
                fputs("Creat", stderr);
            else if(cfg.rebuildactive)
                fputs("Rebuild", stderr);
            else
                fputs("Load", stderr);
            fputs("ing active...", stderr);
        }
        if(init_dircache(cfg.newsdir, 32) || init_active(cfg.newsdir)) {
            fprintf(stderr, "Init failed\n");
            result = 3;
        }
        else if(!DOLOG(LOG_TESTMODE) && cfg.createactive && new_active(cfg.activefile)) {
            fprintf(stderr, "Create failed\n");
            result = 3;
        }
        else {
            if(load_active(cfg.activefile) < 1) {
                fprintf(stderr, "Load0 failed\n");
                if(cfg.rebuildactive) {
                    if(!DOLOG(LOG_TESTMODE) && new_active(cfg.activefile)) {
                        fprintf(stderr, "Load1 failed\n");
                        result = 3;
                    }
                    else if(load_active(cfg.activefile)) {
                        fprintf(stderr, "Load2 failed\n");
                        result = 3;
                    }
                }
                else {
                    fprintf(stderr, "No rebuild failure\n");
                    result = 3;
                }
            }
            if(cfg.rebuildactive)
                scan_groups("");
        }
        if(cfg.verbose)
            fputs(result ? "failed\n" : "done\n", stderr);

        /*
         * Read expire control file
         */
        if(!result) {
            /*
             * Set all groups to default expiration time
             */
            set_expiration("*", cfg.expiresecs, cfg.expiresecs);

            if(cfg.expireconf[0]) {
                if(read_expcnf(cfg.expireconf))
                    result = 6;
            }
            else if(access("expire.conf", 0) == 0) {
                if(read_expcnf("expire.conf"))
                    result = 6;
            }
        }
    }

    if(!result) {
        if(cfg.verify) {
#ifdef USE_REBUILD
            if(test_history())
                result = 4;
#endif
        }
        else if(cfg.nohistory) {
            if(purge_by_active())
                result = 4;
        }
        else if(cfg.cancel) {
            if(purge_by_cancel())
                result = 4;
        }
        else if(cfg.copyfrom[0]) {
            if(cfg.verbose)
                fputs("Rebuilding history....", stderr);
            if(!DOLOG(LOG_TESTMODE)) {
                /*
                 * Pause the server
                 */
                if(!cfg.noserver) {
                    sprintf(server_cmd, "%s\001m\001%s", cfg.logname, cfg.expreason);
                    if(SoDtgramTransact(cfg.controlhost, cfg.controlport, 119,
                                        server_cmd, server_cmd, 30) == -1)
                        syslog(LOG_ERR, "Can't pause server");
                }

#ifdef USE_REBUILD
                if(cfg.oldindex)
                    result = rebuild_oldhistory(cfg.copyfrom);
                else
                    result = rebuild_history(cfg.copyfrom);
#else
                if(!cfg.oldindex)
                    result = rebuild_memory(cfg.copyfrom);
#endif
            }
            if(cfg.verbose)
                fputs(result ? "failed\n" : "done\n", stderr);
            if(result)
                result = 4;
        }
        else {
            if(purge_by_history())
                result = 4;
        }
    }

    if(!cfg.verify) {
        if(zapfile && fclose(zapfile))
            syslog(LOG_ERR, "%s", _strerror("zapfile"));

        if(!result && !cfg.noactive && !DOLOG(LOG_TESTMODE)) {
            /*
             * Pause the server
             */
            if(!cfg.noserver) {
                sprintf(server_cmd, "%s\001m\001%s", cfg.logname, cfg.expreason);
                if(SoDtgramTransact(cfg.controlhost, cfg.controlport, 119,
                                    server_cmd, server_cmd, 30) == -1)
                    syslog(LOG_ERR, "Can't pause server");
            }

            /*
             * Save active file
             */
            if(cfg.verbose)
                fputs("Saving active...", stderr);
            if(save_active(cfg.activefile, 0))
                result = 5;
            if(cfg.verbose) {
                if(result)
                    fputs("failed\n", stderr);
                else
                    fputs("done\n", stderr);
            }
        }
        exit_active();
    }

    /*
     * Restart the server
     */
    if(!cfg.noserver) {
        sprintf(server_cmd, "%s\001h\001%s", cfg.logname, cfg.expreason);
        if(SoDtgramTransact(cfg.controlhost, cfg.controlport, 119,
                            server_cmd, server_cmd, 30) == -1)
            syslog(LOG_ERR, "Can't restart server");
    }

    if((!result || result == 5) && !cfg.verify) {
        /*
         * Save statistics
         */
        if(cfg.summary[0]) {
            FILE *fps;

            if((fps = xfopen(cfg.summary, "a")) != NULL) {
                fprintf(fps, "%ld %ld %ld %ld %ld %ld %ld\n", now,
                        ids_processed, ids_kept, ids_expired, files_deleted,
                        old_dropped, old_kept);
                fclose(fps);
            }
            else
                syslog(LOG_ERR, "%s", _strerror(cfg.summary));
        }
        if(cfg.verbose) {
            fprintf(stderr, "Removed approximately   %8luk\n", bytes_removed / 1024);
            fprintf(stderr, "Article lines processed %8lu\n", ids_processed);
            fprintf(stderr, "Articles retained       %8lu\n", ids_kept);
            fprintf(stderr, "Entries expired         %8lu\n", ids_expired);
            fprintf(stderr, "Files unlinked          %8lu\n", files_deleted);
            if(dirs_removed)
                fprintf(stderr, "Directories removed     %8lu\n", dirs_removed);
            fprintf(stderr, "Old entries dropped     %8lu\n", old_dropped);
            fprintf(stderr, "Old entries retained    %8lu\n", old_kept);
        }
        if(ids_expired == 0)
            result = 2;
    }
    if(result == 2)
        syslog(LOG_INFO, "Nothing to expire after %lu second(s)",
               time(NULL) - now);
    else if(result)
        syslog(LOG_INFO, "Expire returns error %d after %lu second(s)",
               result, time(NULL) - now);
    else
        syslog(LOG_INFO, "Expire processed %lu entries in %lu second(s)",
               ids_processed + old_dropped + old_kept, time(NULL) - now);

    if(fppid) {
        fclose(fppid);
        unlink(cfg.pidfile);
    }

    return(result);
}

/************************************************************************/
/*                                                                      */
/* Parse debug options                                                  */
/*                                                                      */
/* This has been moved to an own procedure in order to overcome trouble */
/* with VisualAge for Windows.                                          */
/*                                                                      */
/************************************************************************/
static void parsedbgopt(char *opt)
{
    while(*opt) {
        switch(*opt) {
           case 'a':
               debug_active(1);
               break;
           case 'e':
               cfg.logflg |= LOG_EXPIRE;
               break;
           case 'h':
#ifdef USE_REBUILD
               ndbm_debug(1);
#endif
               break;
           case 't':
               cfg.logflg |= LOG_TESTMODE;
               fprintf(stderr, "- Testmode -\n");
               break;
           case 'o':
               syslog_opt |= LOG_PERROR;
               break;
           case 'p':
               syslog_opt |= LOG_PID;
               break;
           default:
               fprintf(stderr, "Unknown option -d%c ignored\n", *opt);
               break;
        }
        opt++;
    }
}

/************************************************************************/
/*                                                                      */
/* Parse exclude options                                                */
/*                                                                      */
/* This has been moved to an own procedure in order to overcome trouble */
/* with VisualAge for Windows.                                          */
/*                                                                      */
/************************************************************************/
static void parsexclopt(char *opt)
{
    while(*opt) {
        switch(*opt) {
           case 'a':
               cfg.noactive = 1;
               break;
           case 'h':
               cfg.noupdate = 1;
               break;
           case 'r':
               cfg.normempty = 1;
               break;
           case 'S':
               cfg.xcl_size = 1;
               break;
           case 's':
               cfg.noserver = 1;
               break;
           case 'T':
               cfg.xcl_time = 1;
               break;
           default:
               fprintf(stderr, "Unknown option -x%c ignored", *opt);
               break;
        }
        opt++;
    }
}

/************************************************************************/
/*                                                                      */
/*  Usage display                                                       */
/*                                                                      */
/************************************************************************/
static void usage(void)
{
    fputs("usage: expire [options] [newsgroups]\n\n"
          "  newsgroups is a comma separated list of groups, which may\n"
          "             include '!', '*', '?' and '[...]'\n\n"
          "options:\n"
          "  -c<fname>  control file      -a rebuild active\n"
          "  -d<dflags> logging flags     -A create active\n"
          "  -E<time>   remember time     -b remove bad arts\n"
          "  -e<time>   expire time       -C cancel expired articles\n"
          "  -q<lines>  max bad lines     -h ignore history\n"
          "  -R<reason> define reason     -i ignore expire lines\n"
          "  -s<call>   sort program      -L with low priority\n"
          "  -S<fname>  summary file      -M use last write\n"
          "  -t<dir>    temp directory    -p use posting date\n"
          "  -u<user>   user name         -r rebuild\n"
          "  -w<time>   time warp         -T sort active\n"
          "  -x<xflags> exclusion flags   -U old index\n"
          "  -z<fname>  use zapfile       -V verify history only\n"
          "                               -v verbose\n\n"
          "logging flags:               exclusion flags:\n"
          "  a  log active update         a  no active update\n"
          "  e  log expiration            h  no history update\n"
          "  h  log history               r  no directory remove\n"
          "  o  log to console            S  exclude size\n"
          "  p  add pid to log            s  no server control\n"
          "  t  test mode                 T  exclude time\n", stderr);
}

/************************************************************************/
/*                                                                      */
/* Signal handler                                                       */
/*                                                                      */
/************************************************************************/
static void cleanup(int sig)
{
    char *msg = "Unknown signal";
    int level = LOG_INFO;

    signal(sig, SIG_IGN);
    switch(sig) {
    case SIGSEGV:
        level = LOG_CRIT;
        msg = "Internal error";
        break;
    case SIGTERM:
        msg = "Killed";
        break;
    case SIGABRT:
        level = LOG_CRIT;
        msg = "Aborted";
        break;
    case SIGBREAK:
    case SIGINT:
        msg = "Interrupted";
        break;
    }
    syslog(level, msg);
    fprintf(stderr, "\n%s\n", msg);
    if(cfg.pidfile[0] != '\0') {
        if(fppid)
            fclose(fppid);
        unlink(cfg.pidfile);
    }
    signal(sig, SIG_DFL);
    signal(SIGABRT, SIG_DFL);
    abort();
}

/*
 * $Log: expire.c,v $
 * Revision 1.1  1997/02/13 11:21:45  oliver
 * initial revision
 *
 */
