/*
 * This program was written by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 * This module contains a chaotic collection of useful routines.
 *
 */

const char misc_rcsid[]="$Id: misc.c,v 1.1 1997/02/13 11:21:49 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <sys/stat.h>

#include <syslog.h>
#include <xfile.h>

#include <chlimits.h>
#include <nactive.h>

#include "config.h"
#include "expire.h"

extern FILE *zapfile;

/************************************************************************/
/*                                                                      */
/************************************************************************/
int hist_print(FILE *fp, char *id, long arrived, long expires, long posted,
               long artsize, char **groups)
{
    char **gp;

    if(fprintf(fp, "%s %lu", id, arrived) <= 0)
        abort();
    if(!cfg.xcl_time) {
        if(expires) {
            if(fprintf(fp, "~%ld", expires) <= 0)
                abort();
        }
        else if(fputs("~-", fp) < 0)
            abort();
        if(fprintf(fp, "~%lu", posted) <= 0)
            abort();
    }
    if(!cfg.xcl_size && fprintf(fp, " %lu", artsize) <= 0)
        abort();
    if(fprintf(fp, " %s", *groups) <= 0)
        abort();
    for(gp = groups + 1; *gp; gp++)
        if(fprintf(fp, ",%s", *gp) <= 0)
            abort();
    if(fputc('\n', fp) == EOF)
        abort();

    return(0);
}

/************************************************************************/
/*                                                                      */
/*                                                                      */
/* Returns 0 if at least one article file has been successfully removed */
/*                                                                      */
/*                                                                      */
/************************************************************************/
int remove_art(char **groups, long artsize)
{
    int result = -1;
    int rc = 0;
    int first = 0;
    char *cp;
    struct stat ast;
    char **gp;
    char *pathname = malloc(_MAX_PATH);

    for(gp = groups; *gp; gp++) {
        if(grouppath(pathname, cfg.newsdir, *gp) == NULL) {
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "Groupname too long %s", *gp);
            continue;
        }
        if((cp = strrchr(pathname, ':')) == NULL) {
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "Bad path %s", pathname);
            continue;
        }
        *cp = '\\';
        if(artsize == 0 && stat(pathname, &ast) == 0)
            artsize = ast.st_size;
        if(DOLOG(LOG_EXPIRE))
            syslog(LOG_DEBUG, "%selete %s",
                   DOLOG(LOG_TESTMODE) ? "Would d" : "D", pathname);
        if(zapfile)
            fprintf(zapfile, "DEL %s\n", pathname);
        else if(!DOLOG(LOG_TESTMODE))
            rc = xremove(pathname);
        if(rc == 0) {
            result = 0;
            files_deleted++;
            bytes_removed += artsize;
            if((cp = strrchr(strcpy(pathname, *gp), ':')) != NULL)
                *cp = '\0';
        }
        else if(DOLOG(LOG_EXPIRE))
            syslog(LOG_DEBUG, "%s", _strerror(pathname));

        /*
         * Clear 'artsize' after the first entry to force
         * redetermination of the files size, because crossposts
         * might be soft-linked and are therefore smaller then
         * the first article file.
         */
        if(first) {
            artsize = 0;
            first = 0;
        }
    }
    free(pathname);

    return(result);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int is_expired(char **groups, long now, long arrived, long expires, long posted)
{
    char **gp;
    char *cp;
    long minkeep, maxkeep;

    /*
     * After assuming that the article has expired we will check
     * the time to keep this article for every group and eventually
     * revise this decision.
     */
    for(gp = groups; *gp; gp++) {
        /*
         * We accept a colon following the groupname
         */
        if((cp = strrchr(*gp, ':')) != NULL)
            *cp = '\0';

        /*
         * Ignore unknown groups
         */
        if(query_expiration(*gp, &minkeep, &maxkeep)) {
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "Ignore unknown group %s", *gp);
            if(cp)
                *cp = ':';
            continue;
        }
        if(cp)
            *cp = ':';
        if(DOLOG(LOG_EXPIRE))
            syslog(LOG_DEBUG, "Keep %s %lds - %lds", *gp, minkeep, maxkeep);

        /*
         * First check the minimum time to keep this
         * article.
         */
        if(cfg.usepost) {
            if(posted + minkeep > now) {
                if(DOLOG(LOG_EXPIRE))
                    syslog(LOG_DEBUG, "Posted %lds, keep %lds", posted, posted + minkeep - now);
                break;
            }
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "Posted %lds, %lds too old", posted, now - posted - minkeep);
        }
        else if(arrived + minkeep > now) {
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "Arrived %lds, keep %lds", arrived, arrived + minkeep - now);
            break;
        }
        if(DOLOG(LOG_EXPIRE))
            syslog(LOG_DEBUG, "Arrived %lds, %lds too old", arrived, now - arrived - minkeep);

        /*
         * See if this article contains an explicit
         * expiration time and hasn't reached neither
         * this one nor the maxkeep time.
         */
        else if(!cfg.ignorexp && expires && expires > now &&
                (arrived + maxkeep) > now) {
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "Expire/max date not reached");
            break;
        }
        if(DOLOG(LOG_EXPIRE))
            syslog(LOG_DEBUG, "Drop %s", *gp);
    }
    return(*gp == NULL);
}

/************************************************************************/
/*                                                                      */
/* parse_historyline                                                    */
/*                                                                      */
/* Parse a history lines. We assume the following format:               */
/*                                                                      */
/* <id> <time> <size> <group:number>[[,<group:number>]...]              */
/*                                                                      */
/* Remembered articles do only contain                                  */
/*                                                                      */
/* <id> <time>                                                          */
/*                                                                      */
/* followed by any number of spaces.                                    */
/*                                                                      */
/* <id> is the message id                                               */
/* <time> contains three time values representing the number of         */
/*        seconds since 01/01/70 and separated by a tilde               */
/*        character: arrived~expires~posted                             */
/* <size> is the article's size in bytes                                */
/* <group:number> is the name of newsgroup and the local number         */
/*                of the article in that group                          */
/*                                                                      */
/* However, this routine is very robust. It will accept lines without   */
/* the <size> field and without expire and/or posting dates.            */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* NONE                                                                 */
/* ------------ ------------------------------------------------------- */
/* Return value -1 if the line is broken and should be discarded        */
/*               0 if the line specifies a normal entry                 */
/*               1 if the line specifies a remembered article           */
/*               2 if the line specifies a canceled article             */
/*                                                                      */
/*                                                                      */
/************************************************************************/
int parse_historyline(char *line, long *arrived, long *expires, long *posted,
                      long *artsize, char **grouplist, char **wipe)
{
    register char *cp = line;

    /*
     * Preset globals
     */
    *expires = 0;
    *posted = 0;
    *artsize = 0;

    /*
     * Each line must begin with the open angle bracket,
     * the first character of each message-id.
     */
    if(*cp != '<')
        return(-1);   /* illegal line */

    /*
     * Convert tabs to spaces and chop off carriage return and line feed.
     */
    while(*cp) {
        if(*cp == '\t')
            *cp = ' ';
        else if(*cp == '\n' || *cp == '\r') {
            *cp = '\0';
            break;
        }
        cp++;
    }

    /*
     * Skip the message-id
     */
    cp = line;
    while(*cp && *cp != ' ')
        cp++;
    if(*cp == '\0')
        return(2);   /* msg-id only, article canceled with old method */

    if(wipe == NULL)
        *cp++ = '\0';

    /*
     * Skip delimiter
     */
    while(*cp == ' ')
        cp++;
    if(*cp == '\0')
        return(2);  /* msg-id only, article canceled with old method */

    /*
     * Read and skip the first time field, the arrival time
     */
    *arrived = atol(cp);
    while(*cp && *cp != ' ' && *cp != '~')
        cp++;
    if(*cp == '\0')
        return(1);   /* msg-id plus single time, article remembered */

    if(wipe)
        *wipe = cp;

    /*
     * Read and skip additional time fields.
     */
    if(*cp == '~') {
        cp++;
        *expires = atol(cp);
        while(*cp && *cp != ' ' && *cp != '~')
            cp++;
        if(*cp == '~') {
            cp++;
            *posted = atol(cp);
            while(*cp && *cp != ' ')
                cp++;
        }
    }
    while(*cp == ' ')
        cp++;
    if(*cp == '\0')
        return(1);   /* msg-id plus time, article remembered */

    /*
     * Read and skip optional size field
     */
    if(isdigit(*cp)) {
        *artsize = atol(cp);
        while(*cp && *cp != ' ')
            cp++;
        if(*cp == '\0')
            return(1);   /* somehow broken, but treated as remembered */
        while(*cp == ' ')
            cp++;
    }
    *grouplist = cp;

    return(0);
}

/************************************************************************/
/* grouppath                                                            */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* path         Points to an array of at least _MAX_PATH bytes in size. */
/************************************************************************/
char *grouppath(char *path, char *newsdir, char *group)
{
    char *ngpath = NULL;
    register char *cp;

    if((strlen(newsdir) + strlen(group) + 12) >= _MAX_PATH)
        fprintf(stderr, "File path too long for %s", group);
    else {
        strcpy(path, newsdir);
        strcat(path, "\\");
        strcat(path, group);
        cp = strrchr(path, '\\');
        while((cp = strchr(cp, '.')) != NULL)
            *cp = '\\';
        ngpath = path;
    }
    return(ngpath);
}

/*
 * $Log: misc.c,v $
 * Revision 1.1  1997/02/13 11:21:49  oliver
 * initial revision
 *
 */
