/*
 * This program was written by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 * This module contains routines to purge articles by scanning
 * through all groups listed in the active file.
 *
 */

const char purgebya_rcsid[]="$Id: purgebya.c,v 1.1 1997/02/13 11:21:50 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <direct.h>

#include <sys\types.h>
#include <sys\stat.h>

#include <chlimits.h>
#ifdef __OS2__
#include <pipe.h>
#endif
#include <match.h>

#include <syslog.h>
#include <xfile.h>
#include <split.h>
#include <arpatime.h>
#include <sodtgram.h>
#include <dircache.h>
#include <nactive.h>

#include "config.h"
#include "rebuild.h"
#include "histhead.h"
#include "expire.h"

static char headitem[HDI_LAST][MAX_BUFSIZ];

static long arrived = 0;
static long expires = 0;
static long posted = 0;
static long artsize = 0;

/************************************************************************/
/*                                                                      */
/*  purge_by_active                                                     */
/*                                                                      */
/*  Purges articles following the active file. Ignores previously build */
/*  history.                                                            */
/*                                                                      */
/************************************************************************/
int purge_by_active(void)
{
    int result = 0;
    char ngname[MAX_GROUPNAME];
    char ngdir[_MAX_PATH];
    char pathname[_MAX_PATH];
    char sortcall[_MAX_PATH];
    char *artpath;
    long *art_array = NULL;
    unsigned int i;
    int num_arts;
    FILE *fphist;
    char **groups;
    int skipped = 0;
    int sort_result;
#ifndef __OS2__
    char *tname;
#endif

    /*
     * Open a temporary history file. We'll build
     * it's index later.
     */
#ifdef __OS2__
    sprintf(sortcall, "%s%s.seq", cfg.sortcall, cfg.historyfile);
    if((fphist = popen(sortcall, "w")) == NULL) {
        syslog(LOG_ERR, "Failed to spawn %s", sortcall);
        return(-1);
    }
#else
    tname = tempnam(cfg.workdir, "PIP");
    sprintf(sortcall, "%s%s.seq < %s", cfg.sortcall, cfg.historyfile, tname);
    if((fphist = fopen(tname, "w")) == NULL) {
        syslog(LOG_ERR, "%s", _strerror(tname));
        return(-1);
    }
#endif

    /*
     * Loop for all lines in active file
     */
    if(cfg.verbose)
        fputs("Scanned Expired Remembr    Kept Bad Group\n", stderr);
    ngname[0] = '\0';
    while(next_group(ngname)) {
        if(cfg.verbose)
            fprintf(stderr, "\r%7u %7u %7u %7u %3u %-42.42s",
                    ids_processed, files_deleted, old_kept, ids_kept, skipped, ngname);

        /*
         * Process specified newsgroups only
         */
        if(!ngmatch(ngname, cfg.groups)) {
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "Group %s ignored", ngname);
            continue;
        }

        /*
         * If no articles were found in this group, try to remove
         * the overview file and the newsgroup directory.
         */
        if((num_arts = query_groupdir(ngname, &art_array)) < 1) {
            if(DOLOG(LOG_EXPIRE))
                syslog(LOG_DEBUG, "No articles found in %s", ngname);
            if(!cfg.normempty && grouppath(ngdir, cfg.newsdir, ngname)) {
                artpath = ngdir + strlen(ngdir);
                strcpy(artpath, "\\.overview");
                remove(ngdir);
                *artpath = '\0';
                if(rmdir(ngdir) == 0)
                    dirs_removed++;
            }
            continue;
        }
        if(DOLOG(LOG_EXPIRE))
            syslog(LOG_DEBUG, "%u article%s found in %s", num_arts,
                   num_arts != 1 ? "s" : "", ngname);

        /*
         * Loop to process each article found in this
         * newsgroup directory.
         * Note, that we let 'artpath' point to the end
         * of the directory name, so we can easily build
         * absolute paths.
         */
        grouppath(ngdir, cfg.newsdir, ngname);
        artpath = ngdir + strlen(ngdir);
        *artpath++ = '\\';
        for(i = 0; i < num_arts; i++) {

            /*
             * Read all header lines we are interested in.
             * 'read_header' returns -1, if we can't find
             * all needed headerlines or even can't read
             * the article file itself.
             */
            ltoa(art_array[i], artpath, 10);
            if(read_histheader(ngdir, headitem, &arrived, &artsize)) {
                if(cfg.removebad) {
                    files_deleted++;
                    bytes_removed += artsize;
                    xremove(ngdir);
                }
                if(++skipped >= cfg.quitbad && cfg.quitbad) {
                    syslog(LOG_ERR, "Too many errors");
                    result = -1;
                    break;
                }
                continue;
            }

            /*
             * Split the Xref line into an array of article
             * references.
             */
            if(headitem[HDI_Xref][0])
                groups = split_list(headitem[HDI_Xref], ' ');
            else {
                /*
                 * If this is not a crossposting, then we won't have
                 * an Xref line. Let's create a list with one group
                 * only. The first item must contain the nodename,
                 * the last is set to NULL.
                 */
                groups = malloc(3 * sizeof(char *));
                groups[0] = "dummy";
                groups[1] = malloc(_MAX_PATH);
                strcat(strcat(strcpy(groups[1], ngname), ":"), artpath);
                groups[2] = NULL;
            }
            ids_processed++;

            posted = getadate(headitem[HDI_Date]);
            expires = getadate(headitem[HDI_Expires]);
            if(is_expired(groups + 1, now + cfg.timewarp, arrived, expires, posted)) {
                /*
                 * Delete article. This will delete all crosspostings
                 * too, so we won't come across this article again.
                 * Therefore we store ids to remember at this point,
                 * even if the current group is not the one in front
                 * of our Xref line.
                 */
                ids_expired++;
                remove_art(groups + 1, artsize);
                if(cfg.forgetsecs && (arrived + cfg.forgetsecs) > (now + cfg.timewarp)) {
                    if(DOLOG(LOG_EXPIRE))
                        syslog(LOG_DEBUG, "Remember %s", headitem[HDI_Message_ID]);
                    if(fprintf(fphist, "%s %lu\n", headitem[HDI_Message_ID], arrived) <= 0)
                        abort();
                    old_kept++;
                }
            }
            else {
                /*
                 * Keep article
                 */
                ids_kept++;

                /*
                 * Add kept article to the history file only if this
                 * is the first existing article file of all crossposts
                 * mentioned in the Xref line.
                 */
                hist_print(fphist, headitem[HDI_Message_ID], arrived, expires,
                           posted, artsize, groups + 1);
            }
            if(cfg.verbose  && !(ids_processed % 100))
                fprintf(stderr, "\r%7u %7u %7u %7u %3u %-42.42s",
                        ids_processed, files_deleted, old_kept, ids_kept, skipped, ngname);

            if(headitem[HDI_Xref][0] == '\0')
                free(groups[1]);
            free(groups);
        } /* Loop articles */
        free(art_array);
        if(result)
            break;

    } /* Loop newsgroups */

    if(cfg.verbose)
        fprintf(stderr, "\r%7u %7u %7u %7u %3u %-42.42s\nSorting....",
                ids_processed, files_deleted, old_kept, ids_kept, skipped, ngname);

#ifdef __OS2__
    if((sort_result = pclose(fphist)) != 0) {
        syslog(LOG_ERR, "'%s' returns error %d", sortcall,
               (sort_result & 0xffff) >> 8);
        return(-1);
    }
#else
    if(fclose(fphist)) {
        syslog(LOG_ERR, "%s", _strerror(tname));
        return(-1);
    }
    if((sort_result = system(sortcall)) != 0) {
        syslog(LOG_ERR, "'%s' returns error %d", sortcall, sort_result);
        return(-1);
    }
    if(unlink(tname))
        syslog(LOG_WARNING, "%s", _strerror(tname));
    free(tname);
#endif

    /*
     * Rebuild history
     */
    strcat(strcpy(pathname, cfg.historyfile), ".seq");
    if(!result) {
        if(cfg.verbose)
            fputs("Rebuilding history....", stderr);
        if(!cfg.noupdate && !DOLOG(LOG_TESTMODE)) {
            /*
             * Pause the server
             */
            if(!cfg.noserver) {
                sprintf(server_cmd, "%s\001m\001%s", cfg.logname, cfg.expreason);
                if(SoDtgramTransact(cfg.controlhost, cfg.controlport, 119,
                                    server_cmd, server_cmd, 30) == -1)
                    syslog(LOG_ERR, "Can't pause server");
            }

#ifdef USE_REBUILD
            if(cfg.oldindex)
                result = rebuild_oldhistory(pathname);
            else
                result = rebuild_history(pathname);
#else
            if(!cfg.oldindex)
                result = rebuild_memory(pathname);
#endif
        }
        if(cfg.verbose)
            fputs(result ? "failed\n" : "done\n", stderr);
    }
    if(xremove(pathname))
        syslog(LOG_ERR, "Delete %s", _strerror(pathname));

    return(result);
}

/*
 * $Log: purgebya.c,v $
 * Revision 1.1  1997/02/13 11:21:50  oliver
 * initial revision
 *
 */
