/*
 * This program was written by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 * This module contains routines to purge article by scanning
 * the history file.
 *
 */

const char purgebyh_rcsid[] = "$Id: purgebyh.c,v 1.1 1997/02/13 11:21:51 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include <sys\types.h>
#include <sys\stat.h>

#include <syslog.h>
#include <chlimits.h>
#include <split.h>
#include <sodtgram.h>
#include <xfile.h>
#include <arpatime.h>

#include "config.h"
#include "rebuild.h"
#include "histhead.h"
#include "expire.h"

static char headitem[HDI_LAST][MAX_BUFSIZ];

static char line[MAX_HISTORYLINE];
static char *grouplist = NULL;
static long arrived = 0;
static long expires = 0;
static long posted = 0;
static long artsize = 0;

/************************************************************************/
/*                                                                      */
/*  purge_by_history                                                    */
/*                                                                      */
/*  Sequentially reads the history file to purges outdated articles.    */
/*                                                                      */
/*  Returns 0 if the history file has been rebuild, -1 on errors.       */
/*                                                                      */
/************************************************************************/
int purge_by_history(void)
{
    int result;
    FILE *fp, *fpnew;
    char filename[_MAX_PATH];
    char pathname[_MAX_PATH];
    int skipped = 0;
    char **groups;
    char *cp;
    long remember;

    /*
     * Open an existing history file for reading
     */
    strcat(strcpy(filename, cfg.historyfile), ".pag");
    if((fp = xfopen(filename, "r")) == NULL) {
        syslog(LOG_ERR, "%s", _strerror(filename));
        return(-1);
    }

    /*
     * Open a temporary history file. We'll build
     * it's index later.
     */
    strcat(strcpy(filename, cfg.historyfile), ".seq");
    if((fpnew = xfopen(filename, "w")) == NULL) {
        syslog(LOG_ERR, "%s", _strerror(filename));
        fclose(fp);
        return(-1);
    }

    /*
     * Calculate remember time.
     */
    if(cfg.forgetsecs)
        remember = now + cfg.timewarp - cfg.forgetsecs;
    else
        remember = now + cfg.timewarp - cfg.expiresecs;

    /*
     * Loop for all lines in the history file
     */
    if(cfg.verbose)
        fputs("Scanned Expired Remembr    Kept Bad Age\n", stderr);
    while(fgets(line, MAX_HISTORYLINE, fp)) {

        /*
         * Discard broken lines
         */
        if((result = parse_historyline(line, &arrived, &expires, &posted,
                                       &artsize, &grouplist, NULL)) == -1) {
            if(++skipped >= cfg.quitbad && cfg.quitbad)
                break;
            result = 0;
            continue;
        }

        /*
         * Discard cancels with old format, which do not
         * specify the arrival time.
         */
        if(result == 2) {
            result = 0;
            old_dropped++;
            continue;
        }

        /*
         * See if we should remember expired msg-ids
         */
        if(result) {
            result = 0;
            if(arrived < remember) {
                old_dropped++;
                continue;
            }
            old_kept++;
            fprintf(fpnew, "%s %lu\n", line, arrived);
        }

        else {
            ids_processed++;

            /*
             * Check any group of crossposts
             */
            groups = split_list(grouplist, ',');

            /*
             * If posting date is 0 then we probably got an
             * old style entry. In that case we must get
             * posting and expiration date from the article
             * file itself.
             */
            if(posted == 0) {
                if(grouppath(pathname, cfg.newsdir, *groups) != NULL) {
                    if((cp = strrchr(pathname, ':')) != NULL) {
                        *cp = '\\';
                        if(read_histheader(pathname, headitem, &arrived, &artsize) == 0) {
                            posted = getadate(headitem[HDI_Date]);
                            expires = getadate(headitem[HDI_Expires]);
                        }
                    }
                }
                if(posted == 0)
                    posted = arrived;
            }

            if(is_expired(groups, now + cfg.timewarp, arrived, expires, posted)) {
                /*
                 * Delete article
                 */
                ids_expired++;
                remove_art(groups, artsize);

                /*
                 * Add msg-ids to remember to the history file
                 */
                if(arrived > remember) {
                    old_kept++;
                    fprintf(fpnew, "%s %lu\n", line, arrived);
                }
            }
            else {
                /*
                 * Keep article
                 */
                ids_kept++;
                hist_print(fpnew, line, arrived, expires, posted, artsize, groups);
            }
            if(cfg.verbose  && !(ids_processed % 100))
                fprintf(stderr, "\r%7u %7u %7u %7u %3u%4d  ", ids_processed,
                        files_deleted, old_kept, ids_kept, skipped,
                        (now - arrived) / 86400);

            free(groups);
        }
    } /* Loop history lines */

    if(cfg.verbose)
        fprintf(stderr, "\r%7u %7u %7u %7u %3u%4d\n", ids_processed,
                files_deleted, old_kept, ids_kept, skipped,
                (now - arrived) / 86400);

    if(fclose(fpnew)) {
        syslog(LOG_ERR, "%s", _strerror("Closing new history"));
        return(-1);
    }
    if(fclose(fp)) {
        syslog(LOG_ERR, "%s", _strerror("closing history"));
        return(-1);
    }

    /*
     * Rebuild history
     */
    strcat(strcpy(pathname, cfg.historyfile), ".seq");
    if(!result && (ids_expired || old_dropped || skipped)) {
        if(cfg.verbose)
            fputs("Rebuilding history....", stderr);
        if(!cfg.noupdate && !DOLOG(LOG_TESTMODE)) {
            /*
             * Pause the server
             */
            if(!cfg.noserver) {
                sprintf(server_cmd, "%s\001m\001%s", cfg.logname, cfg.expreason);
                if(SoDtgramTransact(cfg.controlhost, cfg.controlport, 119,
                                    server_cmd, server_cmd, 30) == -1)
                    syslog(LOG_ERR, "Can't pause server");
            }

#ifdef USE_REBUILD
            if(cfg.oldindex)
                result = rebuild_oldhistory(pathname);
            else
                result = rebuild_history(pathname);
#else
            if(!cfg.oldindex)
                result = rebuild_memory(pathname);
#endif
        }
        if(cfg.verbose)
            fputs(result ? "failed\n" : "done\n", stderr);
    }
    if(xremove(pathname))
        syslog(LOG_ERR, "Delete %s", _strerror(pathname));

    return(result);
}

/*
 * $Log: purgebyh.c,v $
 * Revision 1.1  1997/02/13 11:21:51  oliver
 * initial revision
 *
 */
