const char dircache_rcsid[] = "$Id$";

/*
 * This part was done by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 * This module maintains a directory cache for newsgroup directories.
 * The routines are called by the active and overview file procedures.
 */

/*
 * $Log$
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <syslog.h>

#include <mutex.h>

#include <scangrpd.h>
#include <dircache.h>

static long mutex_wait = 30000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

typedef struct _DIRBUFF {
    struct _DIRBUFF *next;
    struct _DIRBUFF *prev;
    char group[256];
    int  numarts;
    long *art_array;
} DIRBUFF;

typedef struct {
    int n_buffs;
    DIRBUFF *head;
    DIRBUFF *tail;
} DIRCACHE;

static DIRCACHE gdc;
static char newspath[_MAX_PATH];
static int debug_on = 0;
static int  maxcache = 32;
static long mutex = -1;

/*======================================================================*/
/* LOCAL ROUTINES                                                       */
/*======================================================================*/

/************************************************************************/
/*                                                                      */
/* load_groupdir                                                        */
/*                                                                      */
/* Creates a directory buffer for a specified newsgroup and scans the   */
/* corresponding directory.                                             */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to a zero teminated string which specifies the   */
/*              name of the newsgroup.                                  */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to a directory buffer structure on success or   */
/*              NULL in case of any error.                              */
/*                                                                      */
/************************************************************************/
static DIRBUFF *load_groupdir(const char *group)
{
    int sz = 0;
    DIRBUFF *gdp = malloc(sizeof(DIRBUFF));

    if(gdp) {
        memset(gdp, 0, sizeof(DIRBUFF));
        strcpy(gdp->group, group);
        gdp->numarts = scan_groupdir(group, newspath, &(gdp->art_array), &sz);
        if(debug_on)
            syslog(LOG_DEBUG, "Found %d articles in %s", gdp->numarts, group);
    }
    else
        syslog(LOG_ERR, "Out of memory");

    return(gdp);
}

/************************************************************************/
/*                                                                      */
/* fetch_groupdir                                                       */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to a zero teminated string which specifies the   */
/*              name of the newsgroup.                                  */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to a directory buffer structure on success or   */
/*              NULL in case of any error.                              */
/*                                                                      */
/************************************************************************/
static DIRBUFF *fetch_groupdir(const char *group)
{
    DIRBUFF *gdp;

    /*
     * If our cache is empty, then load the first directory.
     */
    if(gdc.n_buffs == 0) {
        if(debug_on)
            syslog(LOG_DEBUG, "First dircache entry %s", group);
        if((gdc.head = gdc.tail = load_groupdir(group)) != NULL)
            gdc.n_buffs = 1;
        return(gdc.head);
    }

    /*
     * Our group is the last recently used.
     */
    if(strcmp(gdc.head->group, group) == 0) {
        if(debug_on)
            syslog(LOG_DEBUG, "Hit recent dircache entry %s", group);
        return(gdc.head);
    }

    /*
     * Our group is _not_ the recently used one. See, if we can locate
     * it in our cache.
     */
    for(gdp = gdc.head; (gdp = gdp->next) != NULL; ) {
        /*
         * Search the requested group.
         */
        if(strcmp(gdp->group, group))
            continue;

        /*
         * We have found it. Maintain our linked list by moving this
         * directory info to the head.
         */
        if(gdp->next)
            gdp->next->prev = gdp->prev;
        else
            gdc.tail = gdp->prev;
        gdp->prev->next = gdp->next;
        gdp->next = gdc.head;
        gdp->prev = NULL;
        gdc.head->prev = gdp;
        if(debug_on)
            syslog(LOG_DEBUG, "Hit dircache entry %s", group);
        return(gdc.head = gdp);
    }

    /*
     * We reach this point, if the requested group is not in our
     * cache. If our cache is full, we save and remove the tail.
     */
    if(gdc.n_buffs >= maxcache) {
        gdp = gdc.tail;
        if(debug_on)
            syslog(LOG_DEBUG, "Remove %s from dircache", gdp->group);
        gdc.tail = gdp->prev;
        gdp->prev->next = NULL;
        gdc.n_buffs--;
        gdc.head->prev = NULL;
        free(gdp->art_array);
        free(gdp);
    }
    if(debug_on)
        syslog(LOG_DEBUG, "Add %s to dircache", group);
    gdp = load_groupdir(group);
    gdp->next = gdc.head;
    gdc.head->prev = gdp;
    gdc.head = gdp;
    gdc.n_buffs++;
    return(gdp);
}


/*======================================================================*/
/* GLOBAL ROUTINES                                                      */
/*======================================================================*/

/************************************************************************/
/*                                                                      */
/* debug_dircache                                                       */
/*                                                                      */
/* Switch directory cache debugging on/off.                             */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* enable       A value of 0 will switch off debugging, otherwise it    */
/*              is switched on.                                         */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void debug_dircache(int enable)
{
    debug_on = enable;
}

/************************************************************************/
/*                                                                      */
/* dircache_maxwait                                                     */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void dircache_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/*                                                                      */
/* init_dircache                                                        */
/*                                                                      */
/* Initializes directory cache processing.                              */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* newsdir      Points to the string that specifies the path to the     */
/*              news home directory.                                    */
/*                                                                      */
/* cachesize    Specifies the maximum number of directories being       */
/*              cached.                                                 */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on failures                            */
/*                                                                      */
/************************************************************************/
int init_dircache(const char *newsdir, int cachesize)
{
    if(debug_on)
        syslog(LOG_DEBUG, "Init cache for %d entries below %s", cachesize, newsdir);
    strcpy(newspath, newsdir);
    maxcache = cachesize;

    if((mutex = create_mutex()) == -1)
        return(-1);
    return(0);
}

/************************************************************************/
/*                                                                      */
/* flush_dircache                                                       */
/*                                                                      */
/* Releases any resources claimed by the directory cache processing.    */
/*                                                                      */
/************************************************************************/
void flush_dircache(void)
{
    DIRBUFF *gdp, *gdn;
    int rc;
    int retries = 3;

    if(mutex != -1) {
        if(debug_on)
            syslog(LOG_DEBUG, "Flush directory cache");

        /*
         * Try to get exclusive access several times.
         * Do not flush, if we fail.
         */
        while(retries--)
            if((rc = request_mutex(&mutex, MUTEXWAIT)) == 0)
                break;
        if(rc)
            return;

        /*
         * Free allocated storage.
         */
        for(gdp = gdc.head; gdp != NULL; gdp = gdn) {
            gdn = gdp->next;
            free(gdp->art_array);
            free(gdp);
        }

        /*
         * Reset all to zero.
         */
        gdc.n_buffs = 0;
        gdc.head = gdc.tail = NULL;

        /*
         * Release exclusive access.
         */
        release_mutex(mutex);
    }
}

/************************************************************************/
/*                                                                      */
/* exit_dircache                                                        */
/*                                                                      */
/* This routine should be called before program termination.            */
/*                                                                      */
/************************************************************************/
void exit_dircache(void)
{
    if(mutex != -1) {
        flush_dircache();
        destroy_mutex(mutex);
        mutex = -1;
    }

}

/************************************************************************/
/*                                                                      */
/* query_groupinfo                                                      */
/*                                                                      */
/* Requests the number of articles and the lowest and highest article   */
/* number of a given newsgroup.                                         */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be queried.                                */
/*                                                                      */
/* lo           Pointer to a long, where this routine will store        */
/*              the lowest article number of the specified group.       */
/*              A NULL pointer may be supplied if the caller is not     */
/*              interested in this value.                               */
/*                                                                      */
/* hi           Pointer to a long, where this routine will store        */
/*              the highest article number of the specified group.      */
/*              A NULL pointer may be supplied if the caller is not     */
/*              interested in this value.                               */
/* ------------ ------------------------------------------------------- */
/* Return value The number of articles in the newsgroup directory or    */
/*              -1 in case of any failure.                              */
/*                                                                      */
/************************************************************************/
int query_groupinfo(const char *group, long *lo, long *hi)
{
    DIRBUFF *gdp;
    int n = -1;

    if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        gdp = fetch_groupdir(group);
        if((n = gdp->numarts) > 0) {
            if(lo)
                *lo = gdp->art_array[0];
            if(hi)
                *hi = gdp->art_array[n - 1];
        }
        release_mutex(mutex);
    }
    return(n);
}


/************************************************************************/
/*                                                                      */
/* query_groupdir                                                       */
/*                                                                      */
/* Requests an array of all article numbers of a given newsgroup.       */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be queried.                                */
/*                                                                      */
/* art_array    Points to an array pointer. The array will be allocated */
/*              within this routine and must be freed by the caller.    */
/*              If no articles were found, no memory will be allocated. */
/* ------------ ------------------------------------------------------- */
/* Return value The number of articles in the newsgroup directory or    */
/*              -1 in case of any failure.                              */
/*                                                                      */
/************************************************************************/
int query_groupdir(const char *group, long **art_array)
{
    DIRBUFF *gdp;
    int n = -1;

    if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        gdp = fetch_groupdir(group);
        if((n = gdp->numarts) > 0) {
            *art_array = malloc(n * sizeof(long));
            memcpy(*art_array, gdp->art_array, n * sizeof(long));
        }
        release_mutex(mutex);
    }
    return(n);
}

/************************************************************************/
/*                                                                      */
/* query_nextart                                                        */
/*                                                                      */
/* Requests the next article number of a group.                         */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be queried.                                */
/*                                                                      */
/* art_num      The current article number. The routine will return     */
/*              the number of the article following this one. If this   */
/*              value is set to zero, then the number of the first      */
/*              article will be returned.                               */
/* ------------ ------------------------------------------------------- */
/* Return value The article number following 'art_num' or 0 in case of  */
/*              any failure.                                            */
/*                                                                      */
/************************************************************************/
long query_nextart(const char *group, long art_num)
{
    DIRBUFF *gdp;
    int i;
    int n = 0;

    if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        gdp = fetch_groupdir(group);
        for(i = 0; i < gdp->numarts; i++) {
            if(gdp->art_array[i] > art_num) {
                n = gdp->art_array[i];
                break;
            }
        }
        release_mutex(mutex);
    }
    return(n);
}


/************************************************************************/
/*                                                                      */
/* query_prevart                                                        */
/*                                                                      */
/* Requests the previous article number of a group.                     */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be queried.                                */
/*                                                                      */
/* art_num      The current article number. The routine will return     */
/*              the number of the article preceding this one. If this   */
/*              value is set to zero, then the number of the last       */
/*              article will be returned.                               */
/* ------------ ------------------------------------------------------- */
/* Return value The article number following 'art_num' or 0 in case of  */
/*              any failure.                                            */
/*                                                                      */
/************************************************************************/
long query_prevart(const char *group, long art_num)
{
    DIRBUFF *gdp;
    int i;
    int n = 0;

    if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        gdp = fetch_groupdir(group);
        for(i = gdp->numarts; --i >= 0; ) {
            /*
             * 07/05/97 harald
             * Bug removed, which keeps returning the current
             * article.
             * Changed 'array <= art_num' to 'array < art_num'.
             */
            if(art_num < 1 || gdp->art_array[i] < art_num) {
                n = gdp->art_array[i];
                break;
            }
        }
        release_mutex(mutex);
    }
    return(n);
}
