/*
 * This program was written by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

const char loadline_rcsid[] = "$Id: loadline.c,v 1.2 1997/02/13 17:01:59 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

#include <sysdep.h>

#include <loadline.h>

/************************************************************************/
/*                                                                      */
/* load_lines                                                           */
/*                                                                      */
/* A general purpose routine to load text files into an array of lines. */
/* The routine may optionaly remove comments as well as leading and     */
/* trailing spaces.                                                     */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* name         Points to a zero teminated string which specifies the   */
/*              name of the file to be loaded.                          */
/*                                                                      */
/* lines        Pointer which receives the address of an array of       */
/*              character pointers. The array will be allocated within  */
/*              this routine and must be released by the caller.        */
/*                                                                      */
/*              The first element will point to the buffer containing   */
/*              the complete file contents. This enables the caller     */
/*              to release the buffer using 'free()'. Note that this    */
/*              buffer cannot be used to restore the file because it    */
/*              has been modified by inserting string delimiters.       */
/*                                                                      */
/*              The last element of the array contains a NULL pointer.  */
/*                                                                      */
/*              All other elements point to a line being loaded.        */
/*                                                                      */
/* comment      Pointer to a zero terminated string containing all      */
/*              valid comment characters. This pointer may be NULL if   */
/*              no comments are expected in the file.                   */
/*                                                                      */
/* mode         Specifies the mode on how to interpret the file         */
/*              contents. All possible values are optional and can be   */
/*              given in any (or-ed) combination:                       */
/*                                                                      */
/*              Value               Meaning                             */
/*              ------------------- ----------------------------------- */
/*              LLMOD_SKIPLEADING   Leading spaces are skipped          */
/*              LLMOD_SKIPTRAILING  Trailing spaces are skipped         */
/*              LLMOD_COMMENTINLINE Comments are allowed at any         */
/*                                  position within a line.             */
/* ------------ ------------------------------------------------------- */
/* Return value Number of lines read or -1 in case of any error.        */
/*              Note that the array returned in 'lines' will contain    */
/*              two additional elements.                                */
/*                                                                      */
/************************************************************************/
int load_lines(char *name, char ***lines, char *comment, unsigned long mode)
{
    register char *lp;
    register char *cp;
    register char *cp1;
    char **result;
    int  i;
    char *buf;

    /*
     * Read in the file
     */
    if((buf = load_file(name)) == NULL)
        return(-1);

    /*
     * Get a gros count of the number of lines in the file. We add
     * three to the number of elements, one for the first element
     * containing the buffer address, one for the last element
     * containing a NULL pointer and one for the last line which
     * may not be terminated by a line feed.
     */
    for(cp = buf, i = 3; (cp = strchr(cp, '\n')) != NULL; cp++, i++)
        continue;

    /*
     * Allocate enough storage for a line pointer array
     * to contain all lines.
     */
    result = malloc((size_t)i * sizeof(char *));
    result[0] = buf;

    /*
     * Scan the file
     */
    for(i = 0, cp = buf; *cp; ) {
        for(lp = cp; *cp; cp++) {
            if(*cp == '\r' || *cp == '\n' || *cp == '\032')
                break;
        }
        *cp++ = '\0';
        if(*lp == '\0')
            continue;

        /*
         * Chop off comment.
         */
        if(comment) {
            if((mode & LLMOD_COMMENTINLINE) != 0) {
                for(cp1 = comment; *cp1; cp1++) {
                    if(strchr(lp, *cp1))
                        *cp1 = '\0';
                }
                if(*lp == '\0')
                    continue;
            }
            else if(strchr(comment, *lp) != NULL)
                continue;
        }

        /*
         * Cut trailing spaces.
         */
        if((mode & LLMOD_SKIPTRAILING) != 0) {
            for(cp1 = cp - 2; *lp; cp1--) {
                if(*cp1 == ' ' || *cp1 == '\t')
                    *cp1 = '\0';
                else
                    break;
            }
            if(*lp == '\0')
                continue;
        }

        /*
         * Skip leading spaces.
         */
        if((mode & LLMOD_SKIPLEADING) != 0) {
            while(*lp == ' ' || *lp == '\t')
                lp++;
            if(*lp == '\0')
                continue;
        }

        /*
         * Got a non-empty line.
         */
        result[++i] = lp;
    }
    result[i + 1] = NULL;   /* double stiched */
    *lines = result;

    return(i);
}

/************************************************************************/
/*                                                                      */
/* load_textfile                                                        */
/*                                                                      */
/* A useful routine to load text files into an array of lines. Lines    */
/* starting with character '#' will be recognized as comments and       */
/* removed. After converting all tabs to spaces, leading and trailing   */
/* spaces are removed and consecutive spaces are packed to a single     */
/* space.                                                               */
/*                                                                      */
/* The routine internally allocates all neccessary storage from heap,   */
/* which must be freed by the caller, if at least one valid line has    */
/* been found. Example:                                                 */
/*                                                                      */
/*  char **lines, **lp, *cp;                                            */
/*  int  n, i;                                                          */
/*                                                                      */
/*  if((n = load_textfile("file.txt", &lines)) > 0) {                   */
/*         ... process valid lines by index (backwards here) ...        */
/*      for(i = n; --i >= 0; )                                          */
/*          puts(lines[i]);                                             */
/*         ... or process by pointer ...                                */
/*      for(lp = lines; (cp = *lp++) != NULL; )                         */
/*          puts(cp);                                                   */
/*      free(*lines);                                                   */
/*      free(lines);                                                    */
/*  }                                                                   */
/*                                                                      */
/*                                                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* name         Points to a zero teminated string which specifies the   */
/*              name of the file to be loaded.                          */
/*                                                                      */
/* lines        Pointer which receives the address of an array of       */
/*              character pointers. The array will be allocated within  */
/*              this routine and must be released by the caller.        */
/*                                                                      */
/*              The first element will point to the buffer containing   */
/*              the complete file contents. This enables the caller     */
/*              to release the buffer using 'free()'. Note that this    */
/*              buffer cannot be used to restore the file because it    */
/*              has been modified.                                      */
/*                                                                      */
/*              The last element of the array contains a NULL pointer.  */
/*                                                                      */
/*              All other elements point to a line valid line.          */
/* ------------ ------------------------------------------------------- */
/* Return value Number of lines read or -1 in case of any error.        */
/*                                                                      */
/************************************************************************/
int load_textfile(char *name, char ***lines)
{
    char *buf;
    register char *rp;
    register char *wp;
    char nslast = 0;
    int  newline = 1;
    int  lcount = 0;
    char **lap;

    /*
     * Load the complete file into core.
     */
    if((buf = load_file(name)) == NULL) {
        return(-1);
    }

    for(rp = wp = buf; *rp; rp++) {
        /*
         * Skip comments and leading spaces.
         */
        if(newline) {
            if(*rp == '#') {
                while(*rp && *rp != '\n')
                    rp++;
            }
            while(*rp == ' ' || *rp == '\t')
                rp++;
        }

        /*
         * Skip carriage returns.
         */
        if(*rp == '\r')
            rp++;
        /*
         * Concat continuation lines and skip leading spaces.
         */
        else {
            while(*rp == '\\') {
                if(*++rp == '\r')
                    rp += 2;
                else if(*rp == '\n')
                    rp++;
                else {
                    if(*rp)
                        rp--;
                    break;
                }
                while(*rp == ' ' || *rp == '\t')
                    rp++;
            }
        }

        /*
         * This is the end of a line. Skip empty lines.
         */
        if(*rp == '\n' || *rp == '\032' || *rp == '\0') {
            if(!newline) {
                newline = 1;
                if(nslast)
                    *wp++ = '\n';
                else
                    *(wp - 1) = '\n';
                lcount++;
            }
        }

        /*
         * We're inside a line. Convert tabs to spaces
         * and pack spaces.
         */
        else {
            newline = 0;
            if(*rp == '\t')
                *rp = ' ';
            if(*rp == ' ') {
                if(nslast) {
                    nslast = 0;
                    *wp++ = *rp;
                }
            }
            else {
                nslast = 1;
                *wp++ = *rp;
            }
        }
    }
    *wp = '\0';

    /*
     * Allocate and set an array of line pointers.
     */
    if(lcount) {
        *lines = lap = malloc((size_t)(lcount + 1) * sizeof(char *));
        for(rp = buf, newline = 1; *rp; rp++) {
            if(newline) {
                *lap++ = rp;
                newline = 0;
            }
            if(*rp == '\n') {
                *rp = '\0';
                newline = 1;
            }
        }
        *lap = NULL;
    }
    else
        free(buf);

    return(lcount);
}

/*
 * $Log: loadline.c,v $
 * Revision 1.2  1997/02/13 17:01:59  oliver
 * removed ^M
 *
 * Revision 1.1  1997/02/13 11:09:20  oliver
 * initial revision
 *
 */
