/*
 * This program was written by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

const char loadtext_rcsid[] = "$Id: loadtext.c,v 1.1 1997/02/13 11:09:21 oliver Exp $";

#include <stdlib.h>
#include <string.h>

#include <loadfile.h>
#include <loadtext.h>

/************************************************************************/
/*                                                                      */
/* load_textfile                                                        */
/*                                                                      */
/* A useful routine to load text files into an array of lines. Lines    */
/* starting with character '#' will be recognized as comments and       */
/* removed. After converting all tabs to spaces, leading and trailing   */
/* spaces are removed and consecutive spaces are packed to a single     */
/* space.                                                               */
/*                                                                      */
/* The routine internally allocates all neccessary storage from heap,   */
/* which must be freed by the caller, if at least one valid line has    */
/* been found. Example:                                                 */
/*                                                                      */
/*  char **lines, **lp, *cp;                                            */
/*  int  n, i;                                                          */
/*                                                                      */
/*  if((n = load_textfile("file.txt", &lines)) > 0) {                   */
/*         ... process valid lines by index (backwards here) ...        */
/*      for(i = n; --i >= 0; )                                          */
/*          puts(lines[i]);                                             */
/*         ... or process by pointer ...                                */
/*      for(lp = lines; (cp = *lp++) != NULL; )                         */
/*          puts(cp);                                                   */
/*      free(*lines);                                                   */
/*      free(lines);                                                    */
/*  }                                                                   */
/*                                                                      */
/*                                                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* name         Points to a zero teminated string which specifies the   */
/*              name of the file to be loaded.                          */
/*                                                                      */
/* lines        Pointer which receives the address of an array of       */
/*              character pointers. The array will be allocated within  */
/*              this routine and must be released by the caller.        */
/*                                                                      */
/*              The first element will point to the buffer containing   */
/*              the complete file contents. This enables the caller     */
/*              to release the buffer using 'free()'. Note that this    */
/*              buffer cannot be used to restore the file because it    */
/*              has been modified.                                      */
/*                                                                      */
/*              The last element of the array contains a NULL pointer.  */
/*                                                                      */
/*              All other elements point to a line valid line.          */
/* ------------ ------------------------------------------------------- */
/* Return value Number of lines read or -1 in case of any error.        */
/*                                                                      */
/************************************************************************/
int load_textfile(const char *name, char ***lines)
{
    char *buf;
    register char *rp;
    register char *wp;
    char nslast = 0;
    int  newline = 1;
    int  lcount = 0;
    char **lap;

    /*
     * Load the complete file into core.
     */
    if((buf = load_file(name)) == NULL)
        return(-1);

    for(rp = wp = buf; *rp; rp++) {
        /*
         * Skip comments and leading spaces.
         */
        if(newline) {
            if(*rp == '#') {
                while(*rp && *rp != '\n')
                    rp++;
            }
            while(*rp == ' ' || *rp == '\t')
                rp++;
        }

        /*
         * Skip carriage returns.
         */
        if(*rp == '\r')
            rp++;
        /*
         * Concat continuation lines and skip leading spaces.
         */
        else {
            while(*rp == '\\') {
                if(*++rp == '\r')
                    rp += 2;
                else if(*rp == '\n')
                    rp++;
                else {
                    if(*rp)
                        rp--;
                    break;
                }
                while(*rp == ' ' || *rp == '\t')
                    rp++;
            }
        }

        /*
         * This is the end of a line. Skip empty lines.
         */
        if(*rp == '\n' || *rp == '\032' || *rp == '\0') {
            if(!newline) {
                newline = 1;
                if(nslast)
                    *wp++ = '\n';
                else
                    *(wp - 1) = '\n';
                lcount++;
            }
        }

        /*
         * We're inside a line. Convert tabs to spaces
         * and pack spaces.
         */
        else {
            newline = 0;
            if(*rp == '\t')
                *rp = ' ';
            if(*rp == ' ') {
                if(nslast) {
                    nslast = 0;
                    *wp++ = *rp;
                }
            }
            else {
                nslast = 1;
                *wp++ = *rp;
            }
        }
    }
    *wp = '\0';

    /*
     * Allocate and set an array of line pointers.
     */
    if(lcount) {
        *lines = lap = malloc((size_t)(lcount + 1) * sizeof(char *));
        for(rp = buf, newline = 1; *rp; rp++) {
            if(newline) {
                *lap++ = rp;
                newline = 0;
            }
            if(*rp == '\n') {
                *rp = '\0';
                newline = 1;
            }
        }
        *lap = NULL;
    }
    else
        free(buf);

    return(lcount);
}

/*
 * $Log: loadtext.c,v $
 * Revision 1.1  1997/02/13 11:09:21  oliver
 * initial revision
 *
 */
