#include <stdlib.h>
#include <string.h>

#include <mavltree.h>

/************************************************************************/
/*  t_insert                                                            */
/*                                                                      */
/*  Insert a new node into a subtree. Returns 1 if the tree has grown,  */
/*  2 if the key exists and zero otherwise.                             */
/************************************************************************/
static int t_insert(PNODE *ppTree, PNODE pNew)
{
    int   result;
    PNODE pNode, pLNode, pRNode;
    int   rel;

    /*
     * Insert a new node at this point
     */
    if((pNode = *ppTree) == NULL) {
        pNode = pNew;
        result = 1;
    }

    /*
     * New key already exists
     */
    else if((rel = strcmp((char *)(pNode + 1), (char *)(pNew + 1))) == 0)
        result = 2;

    /*
     * New key is smaller: Continue searching in left subtree.
     */
    else if(rel > 0) {
        if((result = t_insert(&pNode->pLLink, pNew)) == 1) {
            switch(pNode->status) {
            case LEFTGROWN:
                /*
                 * previously left grown tree needs to be rebalanced
                 */
                pLNode = pNode->pLLink;
                if(pLNode->status == LEFTGROWN) {
                    /*
                     * do a single LL rotation
                     */
                    pNode->pLLink = pLNode->pRLink;
                    pLNode->pRLink = pNode;
                    pNode->status = BALANCED;
                    pNode = pLNode;
                }
                else {
                    /*
                     * do a double LR rotation
                     */
                    pRNode = pLNode->pRLink;
                    pLNode->pRLink = pRNode->pLLink;
                    pRNode->pLLink = pLNode;
                    pNode->pLLink = pRNode->pRLink;
                    pRNode->pRLink = pNode;
                    pNode->status = (pRNode->status == LEFTGROWN) ? RIGHTGROWN : BALANCED;
                    pLNode->status = (pRNode->status == RIGHTGROWN) ? LEFTGROWN : BALANCED;
                    pNode = pRNode;
                }
                /* fall */
            case RIGHTGROWN:
                /*
                 * Previously right grown tree becomes balanced
                 */
                pNode->status = BALANCED;
                result = 0;
                break;
            case BALANCED:
                /*
                 * previously balanced tree becomes left grown
                 */
                pNode->status = LEFTGROWN;
                break;
            }
        }
    }

    /*
     * New key is greater: Continue searching in right subtree.
     */
    else {
        if((result = t_insert(&pNode->pRLink, pNew)) == 1) {
            switch(pNode->status) {
            case RIGHTGROWN:
                /*
                 * previously right grown tree needs to be rebalanced
                 */
                pLNode = pNode->pRLink;
                if(pLNode->status == RIGHTGROWN) {
                    /*
                     * do a single RR rotation
                     */
                    pNode->pRLink = pLNode->pLLink;
                    pLNode->pLLink = pNode;
                    pNode->status = BALANCED;
                    pNode = pLNode;
                }
                else {
                    /*
                     * do a double RL rotation
                     */
                    pRNode = pLNode->pLLink;
                    pLNode->pLLink = pRNode->pRLink;
                    pRNode->pRLink = pLNode;
                    pNode->pRLink = pRNode->pLLink;
                    pRNode->pLLink = pNode;
                    pNode->status = (pRNode->status == RIGHTGROWN) ? LEFTGROWN : BALANCED;
                    pLNode->status = (pRNode->status == LEFTGROWN) ? RIGHTGROWN : BALANCED;
                    pNode = pRNode;
                }
                /* fall */
            case LEFTGROWN:
                /*
                 * previously left grown tree becomes balanced
                 */
                pNode->status = BALANCED;
                result = 0;
                break;
            case BALANCED:
                /*
                 * previously balanced tree becomes right grown
                 */
                pNode->status = RIGHTGROWN;
                break;
            }
        }
    }

    /*
     * return the current root to the caller
     */
    *ppTree = pNode;

    return(result);
}


/************************************************************************/
/*                                                                      */
/*  avl_insert                                                          */
/*                                                                      */
/*  Inserts a new item in the tree. Returns 0 on success, -1 on errors. */
/*                                                                      */
/************************************************************************/
int avl_insert(NODE **ppRoot, char *id, long recnum)
{
    NODE *pNew;

    pNew = malloc(sizeof(NODE) + strlen(id) + 1);
    pNew->pLLink = NULL;
    pNew->pRLink = NULL;
    pNew->status = BALANCED;
    pNew->recnum = recnum;
    strcpy((char *)(pNew + 1), id);
    if(t_insert(ppRoot, pNew) == 2) {
        free(pNew);
        return(-1);
    }
    return(0);
}

/************************************************************************/
/*                                                                      */
/*  avl_find                                                            */
/*                                                                      */
/*  Walking down a tree searching for an item.                          */
/*                                                                      */
/************************************************************************/
long avl_find(NODE *pTree, char *id)
{
    int r;

    /*
     * Not found, return -1.
     */
    if(pTree == NULL)
        return(-1);

    /*
     * Item is below tree entry, turn left.
     */
    if((r = strcmp(id, (char *)(pTree + 1))) < 0)
        return(avl_find(pTree->pLLink, id));

    /*
     * Item is above tree entry, turn right.
     */
    else if(r)
        return(avl_find(pTree->pRLink, id));

    /*
     * Found, return record number.
     */
    return(pTree->recnum);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void avl_destroy(PNODE pTree)
{
    if(pTree) {
        avl_destroy(pTree->pLLink);
        avl_destroy(pTree->pRLink);
        free(pTree);
    }
}
