const char mhistory_rcsid[] = "$Id$";

/*
 * $Log$
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <syslog.h>
#include <xfile.h>

#include <chlimits.h>
#include <mutex.h>
#include <mavltree.h>


static FILE *hfp = NULL;
static NODE *root = NULL;

static long nh_mutex = -1;
static long mutex_wait = 32000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

static int debug_on = 0;

/************************************************************************/
/* debug_history                                                                     */
/*                                                                      */
/* Switch debugging on or off.                                          */
/*                                                                      */
/************************************************************************/
void debug_history(int enable)
{
    debug_on = enable;
}

/************************************************************************/
/*                                                                      */
/* history_maxwait                                                      */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void history_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int init_history(const char *name, int read_only)
{
    int  dupe_id = 0;
    int  bad_id = 0;
    int  bad_line = 0;
    int  n = 0;
    long recnum;
    char *cp;
    char *line;
    int  l = strlen(name);
    char *filename = malloc(l + 5);

    if(nh_mutex == -1 && (nh_mutex = create_mutex()) == -1)
        return(-1);

    /*
     * Caller doesn't supply the file extension. Add '.pag'
     * to the filename.
     */
    memcpy(filename, name, l);
    memcpy(filename + l, ".pag", 5);

    /*
     * Open the history file, either read/write or read-only.
     * Create a new one, if in read/write mode.
     */
    if((hfp = xfopen(filename, read_only ? "rb" : "r+b")) == NULL) {
        if(read_only) {
            syslog(LOG_ERR, "Open %s (r/o)", _strerror(filename));
            return(-1);
        }
        if((hfp = xfopen(filename, "w+b")) == NULL) {
            syslog(LOG_ERR, "Create %s (r/w)", _strerror(filename));
            return(-1);
        }
    }

    /*
     * Create the incore index.
     */
    recnum = 0;
    line = malloc(MAX_HISTORYLINE);
    while(fgets(line, MAX_HISTORYLINE, hfp)) {
        if(*line == '<') {
            if((cp = strchr(line, '>')) != NULL) {
                *++cp = '\0';
                if(avl_insert(&root, line, recnum))
                    dupe_id++;
                else
                    n++;
            }
            else
                bad_id++;
        }
        else
            bad_line++;
        recnum = ftell(hfp);
    }
    free(line);

    /*
     * Report problems occured during reading the history.
     */
    if(dupe_id || bad_id || bad_line)
        syslog(LOG_ERR, "history  dupes %d  bad ids %d  bad lines %d",
               dupe_id, bad_id, bad_line);
    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int exit_history(void)
{
    /*
     * Close the history file.
     */
    fclose(hfp);
    hfp = NULL;

    /*
     * Release memory resources used by the incore index.
     */
    avl_destroy(root);
    root = NULL;

    destroy_mutex(nh_mutex);
    nh_mutex = -1;

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int reopen_history(const char *name, int read_only)
{
    if(request_mutex(&nh_mutex, MUTEXWAIT) == 0) {
        fclose(hfp);
        hfp = NULL;
        avl_destroy(root);
        root = NULL;
        init_history(name, read_only);
        release_mutex(nh_mutex);
    }

    exit_history();
    return(init_history(name, read_only));
}


/************************************************************************/
/*                                                                      */
/* addhistentry                                                         */
/*                                                                      */
/* Adds a new line to the history database.                             */
/*                                                                      */
/* Format of our history line is:                                       */
/*                                                                      */
/*      id arrived~expires~posted size group:num[,group:num...]         */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on errors.                             */
/*                                                                      */
/************************************************************************/
int addhistentry(char *id, long t_arr, long t_exp, long t_post, long size, char *artlist)
{
    long recnum;

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return -1;

    if(fseek(hfp, 0, SEEK_END)) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s (Add)", _strerror("Seek to history end"));
        return(-1);
    }
    if((recnum = ftell(hfp)) == -1) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s", _strerror("Get history size"));
        return(-1);
    }

    if(avl_insert(&root, id, recnum)) {
        release_mutex(nh_mutex);
        return(-1);
    }

    if(t_exp)
        fprintf(hfp, "%s %ld~%ld", id, t_arr, t_exp);
    else
        fprintf(hfp, "%s %ld~-", id, t_arr);
    fprintf(hfp, "~%ld %ld %s\n", t_post, size, artlist);
    fflush(hfp);
    release_mutex(nh_mutex);

    return(0);
}

/************************************************************************/
/* gethistart                                                           */
/*                                                                      */
/* Searches the history database for a given id.                        */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* msg_id       Points to the message id string.                        */
/*                                                                      */
/* artpath      Must either point to a string buffer to receive the     */
/*              articles pathname or should be set to NULL if the       */
/*              caller is not interested in this information.           */
/*              In the first case the buffer should usually contain     */
/*              the news home directory excluding a trailing backslash, */
/*              and this routine will append the article's path.        */
/* ------------ ------------------------------------------------------- */
/* result       Returns 1 if the id exists or 0 otherwise.              */
/*                                                                      */
/************************************************************************/
int gethistart(char *msg_id, char *artpath)
{
    register char *cp, *cp1;
    int result = 0;
    long recnum;
    char *line;

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return(0);

    /*
     * Fetch the entry with the given message id
     */
    recnum = avl_find(root, msg_id);

    /*
     * Check for existence only.
     */
    if(artpath == NULL || recnum == -1) {
        release_mutex(nh_mutex);
        return(recnum != -1);
    }

    /*
     * Read the complete history line.
     */
    if(fseek(hfp, recnum, SEEK_SET)) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s (Get %ld)", _strerror("Seek history"), recnum);
        return(0);
    }
    line = malloc(MAX_HISTORYLINE);
    if(fgets(line, MAX_HISTORYLINE, hfp) == NULL) {
        release_mutex(nh_mutex);
        free(line);
        syslog(LOG_ERR, "%s", _strerror("Read history"));
        return(0);
    }
    release_mutex(nh_mutex);

    if((cp = strchr(line, '\n')) == NULL) {
        syslog(LOG_ERR, "History line too long: %.20s...", line);
        free(line);
        return(0);
    }
    *cp = '\0';

    /*
     * Parse the history line.
     */
    cp = line;
    if(*cp != '<') {
        syslog(LOG_ERR, "Bad history entry at %ld", recnum);
        free(line);
        return(-1);
    }

    /*
     * Skip id.
     */
    while(*cp && *cp != ' ' && *cp != '\t')
        cp++;
    while(*cp == ' ' || *cp == '\t')
        cp++;

    /*
     * Skip time fields.
     */
    while(*cp && *cp != ' ' && *cp != '\t')
        cp++;
    while(*cp == ' ' || *cp == '\t')
        cp++;

    /*
     * Skip size.
     */
    while(*cp && *cp != ' ' && *cp != '\t')
        cp++;
    while(*cp == ' ' || *cp == '\t')
        cp++;

    /*
     * Now cp points to the first <grp:num> item.
     * Append the path of this message to artpath.
     */
    if(*cp) {
        cp1 = artpath + strlen(artpath);
        *cp1++ = '\\';
        while(*cp && *cp != ' ' && *cp != '\t' && *cp != ',') {
            if(*cp == '.' || *cp == ':')
                *cp1++ = '\\';
            else
                *cp1++ = *cp;
            cp++;
        }
        *cp1 = '\0';
        result = 1;
    }

    free(line);
    return(result);
}

/************************************************************************/
/* pathlistbyid                                                         */
/*                                                                      */
/* Returns a space separated list of pathnames of article files with    */
/* the specified message-id. Used by XPATH.                             */
/*                                                                      */
/************************************************************************/
int pathlistbyid(char *msg_id, char *pathlist)
{
    register char *cr;
    register char *cw;
    long recnum;

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return(0);

    if((recnum = avl_find(root, msg_id)) == -1) {
        release_mutex(nh_mutex);
        return(0);
    }

    /*
     * Read the complete history line.
     */
    if(fseek(hfp, recnum, SEEK_SET)) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s (Pathlist %ld)", _strerror("Seek history"), recnum);
        return(0);
    }
    if(fgets(pathlist, MAX_HISTORYLINE, hfp) == NULL) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s", _strerror("Read history"));
        return(0);
    }

    release_mutex(nh_mutex);

    if((cr = strchr(pathlist, '\n')) == NULL) {
        syslog(LOG_ERR, "History line too long: %.20s...", pathlist);
        return(0);
    }
    *cr = '\0';
    cr = pathlist;

    while(*cr && *cr != ' ' && *cr != '\t')
        cr++;
    while(*cr == ' ' || *cr == '\t')
        cr++;
    while(*cr && *cr != ' ' && *cr != '\t')
        cr++;
    while(*cr == ' ' || *cr == '\t')
        cr++;

    /*
     * Now cr points to the <grp:num> list.
     */
    for(cw = pathlist; *cr; cr++) {
        if(*cr == '.' || *cr == ':')
            *cw++ = '/';
        else if(*cr == ',')
            *cw++ = ' ';
        else
            *cw++ = *cr;
    }
    *cw = '\0';

    return(*pathlist != '\0');
}

/************************************************************************/
/* cancelhistart                                                        */
/*                                                                      */
/* Cancels an article with a given id in the history base.              */
/*                                                                      */
/*     We assume this structure of our history line:                    */
/*                                                                      */
/*     <msg_id> <date>[ <size>] <grp:num>[,<grp:num>...]                */
/*                                                                      */
/*     Field separators may be spaces or tabs.                          */
/*                                                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* msg_id       Points to the message id string.                        */
/*                                                                      */
/* artlist      Must either point to a string buffer to receive the     */
/*              articles pathname or should be set to NULL if the       */
/*              caller is not interested in this information.           */
/*              In the first case the routine returns a comma separated */
/*              list of article paths relative to the news directory.   */
/* ------------ ------------------------------------------------------- */
/* result       Returns 0 if the article has been canceled, -1 otherwise*/
/*                                                                      */
/************************************************************************/
int cancelhistart(char *msg_id, char *artlist)
{
    register char *cp;
    long recnum;
    char *line;

    if(*msg_id != '<' || *(msg_id + strlen(msg_id) - 1) != '>')
        return(-1);

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return(-1);

    if((recnum = avl_find(root, msg_id)) == -1) {
        release_mutex(nh_mutex);
        return(-1);
    }

    /*
     * Read the complete history line.
     */
    if(fseek(hfp, recnum, SEEK_SET)) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s (Cancel %ld)", _strerror("Seek history"), recnum);
        return(-1);
    }
    line = malloc(MAX_HISTORYLINE);
    if(fgets(line, MAX_HISTORYLINE, hfp) == NULL) {
        release_mutex(nh_mutex);
        free(line);
        syslog(LOG_ERR, "%s", _strerror("Read history"));
        return(-1);
    }

    cp = line;
    if((cp = strchr(line, '\n')) == NULL) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "History line too long: %.20s...", line);
        free(line);
        return(-1);
    }
    *cp = '\0';


    /*
     * Skip message id, don't wipe it out.
     */
    cp = line;
    while(*cp && *cp != ' ' && *cp != '\t')
        cp++;

    /*
     * Skip the first field delimiter.
     */
    while(*cp == ' ' || *cp == '\t')
        cp++;

    /*
     * If we already reached the end of the line, then this
     * article has been canceled previously using the old
     * method.
     */
    if(*cp == '\0') {
        release_mutex(nh_mutex);
        if(debug_on)
            syslog(LOG_DEBUG, "Already canceled (old): %s", (char *)msg_id);
        return(-1);
    }

    /*
     * Skip the first date subfield, the arrival date.
     */
    while(*cp && *cp != ' ' && *cp != '\t' && *cp != '~')
        *cp++;

    /*
     * Wipe out the other date subfields, if they exist.
     */
    if(*cp == '~') {
        while(*cp && *cp != ' ' && *cp != '\t')
            *cp++ = ' ';
    }

    /*
     * Skip the second field delimiter.
     */
    while(*cp == ' ' || *cp == '\t')
        cp++;

    /*
     * If we reached the end of the line at this point, then
     * our article has been canceled previously using the
     * current method.
     */
    if(*cp == '\0') {
        release_mutex(nh_mutex);
        if(debug_on)
            syslog(LOG_DEBUG, "Already canceled (new): %s", (char *)msg_id);
        return(-1);
    }

    /*
     * If the next field starts with a digit, then our
     * line contains an article size field. Wipe it out
     * and skip the third field delimiter.
     */
    if(*cp >= '0' && *cp <= '9') {
        while(*cp && *cp != ' ' && *cp != '\t')
            *cp++ = ' ';
        while(*cp == ' ' || *cp == '\t')
            cp++;
    }

    /*
     * Return the article list and wipe it out.
     */
    if(artlist) {
        while(*cp && *cp != ' ' && *cp != '\t') {
            if(*cp == '.' || *cp == ':')
                *artlist++ = '\\';
            else
                *artlist++ = *cp;
            *cp++ = ' ';
        }
        *artlist = '\0';
    }

    /*
     * At least wipe out the rest of the line.
     */
    while(*cp)
        *cp++ = ' ';

    /*
     * Update the history file
     */
    if(fseek(hfp, recnum, SEEK_SET)) {
        release_mutex(nh_mutex);
        free(line);
        syslog(LOG_ERR, "%s (Cancel %ld)", _strerror("Seek history"), recnum);
        return(-1);
    }

    if(fputs(line, hfp) == EOF) {
        release_mutex(nh_mutex);
        free(line);
        syslog(LOG_ERR, "%s", _strerror("Write history"));
        return(-1);
    }
    fflush(hfp);
    release_mutex(nh_mutex);
    free(line);

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static unsigned long line_value(char *line)
{
    unsigned long result = 0;
    register char *cp = line;

    if(*cp == '<') {
        while(*cp && *cp != ' ' && *cp != '\t')
            cp++;
        result = atol(cp + 1);
    }
    else if(debug_on)
        syslog(LOG_DEBUG, "Invalid history line: %.20s...", line);

    return(result);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static void cut_trailing(char *line)
{
    register char *cp = line + (strlen(line) - 1);

    while(*cp == ' ' || *cp == '\n' || *cp == '\t' || *cp == '\r')
        cp--;

    *++cp = '\0';
}

/************************************************************************/
/* firsthist                                                            */
/*                                                                      */
/* Seeks through the history file looking for a line with date and time */
/* later than 'tkey'.                                                   */
/*                                                                      */
/* Returns the offset of the next line if successful, -1 otherwise.     */
/*                                                                      */
/************************************************************************/
long firsthist(unsigned long tkey, char *line)
{
    long top, bot = 0, mid;
    unsigned long trec;

    if(debug_on)
        syslog(LOG_DEBUG, "Binary search history for time >= %ld", tkey);

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return(-1);

    /*
     * Seek to the end of the file to get the top position.
     */
    if(fseek(hfp, 0L, SEEK_END)) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s", _strerror("Seek history end"));
        return(-1);
    }
    if((top = ftell(hfp)) == -1) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s", _strerror("Query history position"));
        return(-1);
    }

    /*
     * This is the big binary search loop.
     */
    for(;;) {

        /*
         * Set file pointer in the middle of our current
         * top and bottom.
         */
	mid = (top + bot) / 2;
        if(fseek(hfp, mid, SEEK_SET)) {
            release_mutex(nh_mutex);
            syslog(LOG_ERR, "%s (First %ld)", _strerror("Seek history"), mid);
            return(-1);
        }

        /*
         * Most probably we landed in the middle of a
         * line. Read the rest and discard it.
         */
        if(fgets(line, MAX_HISTORYLINE, hfp) == NULL) {
            release_mutex(nh_mutex);
            syslog(LOG_ERR, "%s", _strerror("Read history"));
            return(-1);
        }
        if((mid = ftell(hfp)) == -1) {
            release_mutex(nh_mutex);
            syslog(LOG_ERR, "%s", _strerror("Query history position"));
            return(-1);
        }

        /*
         * Read line by line until a valid one has been found
         */
        while(mid < top) {
            if(fgets(line, MAX_HISTORYLINE, hfp) == NULL) {
                release_mutex(nh_mutex);
                if(debug_on)
                    syslog(LOG_DEBUG, "%s", _strerror("Read history"));
                return(-1);
            }
            if((trec = line_value(line)) != 0)
                break;
            if((mid = ftell(hfp)) == -1) {
                release_mutex(nh_mutex);
                syslog(LOG_ERR, "%s", _strerror("Query history position"));
                return(-1);
            }
        }

        if(top <= mid)
            break;
        if(tkey <= trec)
            top = mid;
        else
            bot = mid;
    }

    if(fseek(hfp, bot, SEEK_SET)) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s (First %ld)", _strerror("Seek history"), bot);
        return(-1);
    }
    while(ftell(hfp) <= top) {
        if(fgets(line, MAX_HISTORYLINE, hfp) == NULL) {
            release_mutex(nh_mutex);
            if(debug_on)
                syslog(LOG_DEBUG, "%s", _strerror("Read history"));
            return(-1);
        }
        if(tkey <= line_value(line))
            break;
    }
    cut_trailing(line);

    mid = ftell(hfp);
    release_mutex(nh_mutex);

    return(mid);
}

/************************************************************************/
/*  nexthist                                                            */
/*                                                                      */
/*  Get the next line which begins at or after a given byte position    */
/*  of the history file. Lines longer than MAX_BUFSIZ will be cut,      */
/*  trailing spaces are removed.                                        */
/*                                                                      */
/*  Returns the offset of the next line if successful, -1 otherwise.    */
/*                                                                      */
/************************************************************************/
long nexthist(long offs, char *line)
{
    long result;

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return(-1);

    if(fseek(hfp, offs, SEEK_SET)) {
        release_mutex(nh_mutex);
        syslog(LOG_ERR, "%s (Next %ld)", _strerror("Seek history"), offs);
        return(-1);
    }
    while(fgets(line, MAX_HISTORYLINE, hfp)) {
        if(line_value(line) != 0) {
            cut_trailing(line);
            result = ftell(hfp);
            release_mutex(nh_mutex);
            return(result);
        }
    }

    release_mutex(nh_mutex);

    return(-1);
}
