const char nactive_rcsid[] = "$Id$";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <io.h>
#include <ctype.h>

#include <syslog.h>
#include <xfile.h>
#include <match.h>
#include <datetime.h>

#include <loadtext.h>
#include <mutex.h>

#include <dircache.h>
#include <nactive.h>

static long mutex_wait = 31000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

#define ACTIVE_TEMP "active.tmp"

typedef struct _ACTIVEINFO {
    long name_off;  /* offset into groupnames                       */
    long lomark;    /* lowest article number                        */
    long himark;    /* highest article number                       */
    long minkeep;   /* minimum number of seconds to keep            */
    long maxkeep;   /* maximum number of seconds to keep            */
    int  numarts;   /* number of articles in group                  */
    char mode;      /* "y"=can post, "n"=cannot post, "m"=moderated */
} ACTIVEINFO;

#define GRPNAMP(p)   (groupnames + (p)->name_off)
#define GRPNAMX(x)   (groupnames + ainfo[x].name_off)

static int n_groups = 0;
static int x_groups = 0;
static ACTIVEINFO *ainfo = NULL;

static int u_names = 0;
static int x_names = 0;
static char *groupnames = NULL;

static int is_dirty = 0;
static long mutex = -1;

static char newspath[_MAX_PATH];
static int debug_on = 0;

/*======================================================================*/
/* LOCAL ROUTINES                                                       */
/*======================================================================*/

/************************************************************************/
/*                                                                      */
/* cmpainfo                                                             */
/*                                                                      */
/* This local routine is called by the qsort library routine.           */
/*                                                                      */
/************************************************************************/
static int cmpainfo(const void *ptr1, const void *ptr2)
{
    return(strcmp(GRPNAMP((ACTIVEINFO *)ptr1), GRPNAMP((ACTIVEINFO *)ptr2)));
}

/************************************************************************/
/*                                                                      */
/* add_group                                                            */
/*                                                                      */
/* Local routine used to add a newsgroup to our active info array.      */
/*                                                                      */
/*                                                                      */
/************************************************************************/
static int add_group(char *nptr, size_t nsiz, long lo, long hi, char mode)
{
    ACTIVEINFO *aip;

    /*
     * Provide enough table space.
     */
    if(n_groups >= x_groups) {
        x_groups += 128;
        if((ainfo = realloc(ainfo, x_groups * sizeof(ACTIVEINFO))) == NULL)
            return(-1);
        if(debug_on)
            syslog(LOG_DEBUG, "Allocated space for %d active groups", x_groups);
    }
    aip = &ainfo[n_groups++];
    aip->name_off = u_names;
    aip->lomark = lo;
    aip->himark = hi;
    aip->numarts = -1;
    aip->mode = mode;

    /*
     * Get the name length and provide enough name space.
     */
    if((u_names += nsiz) >= x_names) {
        x_names += 8192;
        if((groupnames = realloc(groupnames, x_names)) == NULL)
            return(-1);
        if(debug_on)
            syslog(LOG_DEBUG, "Allocated %d characters for groupnames (1)", x_names);
    }
    memcpy(GRPNAMP(aip), nptr, nsiz);

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int add_activetimes(const char *group, const char *active_times, const char *who)
{
    int result = -1;
    FILE *fp;

    if((fp = xfopen(active_times, "a")) == NULL)
        syslog(LOG_ERR, "%s", _strerror((char *)active_times));
    else {
        fprintf(fp, "%s %ld %s\n", group, local_to_gmt(time(NULL)), who);
        if(fclose(fp))
            syslog(LOG_ERR, "%s", _strerror((char *)active_times));
        else
            result = 0;
    }
    return(result);
}

/************************************************************************/
/*                                                                      */
/* get_aip                                                              */
/*                                                                      */
/* Local routine to retrieve a pointer to an active info structure of   */
/* a given newsgroup name. After having done some tests with a hash     */
/* table, we're back with the binary search.                            */
/*                                                                      */
/* Note, that this routine may also return previously deleted groups.   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to search.                                    */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to the corresponding active info structure or   */
/*              NULL, if the group couldn't be found.                   */
/*                                                                      */
/************************************************************************/
static ACTIVEINFO *get_aip(const char *group)
{
    int cond;
    int l = 0, h, m;

    h = n_groups - 1;
    while(l <= h) {
        m = (l + h) / 2;
        if((cond = strcmp(group, GRPNAMX(m))) < 0)
            h = m - 1;
        else if(cond > 0)
            l = m + 1;
        else
            return(&(ainfo[m]));
    }
    return(NULL);
}

/************************************************************************/
/*                                                                      */
/* scan_active                                                          */
/*                                                                      */
/* Checks the newsgroup directory for the number of articles and the    */
/* lowest and highest message number and possibly updates the active    */
/* info structure and sets the dirty flag.                              */
/*                                                                      */
/************************************************************************/
static void scan_active(ACTIVEINFO *aip)
{
    int n;
    long lo, hi;

    if(debug_on)
        syslog(LOG_DEBUG, "Scanning %s: %ld-%ld", GRPNAMP(aip), aip->lomark, aip->himark);
    if((n = query_groupinfo(GRPNAMP(aip), &lo, &hi)) > 0) {
        if(debug_on)
            syslog(LOG_DEBUG, "Scanned %s: %d %ld-%ld", GRPNAMP(aip), n, lo, hi);
        if(lo != aip->lomark) {
            aip->lomark = lo;
            is_dirty = 1;
        }
        if(hi > aip->himark) {
            aip->himark = hi;
            is_dirty = 1;
        }
        aip->numarts = n;
    }
    else {
        if(debug_on)
            syslog(LOG_DEBUG, "No articles in %s", GRPNAMP(aip));
        if(aip->lomark <= aip->himark) {
            aip->lomark = aip->himark + 1;
            is_dirty = 1;
        }
        aip->numarts = 0;
    }
}

/*======================================================================*/
/* GLOBAL ROUTINES                                                      */
/*======================================================================*/

/************************************************************************/
/*                                                                      */
/* debug_active                                                         */
/*                                                                      */
/* Switches active file debugging on or off.                            */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* enable       A value of 0 will switch off debugging, otherwise it    */
/*              is switched on.                                         */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void debug_active(int enable)
{
    debug_on = enable;
}

/************************************************************************/
/*                                                                      */
/* active_maxwait                                                       */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void active_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/*                                                                      */
/* init_active                                                          */
/*                                                                      */
/* Initializes active file processing. This routine must be called      */
/* prior to most other routines of this module.                         */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* newsdir      Points to the string that specifies the path to the     */
/*              news home directory.                                    */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on failures                            */
/*                                                                      */
/************************************************************************/
int init_active(const char *newsdir)
{
    if(debug_on)
        syslog(LOG_DEBUG, "Init active for %s", newsdir);
    strcpy(newspath, newsdir);
    if((mutex = create_mutex()) == -1)
        return(-1);
    return(0);
}

/************************************************************************/
/*                                                                      */
/* exit_active                                                          */
/*                                                                      */
/* Releases any resources claimed by active file processing. This       */
/* routine should be called before program termination. After calling   */
/* no other procedure within this module may be called.                 */
/*                                                                      */
/************************************************************************/
void exit_active(void)
{
    if(debug_on)
        syslog(LOG_DEBUG, "Exit active");
    if(mutex != -1) {
        destroy_mutex(mutex);
        mutex = -1;
    }
}

/************************************************************************/
/*                                                                      */
/* new_active                                                           */
/*                                                                      */
/* Creates a new active file. Overwrites any existing one.              */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* pathname     Points to the string that specifies the pathname of     */
/*              the active file to be created.                          */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on failures                            */
/*                                                                      */
/************************************************************************/
int new_active(const char *pathname)
{
    int  result = 0;
    FILE *fp;

    syslog(LOG_INFO, "Creating new %s", pathname);

    if((fp = xfopen(pathname, "w")) != NULL) {
        fputs("control 0 1 n\n", fp);
        fputs("control.cancel 0 1 n\n", fp);
        fputs("control.newgroup 0 1 n\n", fp);
        fputs("control.rmgroup 0 1 n\n", fp);
        fputs("junk 0 1 n\n", fp);
        if(fclose(fp))
            result = -1;
    }
    else
        result = -1;

    if(result)
        syslog(LOG_ERR, "Can't create %s", _strerror((char *)pathname));
    else if(debug_on)
        syslog(LOG_DEBUG, "Created new %s", pathname);

    return(result);
}

/************************************************************************/
/*                                                                      */
/* load_active                                                          */
/*                                                                      */
/* Reads the active file and sets up our global array of active         */
/* newsgroups. This routine doesn't support mutual exclusion and should */
/* be called during the initialization phase.                           */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* pathname     Points to the string that specifies the pathname of     */
/*              the active file to load.                                */
/* ------------ ------------------------------------------------------- */
/* Return value -1 on failures, number of loaded groups otherwise.      */
/*              The caller should also be prepared to get a zero result */
/*              if the file doesn't contain any valid group definitions */
/*                                                                      */
/************************************************************************/
int load_active(const char *pathname)
{
    register char *lp;
    char **lines, **lpp;
    int  n_lines;
    long lo, hi;
    size_t nsiz;
    char *nptr;
    int i;

    /*
     * If the active file doesn't exist, let's create a virgin one.
     */
    if(access(pathname, 0)) {
        char *active_bak = malloc(strlen(pathname) + 5);
        strcpy(active_bak, pathname);
        strcat(active_bak, ".bak");
        if(access(active_bak, 0) == 0)
            xrename(active_bak, pathname);
        else if(new_active(pathname))
            return(-1);
        free(active_bak);
    }

    /*
     * Load all lines of the active file.
     */
    if((n_lines = load_textfile(pathname, &lines)) <= 0)
        return(n_lines);

    /*
     * Preallocate enough table and name space.
     */
    if(x_groups < n_lines) {
        x_groups = n_lines;
        if((ainfo = realloc(ainfo, sizeof(ACTIVEINFO) * x_groups)) == NULL)
            return(-1);
        if(debug_on)
            syslog(LOG_DEBUG, "Allocated space for %d active groups", x_groups);
    }
    if(x_names < n_lines * 64) {
        x_names = n_lines * 64;
        if((groupnames = realloc(groupnames, x_names)) == NULL)
            return(-1);
        if(debug_on)
            syslog(LOG_DEBUG, "Allocated %d characters for groupnames (2)", x_names);
    }
    /*
     * 08/20/97 harald
     * Corrected a bug. 'u_name' wasn't set to 0, so memory was
     * allocated each time when active was reloaded.
     */
    u_names = 0;

    /*
     * Fill the info array by parsing the lines of the file.
     */
    n_groups = 0;
    for(lpp = lines; (lp = *lpp++) != NULL; ) {
        nptr = strlwr(lp);

        /*
         * Get the group name.
         */
        for(nsiz = 1; *lp && *lp != ' '; ) {
            lp++;
            nsiz++;
        }
        if(*lp != ' ') {
            syslog(LOG_NOTICE, "Missing high message number for %s", nptr);
            continue;
        }
        *lp++ = '\0';

        /*
         * Get the highest message number.
         */
        if(*lp < '0' || *lp > '9') {
            syslog(LOG_NOTICE, "Bad high message number for %s", nptr);
            continue;
        }
        hi = atol(lp);

        /*
         * Get the lowest message number.
         */
        while(*lp && *lp != ' ')
            lp++;
        if(*lp++ != ' ' || *lp < '0' || *lp > '9') {
            syslog(LOG_NOTICE, "Missing low message number for %s", nptr);
            continue;
        }
        lo = atol(lp);

        /*
         * Get the mode character.
         */
        while(*lp && *lp != ' ')
            lp++;
        if(*lp != ' ') {
            syslog(LOG_NOTICE, "Missing last field for %s", nptr);
            continue;
        }
        switch(*++lp) {
        case 'y':
        case 'm':
        case 'n':
            if(add_group(nptr, nsiz, lo, hi, *lp)) {
                syslog(LOG_ERR, "Can't add %s", nptr);
                return(-1);
            }
            break;
        default:
            syslog(LOG_NOTICE, "Unknown group mode %s", lp);
            break;
        }
    }

    /*
     * Release line space.
     */
    free(*lines);
    free(lines);

    /*
     * Sort the array to allow binary search.
     */
    qsort(ainfo, n_groups, sizeof(ACTIVEINFO), cmpainfo);

    /*
     * Let's now walk through the list to detect duplicate entries.
     * Duplicate names will break all sorts of algorithms. To avoid
     * this, we store the active file with duplicates removed and
     * call us again.
     */
    for(i = 1; i < n_groups; i++) {
        if(ainfo[i - 1].mode && ainfo[i].mode &&
           strcmp(GRPNAMX(i - 1), GRPNAMX(i)) == 0) {
            syslog(LOG_NOTICE, "Duplicate group %s removed", GRPNAMX(i));
            ainfo[i - 1].mode = '\0';
            is_dirty = 1;
        }
    }
    if(is_dirty) {
        save_active(pathname, 0);
        n_groups = load_active(pathname);
    }

    return(n_groups);
}

/************************************************************************/
/*                                                                      */
/* save_active                                                          */
/*                                                                      */
/* Writes the active structure array back to the active file.           */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* pathname     Points to the string that specifies the pathname of     */
/*              the active file.                                        */
/*                                                                      */
/* force        If this value is set to 1, then the active file will    */
/*              be updated unconditionally. If this value is set to 0,  */
/*              the active file will only be updated, if anything       */
/*              has been changed since the last load or save.           */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on failures                            */
/*                                                                      */
/************************************************************************/
int save_active(const char *pathname, int force)
{
    int  result = -1;
    FILE *fp;
    int i;
    ACTIVEINFO *aip;
    char *active_tmp;

    if(!is_dirty && !force) {
        result = 0;
        if(debug_on)
            syslog(LOG_DEBUG, "%s is uptodate", pathname);
    }
    else if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        active_tmp = malloc(strlen(pathname) + 5);
        strcpy(active_tmp, pathname);
        strcat(active_tmp, ".tmp");
        if((fp = xfopen(active_tmp, "w")) != NULL) {
            if(debug_on)
                syslog(LOG_DEBUG, "Writing %d groups to %s", n_groups, active_tmp);
            for(aip = ainfo, i = n_groups; --i >= 0; aip++) {
                if(aip->mode)
                    if(fprintf(fp, "%s %ld %ld %c\n", GRPNAMP(aip),
                               aip->himark, aip->lomark, aip->mode) < 0) {
                        syslog(LOG_ERR, "%s", _strerror(active_tmp));
                        break;
                    }
                aip->numarts = -1;
            }
            for(; --i >= 0; aip++)
                aip->numarts = -1;
            if(fclose(fp) == 0) {
                if(debug_on)
                    syslog(LOG_DEBUG, "Saved %s %s", active_tmp,
                           is_dirty ? "" : "(forced)");
                result = is_dirty = 0;
            }
            else
                syslog(LOG_ERR, "%s", _strerror(active_tmp));
        }
        else
            syslog(LOG_ERR, "%s", _strerror(active_tmp));

        if(result == 0) {
            if(xrename(pathname, "*.bak"))
                syslog(LOG_ERR, "%s", _strerror((char *)pathname));
            if(xrename(active_tmp, pathname))
                syslog(LOG_ERR, "%s", _strerror((char *)pathname));
        }

        release_mutex(mutex);
    }
    return(result);
}

/************************************************************************/
/*                                                                      */
/* test_active                                                          */
/*                                                                      */
/* Tests if a group is in the active file and returns the groups's      */
/* mode character.                                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be queried.                                */
/* ------------ ------------------------------------------------------- */
/* Return value Newsgroup's mode character or zero, if the newsgroup    */
/*              hasn't been found.                                      */
/*                                                                      */
/************************************************************************/
char test_active(const char *group)
{
    ACTIVEINFO *aip;
    char result = '\0';

    if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        if((aip = get_aip(group)) != NULL)
            result = aip->mode;
        release_mutex(mutex);
    }
    return(result);
}

/************************************************************************/
/*                                                                      */
/* query_active                                                         */
/*                                                                      */
/* Requests the number of articles and the lowest and highest article   */
/* number of a given newsgroup. This routine may scan the newsgroup     */
/* directory, if not done so since loading the active file.             */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be queried.                                */
/*                                                                      */
/* num          Points to an integer, where this routine will store     */
/*              the number of articles found in the specified group.    */
/*              A NULL pointer may be supplied if the caller is not     */
/*              interested in this value.                               */
/*                                                                      */
/* lo           Pointer to a long, where this routine will store        */
/*              the lowest article number of the specified group.       */
/*              A NULL pointer may be supplied if the caller is not     */
/*              interested in this value.                               */
/*                                                                      */
/* hi           Pointer to a long, where this routine will store        */
/*              the highest article number of the specified group.      */
/*              A NULL pointer may be supplied if the caller is not     */
/*              interested in this value.                               */
/* ------------ ------------------------------------------------------- */
/* Return value Newsgroup's mode character or zero, if the newsgroup    */
/*              couldn't be found.                                      */
/*                                                                      */
/************************************************************************/
char query_active(const char *group, int *num, long *lo, long *hi)
{
    ACTIVEINFO *aip;

    /*
     * Try to get exclusive access.
     */
    if(request_mutex(&mutex, MUTEXWAIT))
        return('\0');

    /*
     * Find the group in our info array.
     */
    if((aip = get_aip(group)) == NULL || aip->mode == '\0') {
        syslog(LOG_INFO, "Can't find newsgroup %s", group);
        release_mutex(mutex);
        return('\0');
    }

    /*
     * See if we need to scan the newsgroup directory. The 'scan_active'
     * routine will update the active info structure and set the dirty flag.
     */
    if(aip->numarts == -1)
        scan_active(aip);

    /*
     * Return all requested values.
     */
    if(num)
        *num = aip->numarts;

    if(aip->numarts) {
        if(lo)
            *lo = aip->lomark;
        if(hi)
            *hi = aip->himark;
    }
    else {
        if(lo)
            *lo = 0;
        if(hi)
            *hi = 0;
    }
    release_mutex(mutex);

    return(aip->mode);
}

/************************************************************************/
/*                                                                      */
/* valid_newsgroup                                                      */
/*                                                                      */
/* Check a newsgroup name for validity.                                 */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be chaecked.                               */
/* ------------ ------------------------------------------------------- */
/* Return value 1, if the newsgroup name is valid, 0 otherwise.         */
/*                                                                      */
/************************************************************************/
int valid_newsgroup(const char *group)
{
    const char *cp;

    if(*group == '.' || strspn(group, "0123456789") == strlen(group))
        return(0);
    for(cp = group; *cp; cp++) {
        if(*cp == '.') {
            if(*(cp + 1) == '.' || *(cp + 1) == '\0')
                return(0);
        }
        else if(isspace(*cp) || strchr(":!\\/", *cp) != NULL)
            return(0);
    }
    return(1);
}

/************************************************************************/
/*                                                                      */
/* add_active                                                           */
/*                                                                      */
/* Adds a new group to the list of active newsgroups.                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be updated. This string is converted to    */
/*              lowercase.                                              */
/*                                                                      */
/* mode         Mode character of a new group or new mode character     */
/*              of an existing group.                                   */
/*                                                                      */
/* who          Points to the string that specifies the name of the     */
/*              creator of the new group. If this pointer is NULL       */
/*              or points to an empty string, then the default          */
/*              creator named 'usenet' will be used.                    */
/*                                                                      */
/* active_times Points to a string that specifies the name of the       */
/*              active times file. If this pointer is not NULL and      */
/*              doesn't point to an empty string, then this routine     */
/*              will append a line to the active times file, if the     */
/*              new group has been added successfully.                  */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success                                            */
/*              -1 on illegal newsgroup name                            */
/*              1 if an existing group has been updated                 */
/*              2 if nothing has been changed                           */
/*                                                                      */
/************************************************************************/
int add_active(char *group, long lo, long hi, char mode, char *who, const char *active_times)
{
    int result = -1;
    ACTIVEINFO *aip;

    /*
     * Force group name and mode to lower case
     */
    strlwr(group);
    mode = tolower(mode);

    /*
     * Validate newsgroup name
     */
    if(valid_newsgroup(group) == 0)
        return(-1);

    /*
     * If not explicitely given, use default who
     */
    if(who == NULL || who == '\0')
        who = "usenet";

    if(request_mutex(&mutex, MUTEXWAIT))
        return(-1);

    /*
     * If this group exists then we do a simple change. Note that
     * this includes previously deleted groups.
     */
    if((aip = get_aip(group)) != NULL) {
        if(aip->mode != mode) {
            if(aip->mode == '\0') {
                if(active_times && *active_times)
                    add_activetimes(group, active_times, who);
                syslog(LOG_INFO, "%s added %s with status %c", who, group, mode);
                result = 0;
            }
            else {
                syslog(LOG_INFO, "%s changed %s to %c", who, group, mode);
                result = 1;
            }
            aip->mode = mode;
            is_dirty = 1;
        }
        else
            result = 2;
    }

    /*
     * If this group doesn't exists then append an entry
     * to active.times and add the group to our list of
     * active groups.
     */
    else if(add_group(group, strlen(group) + 1, lo, hi, mode) == 0) {
        qsort(ainfo, n_groups, sizeof(ACTIVEINFO), cmpainfo);
        if(active_times && *active_times) {
            add_activetimes(group, active_times, who);
            syslog(LOG_INFO, "%s added %s with status %c", who, group, mode);
        }
        else if(debug_on)
            syslog(LOG_INFO, "%s added %s with status %c", who, group, mode);
        is_dirty = 1;
        result = 0;
    }

    release_mutex(mutex);
    return(result);
}

/************************************************************************/
/*                                                                      */
/* del_active                                                           */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be removed.                                */
/*                                                                      */
/* who          Points to the string that specifies the name of the     */
/*              destroyer of the new group. If this pointer is NULL     */
/*              or points to an empty string, then the default name     */
/*              'usenet' will be used.                                  */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on failures                            */
/*                                                                      */
/************************************************************************/
int del_active(const char *group, char *who)
{
    int result = -1;
    ACTIVEINFO *aip;

    if(who == NULL || *who == '\0')
        who = "usenet";

    if(request_mutex(&mutex, MUTEXWAIT))
        return(-1);

    if((aip = get_aip(group)) != NULL && aip->mode != '\0') {
        aip->mode = '\0';
        syslog(LOG_INFO, "%s removed newsgroup %s", who, group);
        is_dirty = 1;
        result = 0;
    }
    release_mutex(mutex);

    return(result);
}

/************************************************************************/
/*                                                                      */
/* inc_active                                                           */
/*                                                                      */
/* This routine is used to increment the highest message number of a    */
/* group.                                                               */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to the string that specifies the name of the     */
/*              newsgroup to be updated.                                */
/* ------------ ------------------------------------------------------- */
/* Return value new article number on success, 0 otherwise.             */
/*                                                                      */
/************************************************************************/
long inc_active(const char *group)
{
    ACTIVEINFO *aip;
    long result = 0;

    if(request_mutex(&mutex, MUTEXWAIT))
        return(result);

    if((aip = get_aip(group)) != NULL && aip->mode != '\0') {
        /*
         * If not already done, scan the newsgroup directory and
         * update the active info structure.
         */
        if(aip->numarts == -1)
            scan_active(aip);
        result = ++(aip->himark);
        is_dirty = 1;
    }
    release_mutex(mutex);

    return(result);
}

/************************************************************************/
/*                                                                      */
/* next_group                                                           */
/*                                                                      */
/* This routine is used to step through all active newsgroup names      */
/* in alphabetical order.                                               */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to a buffer for the group name. On entry the     */
/*              caller must place the previous newsgroup name in this   */
/*              buffer. The routine will put the name of the next       */
/*              group in it. In order to request the first group the    */
/*              buffer should contain an empty string.                  */
/*              Note, that the buffer must be large enough to contain   */
/*              the largest possible group name.                        */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to 'group' on success, NULL otherwise.          */
/*                                                                      */
/************************************************************************/
char *next_group(char *group)
{
    int cond;
    int l = 0, h, m;

    if(request_mutex(&mutex, MUTEXWAIT))
        return(NULL);

    /*
     * If the name of the previous group is an empty string, then
     * return the name of the first one.
     */
    if(*group == '\0') {
        for(m = 0; m < n_groups; m++) {
            if(ainfo[m].mode != '\0') {
                strcpy(group, GRPNAMX(m));
                release_mutex(mutex);
                return(group);
            }
        }
    }

    /*
     * Find the specified group in our list and return the name
     * of the next valid newsgroup.
     */
    else {
        h = n_groups - 1;
        while(l <= h) {
            m = (l + h) / 2;
            if((cond = strcmp(group, GRPNAMX(m))) < 0)
                h = m - 1;
            else if(cond > 0)
                l = m + 1;
            else {
                /*
                 * Loop until we found the next valid one.
                 */
                while(++m < n_groups) {
                    if(ainfo[m].mode != '\0') {
                        strcpy(group, GRPNAMX(m));
                        release_mutex(mutex);
                        return(group);
                    }
                }
                break;
            }
        }
    }
    release_mutex(mutex);
    return(NULL);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int get_groupentry(int index, char *info)
{
    int result = -1;

    if(request_mutex(&mutex, MUTEXWAIT))
        return(-1);

    if(index < n_groups && index >= 0) {
        if(ainfo[index].mode != '\0') {
            char *cp = info + strlen(info);
            strcpy(cp, GRPNAMX(index));
            cp += strlen(cp);
            *cp++ = ' ';
            *cp++ = ainfo[index].mode;
            *cp = '\0';
            result = 0;
        }
        else
            result = 1;
    }
    release_mutex(mutex);

    return(result);
}

/************************************************************************/
/*                                                                      */
/* set_expiration                                                       */
/*                                                                      */
/* This routine is used to set the minimum and maximum expiration time  */
/* of groups specified by a wildmatch list.                             */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* ------------ ------------------------------------------------------- */
/* Return value none                                                    */
/*                                                                      */
/************************************************************************/
void set_expiration(char *pattern, long minkeep, long maxkeep)
{
    int i;
    ACTIVEINFO *aip;

    if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        for(aip = ainfo, i = n_groups; --i >= 0; aip++) {
            if(ngmatch(GRPNAMP(aip), pattern)) {
                aip->minkeep = minkeep;
                aip->maxkeep = maxkeep;
            }
        }
        release_mutex(mutex);
    }
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int query_expiration(char *group, long *minkeep, long *maxkeep)
{
    int result = -1;
    ACTIVEINFO *aip;

    if(request_mutex(&mutex, MUTEXWAIT) == 0) {
        if((aip = get_aip(group)) != NULL) {
            if(aip->mode != '\0') {
                *minkeep = aip->minkeep;
                *maxkeep = aip->maxkeep;
                result = 0;
            }
        }
        release_mutex(mutex);
    }
    return(result);
}

/*
 * $Log: nactive.c,v $
 * Revision 1.1  1997/02/13 11:09:24  oliver
 * initial revision
 *
 */
