/*
 *  simple history database manager for UUPC news
 *
 *  Author:  Kai Uwe Rommel <rommel@ars.muc.de>
 *  Created: Sun Aug 15 1993
 *
 *  Bug reports related to THIS modified version should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

const char ndbm_rcsid[] = "$Id: ndbm.c,v 1.1 1997/02/13 11:09:24 oliver Exp $";

/*--------------------------------------------------------------------*/
/*       Parts are Copyright (c) 1989-1994 by Kendra Electronic       */
/*       Wonderworks.                                                 */
/*                                                                    */
/*       All rights reserved except those explicitly granted by       */
/*       the UUPC/extended license agreement.                         */
/*--------------------------------------------------------------------*/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

#include <syslog.h>
#include <chlimits.h>
#include <xfile.h>
#include <ndbm.h>

static debug_on = 0;

/************************************************************************/
/* Switch debugging on or off.                                          */
/*                                                                      */
/* 07/15/1996 harald                                                    */
/************************************************************************/
void ndbm_debug(int enable)
{
    nidx_debug(debug_on = enable);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int ndbm_flush(NDBM *db)
{
    if(db == NULL || db->dbfile == NULL) {
        if(debug_on)
            syslog(LOG_DEBUG, "Flush: Illegal ndbm handle");
        return(-1);
    }
    if(fflush(db->dbfile) == EOF) {
        syslog(LOG_ERR, "%s", _strerror("ndbm flush"));
        return(-1);
    }
    else if(debug_on)
        syslog(LOG_ERR, "ndbm flushed");
    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
NDBM *ndbm_open(const char *name, int read_only)
{
    NDBM *db = memset(malloc(sizeof(NDBM)), 0, sizeof(NDBM));
    char *filename = malloc(_MAX_PATH);
    char *ext;

    db->bufrec = -1;
    strcpy(filename, name);
    strcpy(ext = filename + strlen(filename), NDBM_EXT_DBF);

    if(read_only) {
        if((db->dbfile = xfopen(filename, "rb")) == NULL)
            syslog(LOG_ERR, "Open %s(r/o)", _strerror(filename));
        else if(debug_on)
            syslog(LOG_DEBUG, "Opened %s read-only", filename);
    }
    else if((db->dbfile = xfopen(filename, "r+b")) == NULL) {
        if(debug_on)
            syslog(LOG_DEBUG, "Open %s(r/w)", _strerror(filename));
        if((db->dbfile = xfopen(filename, "w+b")) == NULL)
            syslog(LOG_ERR, "Create %s(r/w)", _strerror(filename));
        else if(debug_on)
            syslog(LOG_DEBUG, "Created %s read/write", filename);
    }
    else if(debug_on)
        syslog(LOG_DEBUG, "Opened %s read/write", filename);

    if(db->dbfile) {
        strcpy(ext, NDBM_EXT_IDX);
        if(nidx_init(db, filename, read_only) == 0) {
            free(filename);
            return(db);
        }
        fclose(db->dbfile);
    }
    free(db);
    free(filename);

    return(NULL);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void ndbm_close(NDBM * db)
{
    if(db == NULL) {
        if(debug_on)
            syslog(LOG_DEBUG, "ndbm close: null handle");
        return;
    }

    nidx_exit(db);

    if(db->dbfile == NULL) {
        if(debug_on)
            syslog(LOG_DEBUG, "Trying to close unopened ndbm");
    }
    else if(fclose(db->dbfile))
        syslog(LOG_ERR, "%s", _strerror("ndbm close"));
    else if(debug_on)
        syslog(LOG_DEBUG, "ndbm closed");
    db->dbfile = NULL;

    free(db);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int ndbm_store(NDBM *db, char *val)
{
    int result = -1;
    char *cp;
    char c = '\0';
    long offset;

    if(db == NULL || db->dbfile == NULL) {
        syslog(LOG_ERR, "Store ndbm: Illegal handle");
        return(-1);
    }

    /*
     * Allow spaces and tabs, even lines with message-id only.
     * 07/15/1996 harald
     */
    for(cp = val; *cp; cp++)
        if(*cp == ' ' || *cp == '\t') {
            c = *cp;
            *cp = '\0';
            break;
        }

    if(fseek(db->dbfile, 0, SEEK_END)) {
        syslog(LOG_ERR, "%s", _strerror("Seek to ndbm end"));
        return(-1);
    }
    if((offset = ftell(db->dbfile)) == -1) {
        syslog(LOG_ERR, "%s", _strerror("Get dbm size"));
        return(-1);
    }

    if(nidx_addkey(db, val, offset))
        *cp = c;
    else {
        *cp = c;
        if(fseek(db->dbfile, 0, SEEK_END))
            syslog(LOG_ERR, "%s", _strerror("Seek to ndbm end"));
        else {
            if(fputs(val, db->dbfile) == EOF && debug_on)
                syslog(LOG_DEBUG, "%s", _strerror("ndbm write"));
            fputc('\n', db->dbfile);
            result = 0;
        }
    }
    return(result);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int ndbm_fetch(NDBM *db, const char *key, char *val)
{
    char *cp;
    long offset;

    if(db == NULL) {
        syslog(LOG_ERR, "Store ndbm: NULL handle");
        return(-1);
    }

    if((offset = nidx_getkey(db, key)) == -1)
        return(-1);

    /*
     * Caller is not interested in the record itself.
     */
    if(val == NULL)
        return(0);

    /*
     * This is the advantage point of our hash coded key.
     * If the index routine found our key, it read the
     * record already into 'bufrec', which is fine, because
     * now we would do it anyway. The following 'if' statement
     * isn't really needed, but we keep it here for approval.
     */
    if(offset != db->bufrec) {
        /*
         * This syslog message should never appear.
         */
        syslog(LOG_INFO, "Re-reading record %ld", offset);
        if(fseek(db->dbfile, offset, SEEK_SET)) {
            syslog(LOG_ERR, "%s", _strerror("Seek ndbm"));
            return(-1);
        }

        if(fgets(db->buffer, sizeof(db->buffer), db->dbfile) == NULL) {
            syslog(LOG_ERR, "%s", _strerror("Read ndbm"));
            return(-1);
        }
        db->bufrec = offset;
        if((cp = strchr(db->buffer, '\n')) != NULL)
            *cp = '\0';
    }

    /*
     * Allow spaces as well as tabs.
     * 07/15/1996 harald
     */
    *val = '\0';
    for(cp = db->buffer; *cp; cp++)
        if(*cp == ' ' || *cp == '\t') {
            strcpy(val, cp + 1);
            break;
        }

    return(0);
}

/************************************************************************/
/*                                                                      */
/*                                                                      */
/* This routine is now very liberal on the contents of a history line.  */
/* The date field may or may not contain subfields and missing size     */
/* fields are silently ignored too. Either TAB or space is accepted     */
/* as field delimiter at any place.                                     */
/* 07/15/1996 harald                                                    */
/************************************************************************/
int __cdecl ndbm_cancel(NDBM *db, const char *key, char *artlist)
{
    register char *cp;
    long offset;

    if(db == NULL) {
        syslog(LOG_ERR, "Cancel ndbm: NULL handle");
        return(-1);
    }

    if((offset = nidx_getkey(db, key)) == -1)
        return(-1);

    /*
     * See ndbm_fetch for further explanation and why the
     * following syslog message should never appear.
     */
    if(offset != db->bufrec) {
        syslog(LOG_INFO, "Re-reading record %ld", offset);
        if(fseek(db->dbfile, offset, SEEK_SET)) {
            syslog(LOG_ERR, "%s", _strerror("Seek dbm"));
            return(-1);
        }

        if(fgets(db->buffer, sizeof(db->buffer), db->dbfile) == NULL) {
            syslog(LOG_ERR, "%s", _strerror("Read ndbm"));
            return(-1);
        }
        db->bufrec = offset;
        if((cp = strchr(db->buffer, '\n')) != NULL)
            *cp = '\0';
    }

    /*
     * We assume this structure of our history line:
     *
     * <msg_id> <date>[ <size>] <grp:num>[,<grp:num>...]
     *
     * Field separators may be spaces or tabs.
     */
    cp = db->buffer;

    /*
     * Skip message id, don't wipe it out.
     */
    while(*cp && *cp != ' ' && *cp != '\t')
        cp++;

    /*
     * Skip the first field delimiter.
     */
    while(*cp == ' ' || *cp == '\t')
        cp++;

    /*
     * If we already reached the end of the line, then this
     * article has been canceled previously using the old
     * method.
     * 07/15/1996 harald
     */
    if(*cp == '\0') {
        if(debug_on)
            syslog(LOG_DEBUG, "Already canceled (old): %s", (char *)key);
        return(-1);
    }

    /*
     * Skip the first date subfield, the arrival date.
     */
    while(*cp && *cp != ' ' && *cp != '\t' && *cp != '~')
        *cp++;

    /*
     * Wipe out the other date subfields, if they exist.
     */
    if(*cp == '~') {
        while(*cp && *cp != ' ' && *cp != '\t')
            *cp++ = ' ';
    }

    /*
     * Skip the second field delimiter.
     */
    while(*cp == ' ' || *cp == '\t')
        cp++;

    /*
     * If we reached the end of the line at this point, then
     * our article has been canceled previously using the
     * current method.
     * 07/15/1996 harald
     */
    if(*cp == '\0') {
        if(debug_on)
            syslog(LOG_DEBUG, "Already canceled (new): %s", (char *)key);
        return(-1);
    }

    /*
     * If the next field starts with a digit, then our
     * line contains an article size field. Wipe it out
     * and skip the third field delimiter.
     */
    if(*cp >= '0' && *cp <= '9') {
        while(*cp && *cp != ' ' && *cp != '\t')
            *cp++ = ' ';
        while(*cp == ' ' || *cp == '\t')
            cp++;
    }

    /*
     * Return the article list and wipe it out.
     */
    if(artlist) {
        while(*cp && *cp != ' ' && *cp != '\t') {
            if(*cp == '.' || *cp == ':')
                *artlist++ = '\\';
            else
                *artlist++ = *cp;
            *cp++ = ' ';
        }
        *artlist = '\0';
    }

    /*
     * At least wipe out the rest of the line.
     */
    while(*cp)
        *cp++ = ' ';

    /*
     * Update the history file
     */
    if(fseek(db->dbfile, offset, SEEK_SET)) {
        syslog(LOG_ERR, "%s", _strerror("Seek ndbm"));
        return(-1);
    }

    if(fputs(db->buffer, db->dbfile) == EOF) {
        if(debug_on)
            syslog(LOG_DEBUG, "%s", _strerror("Write ndbm"));
        return(-1);
    }

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static unsigned long line_value(char *line)
{
    unsigned long result = 0;
    register char *cp = line;

    if(*cp == '<') {
        while(*cp && *cp != ' ' && *cp != '\t')
            cp++;
        result = atol(cp + 1);
    }
    else if(debug_on)
        syslog(LOG_DEBUG, "Invalid ndbm line: %.20s...", line);

    return(result);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static void cut_trailing(char *line)
{
    register char *cp = line + (strlen(line) - 1);

    while(*cp == ' ' || *cp == '\n' || *cp == '\t' || *cp == '\r')
        cp--;

    *++cp = '\0';
}

/************************************************************************/
/*                                                                      */
/* This routine performs a binary search on the history file for a      */
/* specific arrival time.                                               */
/*                                                                      */
/************************************************************************/
long ndbm_first(NDBM *db, unsigned long tkey, char *line)
{
    long top, bot = 0, mid;
    unsigned long trec;

    if(db == NULL) {
        syslog(LOG_ERR, "First ndbm: NULL handle");
        return(-1);
    }

    /*
     * Seek to the end of the file to get the top position.
     */
    if(fseek(db->dbfile, 0L, SEEK_END)) {
        syslog(LOG_ERR, "%s", _strerror("Seek ndbm end"));
        return(-1);
    }
    if((top = ftell(db->dbfile)) == -1) {
        syslog(LOG_ERR, "%s", _strerror("Query ndbm position"));
        return(-1);
    }

    /*
     * This is the big binary search loop.
     */
    for(;;) {

        /*
         * Set file pointer in the middle of our current
         * top and bottom.
         */
	mid = (top + bot) / 2;
        if(fseek(db->dbfile, mid, SEEK_SET)) {
            syslog(LOG_ERR, "%s", _strerror("Seek ndbm"));
            return(-1);
        }

        /*
         * Most probably we landed in the middle of a
         * line. Read the rest and discard it.
         */
        if(fgets(line, MAX_HISTORYLINE, db->dbfile) == NULL) {
            syslog(LOG_ERR, "%s", _strerror("Read ndbm"));
            return(-1);
        }
        if((mid = ftell(db->dbfile)) == -1) {
            syslog(LOG_ERR, "%s", _strerror("Query ndbm position"));
            return(-1);
        }

        /*
         * Read line by line until a valid one has been found
         */
        while(mid < top) {
            if(fgets(line, MAX_HISTORYLINE, db->dbfile) == NULL) {
                if(debug_on)
                    syslog(LOG_DEBUG, "%s", _strerror("Read ndbm"));
                return(-1);
            }
            if((trec = line_value(line)) != 0)
                break;
            if((mid = ftell(db->dbfile)) == -1) {
                syslog(LOG_ERR, "%s", _strerror("Query ndbm position"));
                return(-1);
            }
        }

        if(top <= mid)
            break;
        if(tkey <= trec)
            top = mid;
        else
            bot = mid;
    }

    if(fseek(db->dbfile, bot, SEEK_SET)) {
        syslog(LOG_ERR, "%s", _strerror("Seek ndbm"));
        return(-1);
    }
    while(ftell(db->dbfile) <= top) {
        if(fgets(line, MAX_HISTORYLINE, db->dbfile) == NULL) {
            if(debug_on)
                syslog(LOG_DEBUG, "%s", _strerror("Read ndbm"));
            return(-1);
        }
        if(tkey <= line_value(line))
            break;
    }
    cut_trailing(line);
    return(ftell(db->dbfile));
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
long ndbm_next(NDBM *db, long offs, char *line)
{
    if(db == NULL)
        syslog(LOG_ERR, "Next ndbm: NULL handle");
    else if(fseek(db->dbfile, offs, SEEK_SET))
        syslog(LOG_ERR, "%s", _strerror("Seek ndbm"));
    else {
        while(fgets(line, MAX_HISTORYLINE, db->dbfile)) {
            if(line_value(line) != 0) {
                cut_trailing(line);
                return(ftell(db->dbfile));
            }
        }
    }
    return(-1);
}


#if 0
/************************************************************************/
/*                                                                      */
/* I don't know what to do with this one. May be it's useable sometime. */
/*                                                                      */
/************************************************************************/
int ndbm_reindex(const char *name)
{
    long offset;
    char *cp;
    char *line;
    NDBM *db = memset(malloc(sizeof(NDBM)), 0, sizeof(NDBM));
    char *filename = malloc(_MAX_PATH);

    db->bufrec = -1;
    strcpy(filename, name);
    strcat(filename, NDBM_EXT_DBF);
    if((db->dbfile = xfopen(filename, "rb")) == NULL) {
        syslog(LOG_ERR, "%s", _strerror(filename));
        free(filename);
        free(db);
        return(-1);
    }

    strcpy(filename, name);
    strcat(filename, NDBM_EXT_IDX);
    unlink(filename);
    if(nidx_init(db, filename, 0)) {
        fclose(db->dbfile);
        free(filename);
        free(db);
        return(-1);
    }

    line = malloc(8192);
    for(;;) {
        offset = ftell(db->dbfile);
        if(fgets(line, MAX_HISTORYLINE, db->dbfile) == NULL)
            break;
        if((cp = strchr(line, '\n')) == NULL) {
            fgets(line, MAX_HISTORYLINE, db->dbfile);
            continue;
        }
        *cp = '\0';
        if((cp = strchr(line, ' ')) == NULL)
            continue;
        *cp = '\0';
        nidx_addkey(db, line, offset);
    }
    free(line);

    nidx_exit(db);
    fclose(db->dbfile);
    free(filename);
    free(db);

    return(0);
}

#endif

/*
 * $Log: ndbm.c,v $
 * Revision 1.1  1997/02/13 11:09:24  oliver
 * initial revision
 *
 */
