/*
 * This OS/2 part was written by Harald Kipp
 *
 * Starting with version 0.9p we handle the history file
 * exclusively by the server. This change has global
 * consequences to the whole distribution, as the server
 * is now the main center. Inews, Rnews, Chanx and all
 * the other programs accessing the history need the
 * server running in the background. All article storage
 * and retrieval is done by the server too.
 *
 * New articles are offered to Changi using IHAVE. Changi
 * may store the article and/or call an external program
 * feeded with the article either by stdin or a temporary
 * file.
 *
 * The purpose of this file is to provide all server threads
 * with shared access to the history file.
 *
 * Bug reports related should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

const char nhistory_rcsid[] = "$Id: nhistory.c,v 1.1 1997/02/13 11:09:25 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <chlimits.h>
#include <sysdep.h>
#include <mutex.h>
#include <syslog.h>
#include <arpatime.h>
#include <ndbm.h>

#include <nhistory.h>

static long mutex_wait = 32000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

static NDBM *ndbm;
static long nh_mutex = -1;
static int debug_on = 0;
static char historyline[MAX_HISTORYLINE];

volatile int change_flag;  /* set by API routines and cleared by flush thread */
volatile int flush_active; /* set if flush thread is running */
volatile int flush_exit;   /* set by API routines to terminate flush thread */

/************************************************************************/
/* Switch debugging on or off.                                          */
/*                                                                      */
/* 07/15/1996 harald                                                    */
/************************************************************************/
void debug_history(int enable)
{
    ndbm_debug(debug_on = enable);
}

/************************************************************************/
/*                                                                      */
/* history_maxwait                                                      */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void history_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/* Timer thread for history file flushing. The history will be flushed  */
/* if more than 10 seconds elapsed since it's latest change.            */
/*                                                                      */
/* 07/15/1996 harald                                                    */
/************************************************************************/
void _Optlink flush_timer(void *pv)
{
    int dirty = 0;
    int count = 0;

    /*
     * Signal API routines that we're active.
     */
    flush_active = 1;

    /*
     * Wake up every second in a loop.
     */
    for(;;) {
        ms_sleep(1000);

        /*
         * API routine wants us to terminate.
         */
        if(flush_exit)
            break;

        /*
         * If 'change_flag' has been set, then there
         * was another change within the last second.
         */
        if(change_flag) {
            change_flag = 0;
            dirty = 1;
            count = 0;
        }

        /*
         * If nothing changed within the last 10 seconds and
         * if our file buffer is dirty then let's flush it.
         */
        else if(dirty && ++count >= 10) {
            /*
             * 09/03/1996 harald
             * Changed from 30 to 0 seconds for faster
             * response to 'stop_flushing()' and to avoid
             * syslog message.
             */
            if(request_mutex(&nh_mutex, 0) == 0) {
                change_flag = dirty = count = 0;
                ndbm_flush(ndbm);
                release_mutex(nh_mutex);
            }
        }
    }

    /*
     * Signal API routines that we've stopped our activity.
     */
    flush_active = 0;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static void start_flushing(void)
{
    int retries = 100;

    /*
     * Initialize some flags and start flush timer.
     */
    flush_active = flush_exit = change_flag = 0;
    _beginthread(flush_timer, NULL, 8192, NULL);

    /*
     * Wait until flush timer signals activity.
     */
    while(retries-- > 0) {
        if(flush_active)
            break;
        ms_sleep(100);
    }

    /*
     * Looks like we're out of resources. It will catch us anyway.
     */
    if(!flush_active)
        syslog(LOG_ERR, "Can't create history flushing timer");
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static void stop_flushing(void)
{
    int retries = 100;

    /*
     * Signal flush timer that it should terminate.
     */
    flush_exit = 1;

    /*
     * Wait for flush timer becoming inactive.
     */
    while(retries-- > 0) {
        if(!flush_active)
            break;
        ms_sleep(100);
    }

    /*
     * Something strange has happened.
     */
    if(flush_active)
        syslog(LOG_ERR, "Can't destroy history flushing timer");

    flush_active = flush_exit = change_flag = 0;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int init_history(const char *name, int read_only)
{
    if(nh_mutex != -1 || (nh_mutex = create_mutex()) != -1) {
        if((ndbm = ndbm_open(name, read_only)) != NULL && read_only == 0)
            start_flushing();
    }
    return((ndbm == NULL) ? -1 : 0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int exit_history(void)
{
    stop_flushing();
    ndbm_close(ndbm);
    ndbm = NULL;
    destroy_mutex(nh_mutex);
    nh_mutex = -1;

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int reopen_history(const char *name, int read_only)
{
    stop_flushing();

    if(request_mutex(&nh_mutex, MUTEXWAIT) == 0) {
        ndbm_close(ndbm);
        if((ndbm = ndbm_open(name, read_only)) != NULL && read_only == 0)
            start_flushing();
        release_mutex(nh_mutex);
    }

    return((ndbm == NULL) ? -1 : 0);
}

/************************************************************************/
/*                                                                      */
/* addhistentry                                                         */
/*                                                                      */
/* Adds a new line to the history database.                             */
/*                                                                      */
/* Format of our history line is:                                       */
/*                                                                      */
/*      id arrived~expires~posted size group:num[,group:num...]         */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on errors.                             */
/*                                                                      */
/************************************************************************/
int addhistentry(char *id, long t_arr, long t_exp, long t_post, long size, char *artlist)
{
    int result = -1;

    /*
     * Format of our history line is:
     *
     * <id> arrived~expires~posted size group:num[,group:num...]
     */
    if(request_mutex(&nh_mutex, MUTEXWAIT) == 0) {
        if(t_exp)
            sprintf(historyline, "%s %ld~%ld~%ld %ld %s", id, t_arr, t_exp, t_post, size, artlist);
        else
            sprintf(historyline, "%s %ld~-~%ld %ld %s", id, t_arr, t_post, size, artlist);

        /*
         * Add the formatted line to the database
         */
        if((result = ndbm_store(ndbm, historyline)) == 0)
            change_flag = 1;
        release_mutex(nh_mutex);
    }

    return(result);
}

/************************************************************************/
/* pathlistbyid                                                         */
/*                                                                      */
/* Returns a space separated list of pathnames of article files with    */
/* the specified message-id.                                            */
/*                                                                      */
/************************************************************************/
int pathlistbyid(char *msg_id, char *pathlist)
{
    register char *cr;
    register char *cw;

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return(0);

    if(ndbm_fetch(ndbm, msg_id, pathlist)) {
        release_mutex(nh_mutex);
        return(0);  /* article not found */
    }
    cr = pathlist;
    release_mutex(nh_mutex);

    while(*cr && *cr != ' ' && *cr != '\t')
        cr++;
    while(*cr == ' ' || *cr == '\t')
        cr++;
    while(*cr && *cr != ' ' && *cr != '\t')
        cr++;
    while(*cr == ' ' || *cr == '\t')
        cr++;

    /*
     * Now cr points to the <grp:num> list.
     */
    for(cw = pathlist; *cr; cr++) {
        if(*cr == '.' || *cr == ':')
            *cw++ = '/';
        else if(*cr == ',')
            *cw++ = ' ';
        else
            *cw++ = *cr;
    }
    *cw = '\0';

    return(*pathlist != '\0');
}

/************************************************************************/
/* gethistart                                                           */
/*                                                                      */
/* Searches the history database for a given id.                        */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* msg_id       Points to the message id string.                        */
/*                                                                      */
/* artpath      Must either point to a string buffer to receive the     */
/*              articles pathname or should be set to NULL if the       */
/*              caller is not interested in this information.           */
/*              In the first case the buffer should usually contain     */
/*              the news home directory excluding a trailing backslash, */
/*              and this routine will append the article's path.        */
/* ------------ ------------------------------------------------------- */
/* result       Returns 1 if the id exists or 0 otherwise.              */
/*                                                                      */
/************************************************************************/
int gethistart(char *msg_id, char *artpath)
{
    register char *cp;
    char *cp1;
    char *buf;
    int result = 0;

    if(request_mutex(&nh_mutex, MUTEXWAIT))
        return(0);

    /*
     * Fetch the entry with the given message id
     */
    if(artpath == NULL) {
        result = (ndbm_fetch(ndbm, msg_id, NULL) == 0);
        release_mutex(nh_mutex);
        return(result);
    }

    if(ndbm_fetch(ndbm, msg_id, historyline)) {
        release_mutex(nh_mutex);
        return(0);  /* article not found */
    }
    cp = buf = strdup(historyline);
    release_mutex(nh_mutex);

    while(*cp && *cp != ' ' && *cp != '\t')
        cp++;
    while(*cp == ' ' || *cp == '\t')
        cp++;
    while(*cp && *cp != ' ' && *cp != '\t')
        cp++;
    while(*cp == ' ' || *cp == '\t')
        cp++;

    if(*cp) {
        /*
         * Now cp points to the first <grp:num> item.
         * Append the path of this message to artpath.
         */
        cp1 = strchr(artpath, '\0');
        *cp1++ = '\\';
        while(*cp && *cp != ' ' && *cp != '\t' && *cp != ',') {
            if(*cp == '.' || *cp == ':')
                *cp1++ = '\\';
            else
                *cp1++ = *cp;
            cp++;
        }
        *cp1 = '\0';
        result = 1;
    }

    free(buf);
    return(result);
}

/************************************************************************/
/* cancelhistart                                                        */
/*                                                                      */
/* Cancels an article with a given id in the history base.              */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* msg_id       Points to the message id string.                        */
/*                                                                      */
/* artlist      Must either point to a string buffer to receive the     */
/*              articles pathname or should be set to NULL if the       */
/*              caller is not interested in this information.           */
/*              In the first case the routine returns a comma separated */
/*              list of article paths relative to the news directory.   */
/* ------------ ------------------------------------------------------- */
/* result       Returns 0 if the article has been canceled, -1 otherwise*/
/*                                                                      */
/************************************************************************/
int cancelhistart(char *msg_id, char *artlist)
{
    int result = -1;

    if(*msg_id == '<' && *(msg_id + strlen(msg_id) - 1) == '>') {
        if(request_mutex(&nh_mutex, MUTEXWAIT) == 0) {
            if((result = ndbm_cancel(ndbm, msg_id, artlist)) == 0)
                change_flag = 1;
            release_mutex(nh_mutex);
        }
    }
    return(result);
}

/************************************************************************/
/* firsthist                                                            */
/*                                                                      */
/* Seeks through the history file looking for a line with date and time */
/* later than 'tkey'.                                                   */
/*                                                                      */
/* Returns the offset of the next line if successful, -1 otherwise.     */
/*                                                                      */
/************************************************************************/
long firsthist(unsigned long tkey, char *line)
{
    long result = -1;

    if(request_mutex(&nh_mutex, MUTEXWAIT) == 0) {
        if(debug_on)
            syslog(LOG_DEBUG, "Binary search history for time >= %ld", tkey);
        result = ndbm_first(ndbm, tkey, line);
        release_mutex(nh_mutex);
    }
    return(result);
}

/************************************************************************/
/*  nexthist                                                            */
/*                                                                      */
/*  Get the next line which begins at or after a given byte position    */
/*  of the history file. Lines longer than MAX_BUFSIZ will be cut,      */
/*  trailing spaces are removed.                                        */
/*                                                                      */
/*  Returns the offset of the next line if successful, -1 otherwise.    */
/*                                                                      */
/************************************************************************/
long nexthist(long offs, char *line)
{
    long result = -1;

    if(request_mutex(&nh_mutex, MUTEXWAIT) == 0) {
        result = ndbm_next(ndbm, offs, line);
        release_mutex(nh_mutex);
    }
    return(result);
}

/*
 * $Log: nhistory.c,v $
 * Revision 1.1  1997/02/13 11:09:25  oliver
 * initial revision
 *
 */
