/*  nidx.c
 *
 *  simple index manager for UUPC news
 *
 *  Author:  Kai Uwe Rommel <rommel@ars.muc.de>
 *  Created: Sun Aug 15 1993
 *
 *  Bug reports related to THIS modified version should be sent to
 *
 *  harald@os2point.ping.de
 *
 *  Special thanks to Lionel Widdifield for his suggestions
 *  on key encoding and other useful hints.
 *
 */

const char nidx_rcsid[] = "$Id: nidx.c,v 1.1 1997/02/13 11:09:26 oliver Exp $";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <share.h>
#include <io.h>
#include <errno.h>

#include <syslog.h>
#include <chlimits.h>
#include <sysdep.h>
#include <ndbm.h>

static int debug_on = 0;

/************************************************************************/
/* Switch debugging on or off.                                          */
/*                                                                      */
/* 07/15/1996 harald                                                    */
/************************************************************************/
void nidx_debug(int enable)
{
    debug_on = enable;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static void report_failure(char *act, short number, long offset, int line)
{
    static int first = 1;

    syslog(LOG_ERR, "%s page %d(%ld/%d)", _strerror(act), number, offset, line);
    if(first) {
        first = 0;
        fputs("History may be corrupted, run 'expire -rv'.\n", stderr);
    }
}

/* ==================================================================== */
/* ======================== page level primitives ===================== */
/* ==================================================================== */

/************************************************************************/
/*                                                                      */
/************************************************************************/
static short idx_new_page(NDBM *db)
{
    db->page_dirty = 0;
    db->page_number = db->size++;

    return(db->page_number);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int idx_get_page(NDBM *db, short number)
{
    long offset;

    db->page_dirty = 0;
    db->page_number = number;
    offset = db->page_number * sizeof(PAGE);

    if(lseek(db->fh, offset, SEEK_SET) == -1) {
        report_failure("Seek", number, offset, __LINE__);
        return -1;
    }

    if(read(db->fh, &db->page, sizeof(PAGE)) != sizeof(PAGE)) {
        report_failure("Read", number, offset, __LINE__);
        return -1;
    }

    return 0;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int idx_flush_page(NDBM *db)
{
    long offset;

    if(db->page_dirty) {
        db->page_dirty = 0;
        offset = db->page_number * sizeof(PAGE);

        if(lseek(db->fh, offset, SEEK_SET) == -1) {
            report_failure("Seek", db->page_number, offset, __LINE__);
            return -1;
	}

        if(write(db->fh, &db->page, sizeof(PAGE)) != sizeof(PAGE)) {
            report_failure("Write", db->page_number, offset, __LINE__);
            return -1;
        }
    }

    return 0;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int idx_push_page(NDBM *db, short number)
{
    if(idx_flush_page(db))
        return -1;

    if(db->page_stacksize == IDX_MAXSTACK) {
        syslog(LOG_ERR, "Index stack overflow");
        return -1;
    }

    db->page_stack[db->page_stacksize++] = db->page_number;

    return idx_get_page(db, number);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int idx_pop_page(NDBM *db)
{
    if(idx_flush_page(db))
        return -1;

    if(db->page_stacksize == 0) {
        syslog(LOG_ERR, "Index stack underflow");
        return -1;
    }

    return idx_get_page(db, db->page_stack[--(db->page_stacksize)]);
}

/* ==================================================================== */
/* ========================== the workhorses ========================== */
/* ==================================================================== */


#ifdef IDX_DUMP
/************************************************************************/
/*                                                                      */
/************************************************************************/
static void idx_dump(NDBM *db, char *adv)
{
    int i;

    printf("Page %d: %s\n", db->page_number, adv);
    for(i = 0; i < db->page.items; i++)
        printf("P%04d %-11.11s R%09lu\n", i, db->page.item[i].key, db->page.item[i].offset);
}
#endif

/************************************************************************/
/*                                                                      */
/************************************************************************/
static char *keycode(char *target, const char *source)
{
    unsigned long *code = (unsigned long *)target;
    unsigned long *u = (unsigned long *)source;
    size_t i = strlen(source) / 4;
    register size_t j;
    int adj;
    char *cp = target;

    memset(target, 0, IDX_MAXKEY);

    /*
     * 07/05/97 harald
     * Previous version crashed on keys with 3 or less
     * characters. Now we simply copy small keys.
     */
    if(i == 0)
        return(strcpy(target, source));

    if(i > 8)
        i = 8;
    for(j = 0, adj = 0; j < i - 1; j += 2, adj++) {
        code[0] ^= (u[j] << adj);
        code[1] ^= (u[j + 1] << adj);
    }
    for(j = 0; j < 8; j++, cp++)
        if(*cp < 32)
            *cp += 32;
    /* printf("%08lX%08lX %s %s\n", code[0], code[1], target, source); */

    return(target);
}

/************************************************************************/
/*                                                                      */
/* itmcmp                                                               */
/*                                                                      */
/* Compares an item with a given key.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* db           Database descriptor.                                    */
/*                                                                      */
/* key          Points to a null terminated key string.                 */
/*                                                                      */
/* n            Index into the item page.                               */
/* ------------ ------------------------------------------------------- */
/* Return value A negative value if key is less than item n,            */
/*              0 if key is equal to item n, or a                       */
/*              positive value if key is greater than item n.           */
/*                                                                      */
/************************************************************************/
static int itmcmp(NDBM *db, const char *key, const char *code, int n)
{
    int c;
    short l = (short)strlen(key);
    ITEM *item = &db->page.item[n];

    if((c = l - item->klen) == 0) {
        if((c = memcmp(code, item->key, IDX_MAXKEY)) == 0) {
            if(l > IDX_MAXKEY) {
                if(db->bufrec != item->offset) {
                    char *cp;

                    if(fseek(db->dbfile, item->offset, SEEK_SET))
                        syslog(LOG_ERR, "%s", _strerror("Seek ndbm"));
                    if(fgets(db->buffer, sizeof(db->buffer), db->dbfile) == NULL)
                        syslog(LOG_ERR, "%s", _strerror("Read ndbm"));
                    db->bufrec = item->offset;

                    /*
                     * Keep the buffer clean from linefeeds.
                     */
                    if((cp = strchr(db->buffer, '\n')) != NULL)
                        *cp = '\0';
                }
                c = memcmp(key, db->buffer, (size_t)l);
                /* printf("Re(%+d)'%s'\n", c, db->buffer); */
            }
        }
    }
    return(c);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int idx_search(NDBM *db, const char *key)
{
    int n, cmp;
    char code[IDX_MAXKEY];

    keycode(code, key);
    /* printf("\n  Cmp '%s'\n", key); */
    for (;;) {
        for (n = db->page.items - 1; n >= 0; n--) {
            /* printf(" with '%-11.11s'\n", db->page.item[n].key); */
            if((cmp = itmcmp(db, key, code, n)) == 0)
                return(n);
            if(cmp > 0) {
                if(db->page.item[n].child) {
                    if(idx_push_page(db, db->page.item[n].child))
                        return(-1);
                    break;
                }
                return(-1);
	    }
	}
        if(n < 0) {
            if(db->page.child_0) {
                if(idx_push_page(db, db->page.child_0))
                    break;
            }
	    else
                break;
        }
    }
    return(-1);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int idx_add(NDBM *db, const char *key, ITEM new)
{
    int result = 0;
    int i, n;
    char code[IDX_MAXKEY];

    keycode(code, key);
    for (;;) {
        /* printf("\n  Add '%s'\n", key); */
        for (n = db->page.items; n > 0; n--)
            if(itmcmp(db, key, code, n - 1) > 0)
		break;

        if(db->page.items < IDX_MAXITEM) {
	    /* we are lucky */

            for (i = db->page.items; i > n; i--) {
                db->page.item[i] = db->page.item[i - 1];
                db->page.item[i] = db->page.item[i - 1];
	    }

            db->page.items++;
            db->page.item[n] = new;
            db->page_dirty = 1;

#ifdef IDX_DUMP
            idx_dump(db, "Fits");
#endif
            result = idx_flush_page(db);
            break;
	}
	else {
	    /* split page */

	    ITEM up;
	    PAGE newpage;
            short child_0;

            if(n <= IDX_MINITEM) {
                if(n != IDX_MINITEM) {
                    up = db->page.item[IDX_MINITEM - 1];

		    for (i = IDX_MINITEM - 1; i > n; i--)
                        db->page.item[i] = db->page.item[i - 1];

                    db->page.item[n] = new;
		    new = up;
		}

		for (i = 0; i < IDX_MINITEM; i++)
                    newpage.item[i] = db->page.item[IDX_MINITEM + i];
	    }
	    else {
                up = db->page.item[IDX_MINITEM];
		n -= IDX_MINITEM;

		for (i = 0; i < n - 1; i++)
                    newpage.item[i] = db->page.item[IDX_MINITEM + i + 1];

		newpage.item[n - 1] = new;

		for (i = n; i < IDX_MINITEM; i++)
                    newpage.item[i] = db->page.item[IDX_MINITEM + i];

		new = up;
	    }

	    newpage.child_0 = new.child;
            db->page.items = newpage.items = IDX_MINITEM;

            if(db->page_stacksize == 0) {
		/* grow tree height */

		/* write both pieces onto new pages */

                child_0 = idx_new_page(db);

                db->page_dirty = 1;
#ifdef IDX_DUMP
                idx_dump(db, "Split 1");
#endif
                if((result = idx_flush_page(db)) != 0)
                    break;

                new.child = idx_new_page(db);
                memcpy(&db->page, &newpage, sizeof(PAGE));

                db->page_dirty = 1;
#ifdef IDX_DUMP
                idx_dump(db, "Split 2");
#endif
                if((result = idx_flush_page(db)) != 0)
                    break;

		/* and keep the root on page zero */

                db->page.items = 1;
                db->page.item[0] = new;
                db->page.child_0 = child_0;

                db->page_number = 0;
                db->page_dirty = 1;
#ifdef IDX_DUMP
                idx_dump(db, "New root");
#endif
                result = idx_flush_page(db);
                break;
	    }
	    else {
		/* write lower half onto old page */

                db->page_dirty = 1;
#ifdef IDX_DUMP
                idx_dump(db, "Split 1");
#endif
                if((result = idx_flush_page(db)) != 0)
                    break;

		/* and upper one onto new one */

                new.child = idx_new_page(db);
                memcpy(&db->page, &newpage, sizeof(PAGE));

                db->page_dirty = 1;
#ifdef IDX_DUMP
                idx_dump(db, "Split 2");
#endif
                if((result = idx_flush_page(db)) != 0)
                    break;

		/* and insert middle key into parent page */

                if((result = idx_pop_page(db)) != 0)
                    break;
	    }
	}
    }
    return(result);
}

/* ==================================================================== */
/* ======================= interface functions ======================== */
/* ==================================================================== */

/************************************************************************/
/*                                                                      */
/************************************************************************/
int nidx_init(NDBM *db, char *filename, int read_only)
{
    long size;
    unsigned int oflg;
    unsigned int shflg;
    int i;

    if(read_only) {
        oflg = O_RDONLY | O_BINARY;
        shflg = SH_DENYNO;
    }
    else {
        oflg = O_RDWR | O_CREAT | O_BINARY;
        shflg = SH_DENYWR;
    }
    for(i = 20; --i >= 0; ) {
        if((db->fh = _sopen(filename, (int)oflg, (int)shflg, S_IREAD | S_IWRITE)) != -1)
            break;
        if(errno != EACCES)
            break;
        ms_sleep(1000);
    }
    if(db->fh == -1) {
        syslog(LOG_ERR, "%s", _strerror(filename));
        return(-1);
    }
    if((size = lseek(db->fh, 0, SEEK_END)) == -1) {
        syslog(LOG_ERR, "Seek end %s", _strerror(filename));
        return(-1);
    }
    if(size == 0) {
        if(chsize(db->fh, sizeof(PAGE)) != 0) {
            syslog(LOG_ERR, "Resize %s", _strerror(filename));
            return(-1);
        }
        db->size = 0;
        idx_new_page(db);
    }
    else if(size % sizeof(PAGE)) {
        syslog(LOG_ERR, "Inconsistent index size %ld", size);
        return(-1);
    }
    else
        db->size = (short)(size / sizeof(PAGE));
    db->page_number = -1;

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void nidx_exit(NDBM *db)
{
    if(db) {
        idx_flush_page(db);
        close(db->fh);
    }
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int nidx_addkey(NDBM *db, const char *key, const long offset)
{
    ITEM new;

    if(db == NULL)
        return -1;
    if(idx_get_page(db, 0) == -1)
        return -1;
    db->page_stacksize = 0;

    if(idx_search(db, key) != -1) {
        if(debug_on)
            syslog(LOG_DEBUG, "%s exists, not added", key);
        return(-1);
    }

    new.klen = (short)strlen(key);
    keycode(new.key, key);
    new.offset = offset;
    new.child = 0;

    if(idx_add(db, key, new) == -1) {
        if(debug_on)
            syslog(LOG_DEBUG, "Failed to add %s", key);
        return(-1);
    }
    if(debug_on)
        syslog(LOG_DEBUG, "Added %s", key);

    return 0;

}

/************************************************************************/
/*                                                                      */
/************************************************************************/
long nidx_getkey(NDBM *db, const char *key)
{
    int pos;

    if(db == NULL)
        return -1;

    if(idx_get_page(db, 0) == -1)
        return -1;
    db->page_stacksize = 0;

    if((pos = idx_search(db, key)) == -1) {
        if(debug_on)
            syslog(LOG_DEBUG, "%s not found", key);
        return -1;
    }
    if(debug_on)
        syslog(LOG_DEBUG, "%s found at %ld", key, db->page.item[pos].offset);

    return(db->page.item[pos].offset);
}

/*
 * $Log: nidx.c,v $
 * Revision 1.1  1997/02/13 11:09:26  oliver
 * initial revision
 *
 */
