const char overvfmt_rcsid[] = "$Id$";

/*
 * This part was done by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 */
/*
 * $Log$
 */



#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>

#include <syslog.h>
#include <xfile.h>

#include <loadtext.h>

#include <overview.h>
#include <overvfmt.h>

static OVERVIEW_FMT std_ovfmt[] = {
    { "Subject"   , 0 },
    { "From"      , 0 },
    { "Date"      , 0 },
    { "Message-ID", 0 },
    { "References", 0 },
    { "Bytes"     , 0 },
    { "Lines"     , 0 }
};

extern FILE *openartlink(const char *path);

static int novfmt = sizeof(std_ovfmt) / sizeof(std_ovfmt[0]);
static OVERVIEW_FMT *ovfmt = std_ovfmt;
static int debug_on = 0;
static int ibytes = 6;

/************************************************************************/
/*                                                                      */
/* group2path                                                           */
/*                                                                      */
/************************************************************************/
static char *group2path(char *path, const char *home, const char *group)
{
    register char *cp;
    char *result = NULL;
    size_t l_home = strlen(home);

    if(l_home + strlen(group) + 12 <= _MAX_PATH) {
        result = memcpy(path, home, l_home);
        cp = path + l_home;
        *cp++ = '\\';
        for(; *group; group++)
            if(*group == '.')
                *cp++ = '\\';
            else
                *cp++ = *group;
        *cp = '\0';
    }

    return(result);
}


/************************************************************************/
/*                                                                      */
/* debug_overviewfmt                                                    */
/*                                                                      */
/* Switch overview format debugging on/off.                             */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* enable       A value of 0 will switch off debugging, otherwise it    */
/*              is switched on.                                         */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void debug_overviewfmt(int enable)
{
    debug_on = enable;
}

static char line_none[32];
static char item_none[32];
static char newspath[_MAX_PATH];

/************************************************************************/
/*                                                                      */
/************************************************************************/
int init_overviewfmt(char *formatfile, char *newsdir,
                     char *linedummy, char *itemdummy)
{
    if(debug_on) {
        if(*formatfile)
            syslog(LOG_DEBUG, "Overview format in %s", formatfile);
        else
            syslog(LOG_DEBUG, "Default overview format");
        syslog(LOG_DEBUG, "Overview dummies: '%s' '%s'", linedummy, itemdummy);
    }
    strcpy(newspath, newsdir);
    strcpy(line_none, linedummy);
    strcpy(item_none, itemdummy);
    return(load_overviewfmt(formatfile));
}

/************************************************************************/
/*                                                                      */
/* load_overviewfmt                                                     */
/*                                                                      */
/* Reads the overview schema file and fills a global array of           */
/* OVERVIEW_FMT structures. In case of any errors while reading the     */
/* file, the array is set to a default contents.                        */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* name         Points to the string that specifies the pathname of the */
/*              overview schema file, usually 'overview.fmt'.           */
/* ------------ ------------------------------------------------------- */
/* Return value none                                                    */
/*                                                                      */
/************************************************************************/
int load_overviewfmt(const char *name)
{
    char **lines;
    char *cp;
    int i, n;

    /*
     * Completely load the file into core memory. The array
     * 'lines' will be allocated inside this routine and
     * contain pointers to each line with leading and trailing
     * spaces and also comments removed. Note that comments
     * need a '#' character in the first column.
     */
    if(name && *name && (n = load_textfile(name, &lines)) > 0) {
        if(debug_on)
            syslog(LOG_DEBUG, "Loaded %d items from %s", n, name);
        novfmt = n;
        ibytes = 0;
        ovfmt = malloc(novfmt * sizeof(OVERVIEW_FMT));
        for(i = 0; i < novfmt; i++) {
            /*
             * Get the name of the header line and try to
             * find it in our global KEWORD array.
             */
            if((cp = strchr(lines[i], ':')) != NULL)
                *cp = '\0';

            /*
             * Add the id of this header to our array and determine
             * if the string 'full' has been added to the name of
             * the header. In this case the name of the header line
             * itself will be added to the overview file. This option
             * is recommended for the 'Xref:' header, but will not be
             * used in our standard setup to provide compatibility
             * with previous versions.
             */
            ovfmt[i].name = strdup(lines[i]);
            if(strcmp(lines[i], "Bytes") == 0)
                ibytes = i + 1;
            ovfmt[i].full = (cp != NULL && stricmp(cp + 1, "full") == 0);
        }
        free(*lines);
        free(lines);
    }
    else if(debug_on)
        syslog(LOG_DEBUG, "Using standard overview");
    return(0);
}

/************************************************************************/
/*                                                                      */
/* overview_items                                                       */
/*                                                                      */
/* Returns the number of overview items.                                */
/*                                                                      */
/************************************************************************/
int overview_items(void)
{
    return(novfmt);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int overview_index(char *field)
{
    register int i;

    for(i = 0; i < novfmt; i++)
        if(stricmp(field, ovfmt[i].name) == 0)
            return(i + 1);
    return(0);
}

/************************************************************************/
/*                                                                      */
/* make_overline                                                        */
/*                                                                      */
/* Creates an overview line by reading the header of an article file.   */
/* Note, that this routine is not reentrant.                            */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to a zero teminated string which specifies the   */
/*              name of the newsgroup.                                  */
/*                                                                      */
/* artnum       Article number of the requested overview line.          */
/*                                                                      */
/* ovrline      Points to a buffer, where this routine will store       */
/*              the overview line or item. Its size should be at        */
/*              least MAX_OVERLINE number of bytes.                     */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to ovrline buffer.                              */
/*                                                                      */
/************************************************************************/
char *make_overline(char *group, long artnum, char *ovrline)
{
    register char *cp;
    register char *cp2;
    int  i;
    FILE *fp;
    int  sz = novfmt * sizeof(char *);
    long fsz;
    int  nitems = novfmt;
    char **iarray = memset(malloc(sz), 0, sz);;

    /*
     * Create the pathname of the article file.
     */
    group2path(ovrline, newspath, group);
    cp = ovrline + strlen(ovrline);
    *cp++ = '\\';
    ltoa(artnum, cp, 10);
    if(debug_on)
        syslog(LOG_DEBUG, "Overview from %s", ovrline);

    /*
     * Open the article file and read the header.
     */
    if((fp = openartlink(ovrline)) != NULL) {
        i = 0;

        /*
         * 08/20/97 harald
         * If we need the number of bytes, we take this item
         * from the real filesize. This is done because of
         * two reasons: 1. Usual message headers don't
         * have this item. 2. We don't need to read the complete
         * header, if we already found all remaining fields.
         */
        if(ibytes) {
            fsz = filelength(fileno(fp));
            iarray[ibytes - 1] = strdup(ltoa(fsz, ovrline, 10));
            nitems--;
        }
        while(fgets(ovrline, MAX_OVERLINE, fp)) {
            if((cp = strchr(ovrline, '\n')) == NULL)
                break;
            *cp = '\0';
            if(*ovrline == 0) {
                break;
            }

            /*
             * 07/07/97 harald
             * Allow continuation lines.
             */
            if(*ovrline == ' ' || *ovrline == '\t') {
                if(i) {
                    sz = strlen(iarray[i - 1]);
                    cp2 = ovrline;
                    while(*cp2 == ' ' || *cp2 == '\t')
                        cp2++;

                    cp = iarray[i - 1] = realloc(iarray[i - 1], sz + strlen(cp2) + 2);

                    cp += sz;
                    *cp++ = ' ';
                    for(; *cp2; cp2++)
                        if(*cp2 == '\t')
                            *cp++ = ' ';
                        else
                            *cp++ = *cp2;
                    *cp = '\0';
                }
                continue;
            }
            else if((cp2 = strchr(ovrline, ':')) == NULL) {
                break;
            }
            *cp2++ = '\0';
            while(*cp2 == ' ' || *cp2 == '\t')
                cp2++;
            if((i = overview_index(ovrline)) != 0) {
                if(iarray[i - 1] == NULL) {
                    cp = iarray[i - 1] = malloc(strlen(cp2) + 1);
                    for(; *cp2; cp2++)
                        if(*cp2 == '\t')
                            *cp++ = ' ';
                        else
                            *cp++ = *cp2;
                    *cp = '\0';
                    if(--nitems <= 0) {
                        break;
                    }
                }
                else
                    i = 0;
            }
        }
        fclose(fp);
    }

    /*
     * Now create the line from the array of collected items.
     */
    for(cp = ovrline, i = 0; ; ) {
        if(iarray[i] == NULL)
            strcpy(cp, line_none);
        else {
            /*
             * 05/24/97 harald
             * Corrected a bug: Do only output name of header, if
             * item exists.
             */
            if(ovfmt[i].full) {
                strcpy(cp, ovfmt[i].name);
                cp += strlen(cp);
                *cp++ = ':';
                /*
                 * 07/11/97 harald
                 * Added a space. slrn wasn't able to detect the xref header
                 */
                *cp++ = ' ';
            }
            strcpy(cp, iarray[i]);
            free(iarray[i]);
        }
        if(++i >= novfmt)
            break;
        cp += strlen(cp);
        *cp++ = '\t';
    }
    free(iarray);

    return(ovrline);
}

/************************************************************************/
/*                                                                      */
/* Two things done here: First check for end of line, otherwise         */
/* we are in trouble if overview.fmt had been expanded.                 */
/* Second, use the string '(none)' if any item is missing.              */
/*                                                                      */
/************************************************************************/
char *extract_overitem(char *line, int item)
{
    char *cp, *wp;
    int full;

    if(item > 0 && item <= novfmt) {
        full = ovfmt[item - 1].full;
        for(wp = cp = line; --item; cp++) {
            while(*cp && *cp != '\t')
                cp++;
            if(*cp == '\0')
                break;
        }
        if(full && (wp = strchr(cp, ':')) != NULL) {
            cp = wp + 1;
            while(*cp == ' ')
                cp++;
        }
        wp = line;
        while(*cp && *cp != '\t')
            *wp++ = *cp++;
        *wp = '\0';
        if(*line == '\0')
            strcpy(line, item_none);
    }
    return(line);
}
