/*
 * This part was done by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

const char overview_rcsid[] = "$Id: overview.c,v 1.1 1997/02/13 11:09:33 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>

#include <syslog.h>

#include <mutex.h>
#include <overview.h>
#include <overvfmt.h>
#include <ovrcache.h>

static long mutex_wait = 33000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

static long mutex = -1;
static int  debug_on = 0;

/************************************************************************/
/*                                                                      */
/* find_long                                                            */
/*                                                                      */
/* Finds a long value in a sorted array and returns the zero based      */
/* index.                                                               */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* key          Value to be found.                                      */
/*                                                                      */
/* array        Array of values.                                        */
/*                                                                      */
/* num          Number of values in 'array'.                            */
/* ------------ ------------------------------------------------------- */
/* Return value Retuns an index into array or -1 if the value couldn't  */
/*              be found in the array.                                  */
/*                                                                      */
/************************************************************************/
static int find_long(long key, long array[], int num)
{
    register int m = 0;
    register int l = 0;
    long val;

    num--;
    while(l <= num) {
        if(key < (val = array[m = (l + num) / 2]))
            num = m - 1;
        else if (key > val)
            l = m + 1;
        else
            return(m);
    }
    return(-1);
}

/************************************************************************/
/*                                                                      */
/* debug_overview                                                       */
/*                                                                      */
/* Switch overview debugging on/off.                                    */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* enable       A value of 0 will switch off debugging, otherwise it    */
/*              is switched on.                                         */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void debug_overview(int enable)
{
    debug_on = enable;
    debug_overviewfmt(enable);
    debug_overcache(enable);
}

/************************************************************************/
/*                                                                      */
/* overview_maxwait                                                     */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void overview_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int init_overview(char *formatfile, char *newsdir, char *overdir,
                  char *headerdummy, int cachesize)
{
    if(debug_on)
        syslog(LOG_DEBUG, "Init overview for %s", newsdir);
    if((mutex = create_mutex()) == -1)
        return(-1);

    init_overcache(overdir, newsdir, cachesize);
    return(init_overviewfmt(formatfile, newsdir, headerdummy, "(none)"));
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void exit_overview(void)
{
    if(mutex != -1) {
        flush_overview();
        destroy_mutex(mutex);
    }
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void flush_overview(void)
{
    if(request_mutex(&mutex, MUTEXWAIT))
        return;

    flush_overcache();

    release_mutex(mutex);
}

/************************************************************************/
/*                                                                      */
/* get_overline                                                         */
/*                                                                      */
/* Get a line or specified item from the overview file. If no overview  */
/* data exists for the specified article, it will be created.           */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to a zero teminated string which specifies the   */
/*              name of the newsgroup.                                  */
/*                                                                      */
/* artnum       Number of the article of the requested overview line.   */
/*                                                                      */
/* line         Points to a buffer, where this routine will store       */
/*              the overview line or item. Its size should be at        */
/*              least MAX_OVERLINE number of bytes.                     */
/*                                                                      */
/* item         The index of the requested item. A list of available    */
/*              items may be given by a file named 'overview.fmt'.      */
/*              If this value is zero, the whole line will be returned. */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, 1 if line faked, -1 in case of any error. */
/*                                                                      */
/************************************************************************/
int get_overline(char *group, long artnum, char *line, int item)
{
    int result = 0;
    int i;
    OVERBUFF *ovr;

    /*
     * Request exclusive access to the overview base.
     */
    if(request_mutex(&mutex, MUTEXWAIT))
        return(-1);

    /*
     * Get the overview info of this group from our overview cache.
     * The routine 'fetch_overbuff' will automatically update
     * 'art_array' and maintain the cache.
     */
    ovr = fetch_overbuff(group);

    /*
     * If the requested article number couldn't be found in 'art_array',
     * then the article doesn't exist. Return with an error.
     */
    if((i = find_long(artnum, ovr->art_array, ovr->n_arts)) == -1) {
        release_mutex(mutex);
        if(debug_on)
            syslog(LOG_DEBUG, "No article %s:%ld", group, artnum);
        return(-1);
    }

    /*
     * If the overview line of the requested article is empty, create
     * one from the article's header by calling 'make_overline'.
     */
    if(ovr->line[i] == NULL) {
        if(debug_on)
            syslog(LOG_DEBUG, "Fake overview %s:%ld", group, artnum);
        ovr->line[i] = strdup(make_overline(group, artnum, line));
        ovr->added[i] = 'A';
        ovr->is_dirty = 1;
        result = 1;
    }

    /*
     * If the caller doesn't asked us for the complete line, then
     * extract the requested item.
     */
    strcpy(line, ovr->line[i]);
    if(item > 0)
        extract_overitem(line, item);

    /*
     * Release exclusive access to the overview base and return.
     */
    release_mutex(mutex);

    return(result);
}

/*
 * $Log: overview.c,v $
 * Revision 1.1  1997/02/13 11:09:33  oliver
 * initial revision
 *
 */
