/*
 * This part was done by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

const char ovrcache_rcsid[] = "$Id: ovrcache.c,v 1.1 1997/02/13 11:09:33 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>

#include <syslog.h>
#include <xfile.h>

#include <loadfile.h>

#include <dircache.h>
#include <nactive.h>
#include <overview.h>
#include <ovrcache.h>


typedef struct {
    int n_buffs;
    OVERBUFF *head;
    OVERBUFF *tail;
} OVERCACHE;

static OVERCACHE ovc;
static char newspath[_MAX_PATH];
static char overpath[_MAX_PATH];
static char overname[_MAX_PATH];

static int debug_on = 0;

static int  maxcache = 32;

/*======================================================================*/
/* LOCAL ROUTINES                                                       */
/*======================================================================*/

/************************************************************************/
/*                                                                      */
/* group2path                                                           */
/*                                                                      */
/************************************************************************/
static char *group2path(char *path, const char *home, const char *group)
{
    register char *cp;
    char *result = NULL;
    size_t l_home = strlen(home);

    if(l_home + strlen(group) + 12 <= _MAX_PATH) {
        result = memcpy(path, home, l_home);
        cp = path + l_home;
        *cp++ = '\\';
        for(; *group; group++)
            if(*group == '.')
                *cp++ = '\\';
            else
                *cp++ = *group;
        *cp = '\0';
    }

    return(result);
}

/************************************************************************/
/*                                                                      */
/* load_overview                                                        */
/*                                                                      */
/* Loads the overview info structure of a specified newsgroup.          */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to a zero teminated string which specifies the   */
/*              name of the newsgroup.                                  */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to the loaded overview info structure on        */
/*              success or NULL in case of any failure. The structure   */
/*              storage will be allocated from heap space and must be   */
/*              freed by the caller.                                    */
/*                                                                      */
/* Note: This routine is not reentrant.                                 */
/************************************************************************/
static OVERBUFF *load_overview(const char *group)
{
    OVERBUFF *ovr;
    register char *cp;
    char *lp, *ip;
    long ovrnum;
    int i;

    if(group2path(overname, overpath, group) == NULL) {
        syslog(LOG_NOTICE, "Illegal newsgroup %s", group);
        return(NULL);
    }
    strcat(overname, "\\.overview");

    ovr = memset(malloc(sizeof(OVERBUFF)), 0, sizeof(OVERBUFF));
    strcpy(ovr->group, group);

    if((ovr->n_arts = query_groupdir(group, &(ovr->art_array))) > 0) {
        ovr->line = memset(malloc(ovr->n_arts * sizeof(char *)), 0,
                           ovr->n_arts * sizeof(char *));
        ovr->added = memset(malloc(ovr->n_arts), 0, ovr->n_arts);

        /*
         * 09/05/1996 harald
         * Test existence first to avoid error message
         * from the inside of 'load_file()'.
         */
        if(access(overname, 0) == 0 &&
           (ovr->buf = load_file(overname)) != NULL) {
            if(debug_on)
                syslog(LOG_DEBUG, "Load %s", overname);
            for(i = 0, lp = ovr->buf; ; lp = cp + 1) {
                ip = NULL;
                for(cp = lp; *cp && *cp != '\n'; cp++)
                    if(ip == NULL && *cp == '\t')
                        ip = cp;
                if(*cp == '\0')
                    break;
                *cp = '\0';
                if(*lp == '\0')
                    continue;
                if(ip == NULL) {
                    if(debug_on)
                        syslog(LOG_DEBUG, "Missing tab in %s", overname);
                    ovr->is_dirty = 1;
                    continue;
                }
                *ip = '\0';
                if((ovrnum = atol(lp)) < ovr->art_array[i]) {
                    if(debug_on)
                        syslog(LOG_DEBUG, "Bad line in %s", overname);
                    ovr->is_dirty = 1;
                    continue;
                }
                while(ovrnum > ovr->art_array[i] && i < ovr->n_arts)
                    i++;
                if(i == ovr->n_arts) {
                    ovr->is_dirty = 1;
                    if(debug_on)
                        syslog(LOG_DEBUG, "Too many lines in %s", overname);
                    break;
                }
                if(ovrnum == ovr->art_array[i]) {
                    if(*(cp - 1) == '\r')
                        *(cp - 1) = '\0';
                    ovr->line[i] = ip + 1;
                }
                else if(debug_on)
                    syslog(LOG_DEBUG, "Article %s:%ld vanished", group, ovrnum);
            }
        }
        else if(debug_on)
            syslog(LOG_DEBUG, "Can't load %s", overname);
    }
    else {
        if(debug_on)
            syslog(LOG_DEBUG, "Empty group %s", group);
        ovr->is_dirty = 1;
    }


    return(ovr);
}

/*======================================================================*/
/* GLOBAL ROUTINES                                                      */
/*======================================================================*/

/************************************************************************/
/*                                                                      */
/* debug_overcache                                                      */
/*                                                                      */
/* Switch overview cache debugging on/off.                              */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* enable       A value of 0 will switch off debugging, otherwise it    */
/*              is switched on.                                         */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void debug_overcache(int enable)
{
    debug_on = enable;
}

#if 0
/************************************************************************/
/*                                                                      */
/************************************************************************/
void dumptree(void)
{
    OVERBUFF *obp;

    fprintf(stderr, "DUMP: head=%08lx tail=%08lx", ovc.head, ovc.tail);
    for(obp = ovc.head; obp != NULL; ) {
        fprintf(stderr, "%08lx: %s %08lx %08lx", obp, obp->group, obp->prev, obp->next);
        obp = obp->next;
    }
}
#endif

/************************************************************************/
/*                                                                      */
/************************************************************************/
void init_overcache(char *overdir, char *newsdir, int cachesize)
{
    if(debug_on)
        syslog(LOG_DEBUG, "Overview cache for %d entries in %s", cachesize, overdir);
    strcpy(overpath, overdir);
    strcpy(newspath, newsdir);
    if(cachesize)
        maxcache = cachesize;
}

/************************************************************************/
/*                                                                      */
/* save_overview                                                        */
/*                                                                      */
/* Saves an overview file of a given overview info structure, if the    */
/* dirty bit is set and releases any allocated memory.                  */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* ovr          Pointer to the overview info structure to be saved.     */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 in case of any error.                  */
/*                                                                      */
/************************************************************************/
int save_overview(OVERBUFF *ovr)
{
    int result = 0;
    FILE *fp = NULL;
    int  i;
    char *cp;

    /*
     * Write the overview info back to a file, if the dirty bit is set.
     */
    if(ovr->is_dirty) {
        ovr->is_dirty = 0;
        if(group2path(overname, overpath, ovr->group) != NULL) {
            strcat(overname, "\\.overview");
            if(debug_on)
                syslog(LOG_DEBUG, "%ste %s", (ovr->n_arts ? "Upda" : "Dele"), overname);
            remove(overname);
            for(i = 0; i < ovr->n_arts; i++) {
                if(ovr->line[i] != NULL) {
                    if(fp == NULL && (fp = xfopen(overname, "wb")) == NULL) {
                        syslog(LOG_ERR, "%s", _strerror(overname));
                        result = -1;
                        break;
                    }
                    fprintf(fp, "%ld\t%s\n", ovr->art_array[i], ovr->line[i]);
                }
            }
            if(fp != NULL && fclose(fp)) {
                syslog(LOG_ERR, "%s", _strerror(overname));
                result = -1;
            }
        }
        else {
            syslog(LOG_NOTICE, "Invalid group name: %s", ovr->group);
            result = -1;
        }
    }

    /*
     * Free any allocated heap memory.
     */
    if(ovr->n_arts) {
        for(cp = ovr->added, i = 0; i < ovr->n_arts; cp++, i++)
            if(*cp != '\0')
                free(ovr->line[i]);
        free(ovr->line);
        free(ovr->added);
        free(ovr->art_array);
        if(ovr->buf)
            free(ovr->buf);
    }
    free(ovr);
    return(result);
}



/************************************************************************/
/*                                                                      */
/* fetch_overbuff                                                       */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* group        Points to a zero teminated string which specifies the   */
/*              name of the newsgroup.                                  */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 in case of any error.                  */
/*                                                                      */
/************************************************************************/
OVERBUFF *fetch_overbuff(char *group)
{
    OVERBUFF *ovr;
    int num;
    long lo, hi;

    /*
     * If our cache is empty, then load the first overview info.
     */
    if(ovc.n_buffs == 0) {
        ovc.head = ovc.tail = load_overview(group);
        ovc.n_buffs = 1;
        return(ovc.head);
    }

    /*
     * Our group is the last recently used.
     */
    if(strcmp(ovc.head->group, group) == 0) {
        ovr = ovc.head;

        /*
         * If the number of articles and the lowest article number
         * are unchanged, then we assume that our overview info is
         * still valid.
         */
        if(query_active(group, &num, &lo, &hi) != '\0' &&
           num == ovr->n_arts && (num == 0 || lo == ovr->art_array[0])) {
            return(ovr);
        }

        /*
         * Either articles were added or removed. Save the current
         * overview info (which will also free the allocated heap
         * space) and reload the info (which will add overview lines
         * of new and remove lines of dropped articles).
         */
        if(debug_on)
            syslog(LOG_DEBUG, "Head resync %s", group);
        if((ovc.head = ovr->next) == NULL) {
            save_overview(ovr);
            ovc.head = ovc.tail = load_overview(group);
            ovc.n_buffs = 1;
            return(ovc.head);
        }
        save_overview(ovr);
        ovr = load_overview(group);
        ovc.head->prev = ovr;
        ovr->next = ovc.head;
        return(ovc.head = ovr);
    }

    /*
     * Our group is _not_ the recently used one. See, if we can locate
     * it in our cache.
     */
    for(ovr = ovc.head; (ovr = ovr->next) != NULL; ) {
        /*
         * Search the requested group.
         */
        if(strcmp(ovr->group, group))
            continue;

        /*
         * We have found it. Maintain our linked list by moving this
         * overview info to the head.
         */
        if(ovr->next)
            ovr->next->prev = ovr->prev;
        else
            ovc.tail = ovr->prev;
        ovr->prev->next = ovr->next;
        ovr->next = ovc.head;
        ovr->prev = NULL;
        ovc.head->prev = ovr;
        ovc.head = ovr;

        /*
         * Return the info to the caller, if it's uptodate. Otherwise
         * reload it before returning.
         */
        if(query_active(group, &num, &lo, &hi) != '\0' &&
           num == ovr->n_arts && (num == 0 || lo == ovr->art_array[0]))
            return(ovr);
        if(debug_on)
            syslog(LOG_DEBUG, "Body resync %s", group);
        if((ovc.head = ovr->next) == NULL) {
            save_overview(ovr);
            ovc.head = ovc.tail = load_overview(group);
            ovc.n_buffs = 1;
            return(ovc.head);
        }
        save_overview(ovr);
        ovr = load_overview(group);
        ovc.head->prev = ovr;
        ovr->next = ovc.head;
        return(ovc.head = ovr);
    }

    /*
     * We reach this point, if the requested group is not in our
     * cache. If our cache is full, we save and remove the tail.
     */
    if(ovc.n_buffs >= maxcache) {
        ovr = ovc.tail;
        ovc.tail = ovr->prev;
        ovr->prev->next = NULL;
        ovc.n_buffs--;
        ovc.head->prev = NULL;
        save_overview(ovr);
    }
    ovr = load_overview(group);
    ovr->next = ovc.head;
    ovc.head->prev = ovr;
    ovc.head = ovr;
    ovc.n_buffs++;
    return(ovr);
}


/************************************************************************/
/*                                                                      */
/************************************************************************/
void flush_overcache(void)
{
    OVERBUFF *ovr, *ovrnext;

    for(ovr = ovc.head; ovr != NULL; ovr = ovrnext) {
        ovrnext = ovr->next;
        save_overview(ovr);
    }
    ovc.n_buffs = 0;
    ovc.head = ovc.tail = NULL;
}

/*
 * $Log: ovrcache.c,v $
 * Revision 1.1  1997/02/13 11:09:33  oliver
 * initial revision
 *
 */
