/*----------------------------------------------------------------------*/
/* Copyright (c) 1996-97 by Harald Kipp. All rights reserved.           */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* Parts are copyright by Kendra Electronic Wonderworks.                */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * This module contains the routines to load a linked list of
 * remote hosts from and to save one back to the disk. It
 * is currently used by Changi Lite.
 */

const char rmtlink_rcsid[] = "$Id: rmtlink.c,v 1.1 1997/02/13 11:09:35 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>
#include <direct.h>
#include <time.h>

#include <xfile.h>
#include <loadtext.h>
#include <domnpath.h>
#include <rmtlink.h>

/************************************************************************/
/*                                                                      */
/* insert_rmtlink                                                        */
/*                                                                      */
/* Merges a new linked list into an existing one. Keeps the existing    */
/* list in sorted order.                                                */
/*                                                                      */
/************************************************************************/
RMTLINK *insert_rmtlink(RMTLINK *rmtl, RMTLINK *rmti)
{
    RMTLINK *rmtix;
    RMTLINK *rmtlx;
    int rc;

    if(rmti == NULL)
        return(rmtl);

    /*
     * If the target list is empty, add the first source item.
     */
    if(rmtl == NULL) {
        rmtl = rmti;
        rmtl->next = NULL;
        rmtl->prev = NULL;
        rmti = rmti->next;
    }

    /*
     * Loop for each element of the source list.
     */
    if((rmtix = rmti) != NULL) {

        /*
         * Find the correct entry point in the target list.
         * We must keep the target list in alphabetical order.
         */
        for(rmtlx = rmtl; rmtlx; rmtlx = rmtlx->next) {

            /*
             * If the new element already exists in the target
             * list, it is simply ignored.
             */
            if((rc = strcmp(rmtlx->name, rmtix->name)) == 0)
                break;

            /*
             * If the new element is lower than the current
             * one in the target list, then we found the
             * insertion point.
             */
            if(rc > 0) {
                if((rmtix->prev = rmtlx->prev) != NULL)
                    rmtlx->prev->next = rmtix;
                else
                    rmtl = rmtix;
                rmtix->next = rmtlx;
                rmtlx->prev = rmtix;
                break;
            }

            /*
             * If we reached the end of the target list, then
             * we simply append the new item at the end of the
             * target list.
             */
            else if(rmtlx->next == NULL) {
                rmtlx->next = rmtix;
                rmtix->prev = rmtlx;
                break;
            }
        }
    }

    /*
     * Return the root element.
     */
    return(rmtl);
}

/************************************************************************/
/*                                                                      */
/* add_rmtlink                                                          */
/*                                                                      */
/* Adds a new host to a remote link. Returns the current root           */
/* of the linked list.                                                  */
/*                                                                      */
/************************************************************************/
RMTLINK *add_rmtlink(RMTLINK *rmtl, const char *name)
{
    RMTLINK *rmtp = malloc(sizeof(RMTLINK));
    time_t now;
    struct tm *gmt;

    memset(rmtp, 0, sizeof(RMTLINK));
    strcpy(rmtp->name, name);
    strlwr(rmtp->name);

    tzset();
    time(&now);
    if((gmt = gmtime(&now)) != NULL) {
        rmtp->lastdate = (long)(gmt->tm_year % 100) * 10000L +
                         (long)(gmt->tm_mon + 1) * 100L +
                         (long)gmt->tm_mday;
        rmtp->lasttime = (long)gmt->tm_hour * 10000L +
                         (long)gmt->tm_min * 100L +
                         (long)gmt->tm_sec;
    }
    return(insert_rmtlink(rmtl, rmtp));
}

/************************************************************************/
/*                                                                      */
/* del_rmtlink                                                          */
/*                                                                      */
/* Removes a host from the remote link. Returns the current root        */
/* of the linked list or NULL, if the list is empty.                    */
/*                                                                      */
/************************************************************************/
RMTLINK *del_rmtlink(RMTLINK *rmtl, const char *name)
{
    RMTLINK *rmtp;

    /*
     * List is empty, return immediately.
     */
    if(rmtl == NULL)
        return(NULL);

    /*
     * If we should remove the first entry, then let's return
     * the pointer to the next item.
     */
    if(stricmp(rmtl->name, name) == 0) {
        rmtp = rmtl->next;
        free(rmtl);
        return(rmtp);
    }

    for(rmtp = rmtl; rmtp; rmtp = rmtp->next) {
        if(stricmp(rmtp->name, name) == 0) {
            rmtp->prev->next = rmtp->next;
            if(rmtp->next)
                rmtp->next->prev = rmtp->prev;
            free(rmtp);
            break;
        }
    }
    return(rmtl);
}

/************************************************************************/
/*                                                                      */
/* test_rmtlink                                                         */
/*                                                                      */
/* Returns TRUE, if a hosts.nntp file exists, FALSE if not.             */
/*                                                                      */
/************************************************************************/
int test_rmtlink(void)
{
    int result = 1;
    char *batchfile = malloc(_MAX_PATH);

    getcwd(batchfile, _MAX_PATH);
    strcat(batchfile, "\\hosts.nntp");
    if(access(batchfile, 0))
        result = 0;

    free(batchfile);

    return(result);
}

/************************************************************************/
/*                                                                      */
/* load_rmtlink                                                          */
/*                                                                      */
/* Creates a new linked list from the contents of a watermark file.     */
/*                                                                      */
/************************************************************************/
RMTLINK *load_rmtlink(void)
{
    char **lines, **lpp, *lp;
    RMTLINK *rmtl = NULL;
    RMTLINK *rmtp;
    char *batchfile = malloc(_MAX_PATH);
    int n;

    /*
     * For now we stick with the current working directory.
     */
    getcwd(batchfile, _MAX_PATH);
    strcat(batchfile, "\\hosts.nntp");

    /*
     * Load the watermark file using our beloved text loader.
     */
    if((n = load_textfile(batchfile, &lines)) > 0) {
        for(lpp = lines; (lp = *lpp++) != NULL; ) {
            rmtp = malloc(sizeof(RMTLINK));
            memset(rmtp, 0, sizeof(RMTLINK));

            printf("%s\n", lp);
            switch(sscanf(lp, "%s %ld %ld", rmtp->name, &(rmtp->lastdate), &(rmtp->lasttime))) {
            case 1:
            case 3:
                strlwr(rmtp->name);
                break;
            default:
                printf("Wrong %s\n", rmtp->name);
                break;
            }

            /*
             * Insert created item into linked list. Keep list
             * sorted by name.
             */
            printf("Process '%s'\n", rmtp->name);
            if(rmtp->name[0]) {
                if(rmtp->lastdate == 0L) {
                    time_t now;
                    struct tm *gmt;

                    tzset();
                    time(&now);
                    if((gmt = gmtime(&now)) != NULL) {
                        rmtp->lastdate = (long)(gmt->tm_year % 100) * 10000L +
                                         (long)(gmt->tm_mon + 1) * 100L +
                                         (long)gmt->tm_mday;
                        rmtp->lasttime = (long)gmt->tm_hour * 10000L +
                                         (long)gmt->tm_min * 100L +
                                         (long)gmt->tm_sec;
                    }
                }
                rmtl = insert_rmtlink(rmtl, rmtp);
            }
        }

        /*
         * Free the line buffers created by our beloved text loader.
         */
        free(*lines);
        free(lines);
    }
    free(batchfile);
    return(rmtl);
}

/************************************************************************/
/*                                                                      */
/* save_rmtlink                                                          */
/*                                                                      */
/* Saves a linked list in a watermark file.                             */
/*                                                                      */
/************************************************************************/
int save_rmtlink(RMTLINK *rmtl)
{
    FILE *fp;
    int result = -1;
    char *batchfile = malloc(_MAX_PATH);

    /*
     * Currently no environment; let's use the current directory.
     */
    getcwd(batchfile, _MAX_PATH);
    strcat(batchfile, "\\hosts.nntp");

    /*
     * Open the watermark file and store the complete list.
     */
    if((fp = xfopen(batchfile, "w")) != NULL) {
        while(rmtl) {
            fprintf(fp, "%s %06lu %06lu\n", rmtl->name, rmtl->lastdate, rmtl->lasttime);
            rmtl = rmtl->next;
        }

        /*
         * Here we check for success. To be frankly, all this
         * is quite unsave and we may loose our watermark file.
         * By the time we should write to a temporary file first
         * and rename it in case of success - or let's create
         * a backup.
         */
        if(!fclose(fp))
            result = 0;
    }

    free(batchfile);
    return(result);
}

/************************************************************************/
/*                                                                      */
/* free_rmtlink                                                          */
/*                                                                      */
/* Frees all heap memory occupied by a watermark list created by        */
/* 'load_rmtlink'.                                                       */
/*                                                                      */
/************************************************************************/
int free_rmtlink(RMTLINK *rmtl)
{
    RMTLINK *rmtp;

    while(rmtl) {
        rmtp = rmtl->next;
        free(rmtl);
        rmtl = rmtp;
    }
    return(0);
}

/************************************************************************/
/*                                                                      */
/* This main program is for standalone testing.                         */
/*                                                                      */
/*                                                                      */
/************************************************************************/
#if STANDALONE
int main(void)
{
    RMTLINK *rmtl;

    if((rmtl = load_rmtlink()) != NULL) {
        save_rmtlink(rmtl);
        free_rmtlink(rmtl);
    }

    return(0);
}
#endif

/*
 * $Log: rmtlink.c,v $
 * Revision 1.1  1997/02/13 11:09:35  oliver
 * initial revision
 *
 */
