/*
 * This part was written by Harald Kipp
 *
 * Bug reports should be sent to
 *
 *  harald@os2point.ping.de
 *
 */

/*
 * $Log$
 */

const char split_rcsid[] = "$Id$";

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <split.h>

/************************************************************************/
/*                                                                      */
/* split_list                                                           */
/*                                                                      */
/* Splits a line of text with specified field separators into a vector  */
/* of fields.                                                           */
/*                                                                      */
/* We don't use strchr() since the text is expected to be either        */
/* relatively short or "delimiter-dense."                               */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* text         Points to the string to be split. Note that this string */
/*              will be modified. A NULL pointer or an empty string     */
/*              is accepted.                                            */
/*                                                                      */
/* delimiter    This character specifies the delimiter.                 */
/* ------------ ------------------------------------------------------- */
/* Return value An allocated NULL-terminated array of the fields within */
/*              the modified argument that the caller is expected to    */
/*              save or free.                                           */
/*                                                                      */
/************************************************************************/
char **split_list(char *text, char delimiter)
{
    register int  i;
    register char *cp;
    char **av;
    char **save;

    /*
     * We can be called with a non-existant or empty string.
     */
    if(text == NULL || *text == '\0') {
        save = malloc(sizeof(char *));
        save[0] = NULL;
        return(save);
    }

    /*
     * How much space do we need?
     */
    for (i = 2, cp = text; *cp; cp++)
        /*
         * 07/07/97 harald
         * Allow any type of whitespace if delimiter is a blank.
         */
        if (*cp == delimiter || (delimiter == ' ' && isspace(*cp)))
	    i++;
    save = malloc((size_t)i * sizeof(char *));

    /*
     * Loop over text.
     */
    for (av = save, *av++ = cp = text; *cp; )
        /*
         * 07/07/97 harald
         * Allow any type of whitespace if delimiter is a blank.
         */
        if (*cp == delimiter || (delimiter == ' ' && isspace(*cp))) {
            *cp++ = '\0';
            *av++ = cp;
	}
	else
            cp++;
    *av = NULL;

    return(save);
}

/************************************************************************/
/*                                                                      */
/* split_pathline                                                       */
/*                                                                      */
/* Splits a pathline into a vector. Field separators are according      */
/* RFC 1036, but we allow dash and underline as part of a hostname.     */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* line         Points to the string to be split. Note that this string */
/*              will be modified. A NULL pointer or an empty string     */
/*              is accepted.                                            */
/* ------------ ------------------------------------------------------- */
/* Return value An allocated NULL-terminated array of the fields within */
/*              the modified argument that the caller is expected to    */
/*              save or free.                                           */
/*                                                                      */
/************************************************************************/
char **split_pathline(char *line)
{
    register int  i;
    register char *cp;
    int  got_delim = 0;
    char **av;
    char **save;

    /*
     * Do it the save way. We may be called with a non-existant
     * or empty string as well as leading delimiters.
     */
    if(line != NULL) {
        while(!isalnum(*line) && strchr(".-_", *line) == NULL)
            line++;
    }
    if(line == NULL || *line == '\0') {
        save = malloc(sizeof(char *));
        save[0] = NULL;
        return(save);
    }

    /*
     * How much space do we need?
     */
    for(i = 2, cp = line; *cp; cp++)
        if(!isalnum(*cp) && strchr(".-_", *cp) == NULL) {
            if(!got_delim) {
                i++;
                got_delim = 1;
            }
        }
        else
            got_delim = 0;
    save = malloc((size_t)i * sizeof(char *));

    /*
     * Loop over line.
     */
    for(got_delim = 0, av = save, *av++ = cp = line; *cp; )
        if(!isalnum(*cp) && strchr(".-_", *cp) == NULL) {
            *cp++ = '\0';
            *av = cp;
            got_delim = 1;
	}
        else {
            if(got_delim) {
                av++;
                got_delim = 0;
            }
            cp++;
        }
    *av = NULL;

    return(save);
}
