/*----------------------------------------------------------------------*/
/* Copyright (c) 1996-97 by Harald Kipp. All rights reserved.           */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* Parts are copyright by Kendra Electronic Wonderworks.                */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * This module contains the routines to load a linked list of
 * watermark items from and to save one back to the disk. It
 * is currently used by Changi Lite.
 */

const char wmlink_rcsid[] = "$Id: wmlink.c,v 1.1 1997/02/13 11:09:42 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>
#include <direct.h>

#include <xfile.h>
#include <loadtext.h>
#include <domnpath.h>
#include <wmlink.h>

/************************************************************************/
/*                                                                      */
/* insert_wmlink                                                        */
/*                                                                      */
/* Merges a new linked list into an existing one. Keeps the existing    */
/* list in sorted order.                                                */
/*                                                                      */
/************************************************************************/
WMLINK *insert_wmlink(WMLINK *wml, WMLINK *wmi)
{
    WMLINK *wmix;
    WMLINK *wmlx;
    int rc;

    if(wmi == NULL)
        return(wml);

    /*
     * If the target list is empty, add the first source item.
     */
    if(wml == NULL) {
        wml = wmi;
        wml->next = NULL;
        wml->prev = NULL;
        wmi = wmi->next;
    }

    /*
     * Loop for each element of the source list.
     */
    if((wmix = wmi) != NULL) {

        /*
         * Find the correct entry point in the target list.
         * We must keep the target list in alphabetical order.
         */
        for(wmlx = wml; wmlx; wmlx = wmlx->next) {

            /*
             * If the new element already exists in the target
             * list, it is simply ignored.
             */
            if((rc = strcmp(wmlx->group, wmix->group)) == 0)
                break;

            /*
             * If the new element is lower than the current
             * one in the target list, then we found the
             * insertion point.
             */
            if(rc > 0) {
                if((wmix->prev = wmlx->prev) != NULL)
                    wmlx->prev->next = wmix;
                else
                    wml = wmix;
                wmix->next = wmlx;
                wmlx->prev = wmix;
                break;
            }

            /*
             * If we reached the end of the target list, then
             * we simply append the new item at the end of the
             * target list.
             */
            else if(wmlx->next == NULL) {
                wmlx->next = wmix;
                wmix->prev = wmlx;
                break;
            }
        }
    }

    /*
     * Return the root element.
     */
    return(wml);
}

/************************************************************************/
/*                                                                      */
/* add_wmlink                                                           */
/*                                                                      */
/* Adds a new group to a watermark link. Returns the current root       */
/* of the linked list.                                                  */
/*                                                                      */
/************************************************************************/
WMLINK *add_wmlink(WMLINK *wml, const char *group, WMLINK **wmp)
{
    *wmp = malloc(sizeof(WMLINK));

    memset(*wmp, 0, sizeof(WMLINK));
    strcpy((*wmp)->group, group);
    strlwr((*wmp)->group);

    return(insert_wmlink(wml, *wmp));
}

/************************************************************************/
/*                                                                      */
/* del_wmlink                                                           */
/*                                                                      */
/* Removes a group from the watermark link. Returns the current root    */
/* of the linked list or NULL, if the list is empty.                    */
/*                                                                      */
/************************************************************************/
WMLINK *del_wmlink(WMLINK *wml, const char *group)
{
    WMLINK *wmp;

    /*
     * List is empty, return immediately.
     */
    if(wml == NULL)
        return(NULL);

    /*
     * If we should remove the first entry, then let's return
     * the pointer to the next item.
     */
    if(stricmp(wml->group, group) == 0) {
        wmp = wml->next;
        free(wml);
        return(wmp);
    }

    for(wmp = wml; wmp; wmp = wmp->next) {
        if(stricmp(wmp->group, group) == 0) {
            wmp->prev->next = wmp->next;
            if(wmp->next)
                wmp->next->prev = wmp->prev;
            free(wmp);
            break;
        }
    }
    return(wml);
}

/************************************************************************/
/*                                                                      */
/* test_wmlink                                                          */
/*                                                                      */
/* Returns TRUE, if a host.watermark file exists, FALSE if not.         */
/*                                                                      */
/************************************************************************/
int test_wmlink(const char *hostname)
{
    int result = 1;
    char *batchfile = malloc(_MAX_PATH);

    getcwd(batchfile, _MAX_PATH);
    domain_path(batchfile, hostname, ".watermark");
    if(access(batchfile, 0))
        result = 0;

    free(batchfile);

    return(result);
}

/************************************************************************/
/*                                                                      */
/* load_wmlink                                                          */
/*                                                                      */
/* Creates a new linked list from the contents of a watermark file.     */
/*                                                                      */
/************************************************************************/
WMLINK *load_wmlink(const char *hostname)
{
    char **lines, **lpp, *lp, *cp;
    WMLINK *wml = NULL;
    WMLINK *wmp = NULL;
    char *batchfile = malloc(_MAX_PATH);

    /*
     * For now we stick with the current working directory.
     */
    getcwd(batchfile, _MAX_PATH);
    domain_path(batchfile, hostname, ".watermark");

    /*
     * Load the watermark file using our beloved text loader.
     */
    if(load_textfile(batchfile, &lines) > 0) {
        for(lpp = lines; (lp = *lpp++) != NULL; ) {
            if((cp = strchr(lp, ':')) == NULL)
                continue;
            *cp++ = '\0';
            if(wmp == NULL)
                wmp = malloc(sizeof(WMLINK));
            memset(wmp, 0, sizeof(WMLINK));

            /*
             * We are quite tolerant and will accept the following
             * format:
             *          group:[last[,days[,arts[,size]]]]
             *
             * All unspecified fields are set to zero.
             */
            strcpy(wmp->group, lp);
            strlwr(wmp->group);
            wmp->last_art = atol(cp);
            if((cp = strchr(cp, ',')) != NULL) {
                wmp->days_keep = atoi(++cp);
                if((cp = strchr(cp, ',')) != NULL) {
                    wmp->arts_keep = atoi(++cp);
                    if((cp = strchr(cp, ',')) != NULL)
                        wmp->size_keep = atoi(cp + 1);
                }
            }

            /*
             * Insert created item into linked list. Keep list
             * sorted by groupname.
             */
            wml = insert_wmlink(wml, wmp);
            wmp = NULL;
        }

        /*
         * Free leftovers
         */
        if(wmp)
            free(wmp);

        /*
         * Free the line buffers created by our beloved text loader.
         */
        free(*lines);
        free(lines);
    }
    free(batchfile);
    return(wml);
}

/************************************************************************/
/*                                                                      */
/* save_wmlink                                                          */
/*                                                                      */
/* Saves a linked list in a watermark file.                             */
/*                                                                      */
/************************************************************************/
int save_wmlink(const char *hostname, const WMLINK *wml)
{
    FILE *fp;
    int result = -1;
    char *batchfile = malloc(_MAX_PATH);

    /*
     * Currently no environment; let's use the current directory.
     */
    getcwd(batchfile, _MAX_PATH);
    domain_path(batchfile, hostname, ".watermark");

    /*
     * Remove the watermark file if our list is empty.
     */
    if(wml == NULL)
        result = xremove(batchfile);
    /*
     * Open the watermark file and store the complete list.
     */
    else if((fp = xfopen(batchfile, "w")) != NULL) {
        while(wml) {
            fprintf(fp, "%s: %ld, %d, %d, %ld\n", wml->group, wml->last_art,
                    wml->days_keep, wml->arts_keep, wml->size_keep);
            wml = wml->next;
        }

        /*
         * Here we check for success. To be frankly, all this
         * is quite unsave and we may loose our watermark file.
         * By the time we should write to a temporary file first
         * and rename it in case of success - or let's create
         * a backup.
         */
        if(!fclose(fp))
            result = 0;
    }

    free(batchfile);
    return(result);
}

/************************************************************************/
/*                                                                      */
/* free_wmlink                                                          */
/*                                                                      */
/* Frees all heap memory occupied by a watermark list created by        */
/* 'load_wmlink'.                                                       */
/*                                                                      */
/************************************************************************/
int free_wmlink(WMLINK *wml)
{
    WMLINK *wmp;

    while(wml) {
        wmp = wml->next;
        free(wml);
        wml = wmp;
    }
    return(0);
}

/************************************************************************/
/*                                                                      */
/* This main program is for standalone testing.                         */
/*                                                                      */
/*                                                                      */
/************************************************************************/
#if 0
int main(void)
{
    WMLINK *wml;

    wml = load_wmlink("naxos.ping.de");
    wml = add_wmlink(wml, "9.a.b");
    wml = add_wmlink(wml, "8.a.b");
    wml = add_wmlink(wml, "7.a.b");
    wml = add_wmlink(wml, "6.a.b");
    wml = add_wmlink(wml, "5.a.b");
    wml = add_wmlink(wml, "4.a.b");
    wml = add_wmlink(wml, "3.a.b");
    wml = add_wmlink(wml, "2.a.b");
    save_wmlink("willi.ping.de", wml);
    free_wmlink(wml);

    return(0);
}
#endif

/*
 * $Log: wmlink.c,v $
 * Revision 1.1  1997/02/13 11:09:42  oliver
 * initial revision
 *
 */
