
/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 *
 * This module is used to set access permissions.
 *
 */

const char accessn_rcsid[] = "$Id: accessn.c,v 1.1 1997/02/13 14:03:27 oliver Exp $";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <time.h>

#include <sys/stat.h>

#include <syslog.h>
#include <match.h>
#include <loadtext.h>

#include "config.h"
#include "access.h"

typedef struct {
    char *host;
    char *groups;
    char *nopost;
    int  canpost;
    int  canread;
    int  canxfer;
    int  needreadauth;
    int  needpostauth;
} HOST_PERM;

static char **host_lines = NULL;
static time_t host_mtime = 0;
static int nhost_perms = 0;
static HOST_PERM *host_perm = NULL;

typedef struct {
    char *user;
    char *pass;
    char *groups;
    char *nopost;
} USER_PERM;

static char **user_lines = NULL;
static time_t user_mtime = 0;
static int nuser_perms = 0;
static USER_PERM *user_perm = NULL;

typedef struct {
    char *host;
    char *user;
    char *perm;
} CTRL_PERM;

static char **ctrl_lines = NULL;
static time_t ctrl_mtime = 0;
static int nctrl_perms = 0;
static CTRL_PERM *ctrl_perm = NULL;


/************************************************************************/
/*                                                                      */
/* update_hostperms                                                     */
/*                                                                      */
/************************************************************************/
static void update_hostperms(char *accessfile)
{
    struct stat st;
    int rc;
    int n_lines;
    char **lpp;
    char *lp;
    char *cp;

    /*
     * If host permission file exists and the modification
     * time didn't change, no update is needed.
     */
    if((rc = stat(accessfile, &st)) == 0 && st.st_mtime == host_mtime) {
        syslog(LOG_DEBUG, "%s is uptodate", accessfile);
        return;
    }
    syslog(LOG_DEBUG, "Last=%ld current=%ld", host_mtime, st.st_mtime);

    /*
     * Free previously allocated space.
     */
    if(nhost_perms) {
        syslog(LOG_DEBUG, "free old");
        free(host_lines[0]);
        free(host_lines);
        free(host_perm);
        host_perm = NULL;
        nhost_perms = 0;
        host_mtime = 0;
    }

    /*
     * Load the access permission file and process each line,
     * if it exists.
     */
    syslog(LOG_DEBUG, "rc=%d Load '%s'", rc, accessfile);
    if(rc == 0 && (n_lines = load_textfile(accessfile, &host_lines)) > 0) {
        syslog(LOG_DEBUG, "Loaded %d lines", n_lines);
        host_mtime = st.st_mtime;
        host_perm = malloc(n_lines * sizeof(HOST_PERM));
        for(lpp = host_lines; (lp = *lpp++) != NULL; ) {
            memset(&host_perm[nhost_perms], 0, sizeof(HOST_PERM));
            syslog(LOG_DEBUG, "Process %d: '%s'", nhost_perms, lp);

            /*
             * First field is the host name
             */
            cp = host_perm[nhost_perms].host = lp;
            while(*cp && *cp != ' ')
                cp++;

            /*
             * We need a second field.
             */
            if(*cp != ' ') {
                if(dolog(LOG_PERM))
                    syslog(LOG_DEBUG, "Bad line in %s: '%s'", accessfile, lp);
                continue;
            }
            *cp++ = '\0';

            /*
             * Accept 'default' for backward compatibility.
             */
            if(stricmp(host_perm[nhost_perms].host, "default") == 0)
                strcpy(host_perm[nhost_perms].host, "*");

            /*
             * Second field specifies read/xfer permission. We
             * use the first character only.
             */
            switch(*cp) {
            case 'B':
                host_perm[nhost_perms].needreadauth = 1;
            case 'b':
                host_perm[nhost_perms].canread = 1;
                host_perm[nhost_perms].canxfer = 1;
                break;
            case 'R':
                host_perm[nhost_perms].needreadauth = 1;
            case 'r':
                host_perm[nhost_perms].canread = 1;
                break;
            case 'X':
                host_perm[nhost_perms].needreadauth = 1;
            case 'x':
                host_perm[nhost_perms].canxfer = 1;
                break;
            }
            while(*cp && *cp != ' ')
                cp++;

            /*
             * We need a third field.
             */
            if(*cp != ' ') {
                if(dolog(LOG_PERM))
                    syslog(LOG_DEBUG, "Bad line in %s: '%s'", accessfile, lp);
                continue;
            }

            /*
             * Third field specifies posting permission. We
             * use the first character only.
             */
            switch(*++cp) {
            case 'P':
                host_perm[nhost_perms].needreadauth = 1;
            case 'p':
                host_perm[nhost_perms].canpost = 1;
                break;
            }

            /*
             * The third field might contain a slash separated
             * subfield containing groups to be excluded from
             * posting.
             */
            while(*cp && *cp != ' ' && *cp != '/')
                cp++;
            if(*cp == '/') {
                *cp++ = '\0';
                host_perm[nhost_perms].nopost = cp;
                while(*cp && *cp != ' ')
                    cp++;
            }

            /*
             * The optional fourth field contains a list of
             * newsgroup patterns.
             */
            if(*cp == ' ') {
                *cp++ = '\0';
                host_perm[nhost_perms].groups = cp;
            }
            nhost_perms++;
        }
    }
    return;
}

/************************************************************************/
/*                                                                      */
/* update_userperms                                                     */
/*                                                                      */
/************************************************************************/
static void update_userperms(char *accessfile)
{
    struct stat st;
    int rc;
    int n_lines;
    char **lpp;
    char *lp;
    char *cp;

    /*
     * If user permission file exists and the modification
     * time didn't change, no update is needed.
     */
    if((rc = stat(accessfile, &st)) == 0 && st.st_mtime == user_mtime)
        return;

    /*
     * Free previously allocated space.
     */
    if(nuser_perms) {
        free(user_lines[0]);
        free(user_lines);
        free(user_perm);
        user_perm = NULL;
        nuser_perms = 0;
        user_mtime = 0;
    }

    /*
     * Load the access permission file and process each line,
     * if it exists.
     */
    if(rc == 0 && (n_lines = load_textfile(accessfile, &user_lines)) > 0) {
        user_mtime = st.st_mtime;
        user_perm = malloc(n_lines * sizeof(USER_PERM));
        for(lpp = user_lines; (lp = *lpp++) != NULL; ) {
            memset(&user_perm[nuser_perms], 0, sizeof(USER_PERM));

            /*
             * First field is the user id
             */
            cp = user_perm[nuser_perms].user = lp;
            while(*cp && *cp != ' ')
                cp++;

            /*
             * We need a second field.
             */
            if(*cp != ' ') {
                if(dolog(LOG_PERM))
                    syslog(LOG_DEBUG, "Bad line in %s: '%s'", accessfile, lp);
                continue;
            }
            *cp++ = '\0';

            /*
             * Second field is the unencrypted password.
             */
            user_perm[nuser_perms].pass = cp;
            while(*cp && *cp != ' ')
                cp++;

            /*
             * The third field is optional and may contain a comma
             * separated list of newsgroup patterns, optionally
             * followed by a comma separated list of newsgroup
             * patterns to be excluded from posting.
             */
            if(*cp) {
                *cp++ = '\0';
                user_perm[nuser_perms].groups = cp;
                while(*cp && *cp != '/')
                    cp++;
                if(*cp == '/') {
                    *cp++ = '\0';
                    user_perm[nuser_perms].nopost = cp;
                }
            }
            nuser_perms++;
        }
    }

    return;
}

/************************************************************************/
/*                                                                      */
/* update_ctrlperms                                                     */
/*                                                                      */
/************************************************************************/
static void update_ctrlperms(char *accessfile)
{
    struct stat st;
    int rc;
    int n_lines;
    char **lpp;
    char *lp;
    char *cp;

    /*
     * If ctrl permission file exists and the modification
     * time didn't change, no update is needed.
     */
    if((rc = stat(accessfile, &st)) == 0 && st.st_mtime == ctrl_mtime)
        return;

    /*
     * Free previously allocated space.
     */
    if(nctrl_perms) {
        free(ctrl_lines[0]);
        free(ctrl_lines);
        free(ctrl_perm);
        ctrl_perm = NULL;
        nctrl_perms = 0;
        ctrl_mtime = 0;
    }

    /*
     * Load the access permission file and process each line,
     * if it exists.
     */
    if(rc == 0 && (n_lines = load_textfile(accessfile, &ctrl_lines)) > 0) {
        ctrl_mtime = st.st_mtime;
        ctrl_perm = malloc(n_lines * sizeof(CTRL_PERM));
        for(lpp = ctrl_lines; (lp = *lpp++) != NULL; ) {
            memset(&ctrl_perm[nctrl_perms], 0, sizeof(CTRL_PERM));

            /*
             * First field is the host name
             */
            cp = ctrl_perm[nctrl_perms].host = lp;
            while(*cp && *cp != ' ')
                cp++;

            /*
             * We need a second field.
             */
            if(*cp != ' ') {
                if(dolog(LOG_PERM))
                    syslog(LOG_DEBUG, "Bad line in %s: '%s'", accessfile, lp);
                continue;
            }
            *cp++ = '\0';

            /*
             * Second field is the user-id.
             */
            ctrl_perm[nctrl_perms].user = cp;
            while(*cp && *cp != ' ')
                cp++;

            /*
             * We need a third field.
             */
            if(*cp != ' ') {
                if(dolog(LOG_PERM))
                    syslog(LOG_DEBUG, "Bad line in %s: '%s'", accessfile, lp);
                continue;
            }
            *cp++ = '\0';
            ctrl_perm[nctrl_perms++].perm = cp;
        }
    }
    return;
}

/************************************************************************/
/*                                                                      */
/* host_perms                                                           */
/*                                                                      */
/************************************************************************/
int host_access(char *accessfile, char *remotehost,
                long remoteaddr, ACCESSPERMS *perms)
{
    int i;
    int result = -1;
    char remote_ip[32];
    char *remote_nm;

    syslog(LOG_DEBUG, "%s %s %08lx %08lx", accessfile, remotehost, remoteaddr, perms);
    memset(perms, 0, sizeof(ACCESSPERMS));

    /*
     * Normalize remote name to lower case and convert ip address
     * to a string.
     * 07/06/96 harald
     */
    remote_nm = strlwr(strdup(remotehost));
    sprintf(remote_ip, "%lu.%lu.%lu.%lu",
            ((unsigned long)remoteaddr      ) & 0xff,
            ((unsigned long)remoteaddr >> 8 ) & 0xff,
            ((unsigned long)remoteaddr >> 16) & 0xff,
            ((unsigned long)remoteaddr >> 24) & 0xff);
    syslog(LOG_DEBUG, "ip=%s", remote_ip);

    /*
     * 09/01/1996 harald
     * If accessfile is unspecified, allow local hosts only.
     */
    if(accessfile == NULL || *accessfile == '\0' || access(accessfile, 0)) {
        syslog(LOG_DEBUG, "Local only");
        if(strcmp(remote_ip, "127.0.0.1") == 0 ||
           strcmp(remote_nm, "localhost") == 0 ||
           stricmp(remote_nm, cfg.mydomain) == 0) {
            perms->canpost = perms->canread = perms->canxfer = 1;
            result = 0;
        }
        free(remote_nm);
        return(result);
    }

    if(dolog(LOG_PERM))
        syslog(LOG_DEBUG, "Checking permissions for %s (%s)",
               remote_nm, remote_ip);

    update_hostperms(accessfile);

    for(i = nhost_perms; --i >= 0; ) {
        if(wildmat(remote_nm, host_perm[i].host) ||
           wildmat(remote_ip, host_perm[i].host)) {
            if(!host_perm[i].canread && !host_perm[i].canxfer)
                break;
            perms->canread = host_perm[i].canread;
            perms->canxfer = host_perm[i].canxfer;
            perms->canpost = host_perm[i].canpost;
            perms->needreadauth = host_perm[i].needreadauth;
            perms->needpostauth = host_perm[i].needpostauth;
            if(host_perm[i].groups)
                perms->ngaccess = strdup(host_perm[i].groups);
            if(host_perm[i].nopost)
                perms->ngreadonly = strdup(host_perm[i].nopost);
            result = 0;
            break;
        }
    }
    free(remote_nm);

    return(result);
}

/************************************************************************/
/*                                                                      */
/* user_perms                                                           */
/*                                                                      */
/************************************************************************/
int user_access(char *accessfile, char *username,
                char *password, ACCESSPERMS *perms)
{
    int i;
    int result = -1;

    /*
     * If user_access doesn't exist, we stick with host access permissions.
     */
    if(accessfile == NULL || accessfile[0] == '\0' || access(accessfile, 0)) {
        syslog(LOG_ERR, "%s not found, access denied", accessfile);
        return(result);
    }

    if(dolog(LOG_PERM))
        syslog(LOG_DEBUG, "Checking permissions for %s", username);

    update_userperms(accessfile);

    for(i = nuser_perms; --i >= 0; ) {
        if(wildmat(username, user_perm[i].user)) {
            if(wildmat(password, user_perm[i].pass)) {
                perms->needreadauth = 0;
                perms->needpostauth = 0;
                if(user_perm[i].groups) {
                    if(perms->ngaccess)
                        free(perms->ngaccess);
                    perms->ngaccess = strdup(user_perm[i].groups);
                }
                if(user_perm[i].nopost) {
                    if(perms->ngreadonly)
                        free(perms->ngreadonly);
                    perms->ngreadonly = strdup(user_perm[i].nopost);
                }
                result = 0;
                break;
            }
            else
                break;
        }
    }

    return(result);
}

/************************************************************************/
/*                                                                      */
/* ctrl_perms                                                           */
/*                                                                      */
/************************************************************************/
int ctrl_perms(char *accessfile, char *remotehost,
               long remoteaddr, char *username, char *permissions)
{
    int i;
    char remote_ip[32];
    char *remote_nm;
    int result = -1;

    /*
     * Normalize remote name to lower case and convert ip address
     * to a string.
     */
    remote_nm = strlwr(strdup(remotehost));
    sprintf(remote_ip, "%lu.%lu.%lu.%lu",
            ((unsigned long)remoteaddr      ) & 0xff,
            ((unsigned long)remoteaddr >> 8 ) & 0xff,
            ((unsigned long)remoteaddr >> 16) & 0xff,
            ((unsigned long)remoteaddr >> 24) & 0xff);

    /*
     * If accessfile is unspecified, allow local hosts only.
     */
    if(accessfile == NULL || *accessfile == '\0' || access(accessfile, 0)) {
        if(strcmp(remote_ip, "127.0.0.1") == 0 ||
           strcmp(remote_nm, "localhost") == 0 ||
           stricmp(remote_nm, cfg.mydomain) == 0) {
            strcpy(permissions, "*");
            result = 0;
        }
        free(remote_nm);
        return(result);
    }

    if(dolog(LOG_PERM))
        syslog(LOG_DEBUG, "Checking control permissions for %s (%s)",
               remote_nm, remote_ip);

    update_ctrlperms(accessfile);

    for(i = nctrl_perms; --i >= 0; ) {
        if(wildmat(remote_nm, ctrl_perm[i].host) ||
           wildmat(remote_ip, ctrl_perm[i].host)) {
            if(wildmat(username, ctrl_perm[i].user)) {
                strcpy(permissions, ctrl_perm[i].perm);
                result = 0;
                break;
            }
        }
    }
    free(remote_nm);

    return(result);
}

/*
 * $Log: accessn.c,v $
 * Revision 1.1  1997/02/13 14:03:27  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:32  oliver
 * initial revision
 *
 */
