/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char ahbs_rcsid[] = "$Id: ahbs.c,v 1.1 1997/02/13 14:03:27 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <sostream.h>

#include <syslog.h>

#include "nntp.h"
#include "globals.h"
#include "config.h"
#include "misc.h"
#include "chartran.h"
#include "artfile.h"
#include "client.h"

static char *verbiage[]=
{
    "head and body follow",
    "head follows",
    "body follows",
    "request text separately"
};

/************************************************************************/
/*                                                                      */
/* {ARTICLE,HEAD,BODY,STAT} <messageid>|articlenum                      */
/*                                                                      */
/* Retrieve article, head, body, or stat, depending on the              */
/* command we were invoked with.                                        */
/*                                                                      */
/************************************************************************/
void ahbs(PNEWSCLIENT pnc, int argc, char *argv[])
{
    char art_id[MAX_MSGID];
    int method;
    FILE *fp;

    switch(*argv[0]) {
    case 'A':
        method = WANT_ARTICLE;
        break;
    case 'H':
        method = WANT_HEAD;
        break;
    case 'S':
        method = WANT_STAT;
        break;
    default:
        method = WANT_BODY;
        break;
    }

    /* ------------------------------ */
    /* Retrieve article by message id */
    /* ------------------------------ */
    if (argc == 2 && *argv[1] == '<') {

        if((fp = openartbyid(argv[1])) == NULL) {
            SoStreamPrintf(pnc->s, "%d No article with message-id %s.\r\n",
		      ERR_NOART, argv[1]);
            if(dolog(LOG_CMDFAIL))
                syslog(LOG_DEBUG, "%s not found", argv[1]);
            return;
        }

        /*
         * Unfortunately we have to read the article's header here.
         */
        if (check_ngperm(fp, pnc->perms.ngaccess, pnc->remotehost, pnc->remoteaddr) == 0) {
            SoStreamPrintf(pnc->s, "%d Can't give that to you, sorry.\r\n",
		      ERR_ACCESS);
            fclose(fp);
            syslog(LOG_INFO, "%s: Illegal newsgroup access", pnc->remotehost);
	    return;
        }

        /*
         * This is not strictly following RFC 977 because we
         * return zero instead of the article number. We would
         * have to compare the current group with the references
         * given in the history line. However, as we might not
         * be in any group at all to process this command, this
         * problem could not be solved completely anyway. so we
         * stick with zero until an updated RFC will give more
         * advise on how to handle it.
         */
        if (SoStreamPrintf(pnc->s, "%d 0 %s Article retrieved, %s.\r\n",
                      OK_ARTICLE + method, argv[1],
                      verbiage[method]) != -1) {
            spew_artfile(pnc->s, fp, method, cfg.char_translate ? tran2rdr : NULL);
            pnc->art_count++;
        }

        fclose(fp);
        return;
    }

    /* -------------------------- */
    /* Retrieve article by number */
    /* -------------------------- */

    /*
     * We need read permission here.
     * 07/09/1996 harald
     */
    if(!pnc->perms.canread) {
        SoStreamPrintf(pnc->s, "%d You only have permission to transfer, sorry.\r\n",
                       ERR_ACCESS);
        syslog(LOG_INFO, "%s: Illegal read access", pnc->remotehost);
        return;
    }

    /*
     * Retrieving by number requires a selected group
     */
    if (!pnc->ingroup) {
        SoStreamPrintf(pnc->s, "%d You are not currently in a newsgroup.\r\n",
                       ERR_NCING);
        if(dolog(LOG_CMDFAIL))
            syslog(LOG_INFO, "%s: Requested %s without newsgroup",
                   pnc->remotehost, argv[0]);
        return;
    }

    /*
     * Get either specified article number or currently selected
     */
    if(argc == 1) {
        if(pnc->art_num == 0) {
            SoStreamPrintf(pnc->s, "%d No article is currently selected.\r\n",
		      ERR_NOCRNT);
            if(dolog(LOG_CMDFAIL))
                syslog(LOG_INFO, "%s: Requested %s without article number",
                        pnc->remotehost, argv[0]);
            return;
	}
    }
    else if((pnc->art_num = atol(argv[1])) == 0) {
        SoStreamPrintf(pnc->s, "%d Invalid article number %ld.\r\n",
                       ERR_NOARTIG, pnc->art_num);
        if(dolog(LOG_CMDFAIL))
            syslog(LOG_INFO, "%s: Requested %s with illegal article number %s",
                   pnc->remotehost, argv[0], argv[1]);
	return;
    }

    if((fp = openartbynum(pnc->mydir, pnc->art_num, art_id)) == NULL) {
        SoStreamPrintf(pnc->s, "%d Missing article number %ld.\r\n",
                       ERR_NOARTIG, pnc->art_num);
        if(dolog(LOG_CMDFAIL))
            syslog(LOG_DEBUG, "Article %ld not found", pnc->art_num);
        return;
    }

    if(SoStreamPrintf(pnc->s, "%d %ld %s Article retrieved; %s.\r\n",
               OK_ARTICLE + method, pnc->art_num, art_id, verbiage[method]) != -1) {
        spew_artfile(pnc->s, fp, method, cfg.char_translate ? tran2rdr : NULL);
        pnc->art_count++;
    }

    fclose(fp);
}

/*
 * $Log: ahbs.c,v $
 * Revision 1.1  1997/02/13 14:03:27  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:32  oliver
 * initial revision
 *
 */
