/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 *
 * This module will keep track of all connected clients.
 *
 */

const char clientab_rcsid[] = "$Id: clientab.c,v 1.1 1997/02/13 14:03:35 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <time.h>
#include <ctype.h>

#include <syslog.h>

#include <sysdep.h>
#include <mutex.h>
#include <sostream.h>
#include <overview.h>
#include <dircache.h>
#include <nactive.h>

#include "nntp.h"

#include "config.h"
#include "rescalc.h"
#include "access.h"
#include "serve.h"
#include "client.h"

static long mutex_wait = 41000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

static long hb_mutex = -1;

typedef struct {
    long remoteaddr;    /* host's ip address, used as our key       */
    char remotehost[MAX_DOMAINNAME];    /* domain name this host    */
    int  art_count;
    int  grp_count;
    int  post_count;
    int  rejct_count;
    unsigned long usr_clock;
    long conn_time;
    int  received;
    int  refused;
    int  rejected;
    int  npnc;          /* number of connected clients at this host */
    int  npnu;          /* number of connected users at this host   */
} NEWSHOST;

typedef struct {
    long remoteaddr;    /* host's ip address, used as a main key    */
    char user[128];     /* user's id, used as a secondary key       */
    char remotehost[MAX_DOMAINNAME];    /* domain name this host    */
    int  npnc;          /* number of connections                    */
} NEWSUSER;


static NEWSCLIENT **apnc;  /* array of pointers to newsclient's local data */
static int npnc;
static int xpnc;

static NEWSHOST **apnh;    /* array of pointers to connected host's data   */
static int npnh;
static int xpnh;

static NEWSUSER **apnu;    /* array of pointers to logged in user's data   */
static int npnu;
static int xpnu;

static int debug_on = 0;

static void _Optlink delay_hoststat(void *pv);

/************************************************************************/
/*                                                                      */
/* debug_clientab                                                       */
/*                                                                      */
/* Switches client table debugging on or off.                           */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* enable       A value of 0 will switch off debugging, otherwise it    */
/*              is switched on.                                         */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void debug_clientab(int enable)
{
    debug_on = enable;
}

/************************************************************************/
/*                                                                      */
/* clientab_maxwait                                                     */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void clientab_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int init_clientab(void)
{
    if((hb_mutex = create_mutex()) == -1)
        return(-1);
    else
        return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int exit_clientab(void)
{
    int i;

    /*
     * Free allocated memory.
     */
    if(request_mutex(&hb_mutex, MUTEXWAIT)) {
        if(xpnc) {
            if(debug_on)
                syslog(LOG_DEBUG, "Destroy client table (%d entries)", npnc);
            for(i = 0; i < npnc; i++)
                free(apnc[i]);
            free(apnc);
        }
        if(xpnh) {
            if(debug_on)
                syslog(LOG_DEBUG, "Destroy host table (%d entries)", npnh);
            for(i = 0; i < npnh; i++)
                free(apnh[i]);
            free(apnh);
        }
        if(xpnu) {
            if(debug_on)
                syslog(LOG_DEBUG, "Destroy user table (%d entries)", npnu);
            for(i = 0; i < npnu; i++)
                free(apnu[i]);
            free(apnu);
        }
        release_mutex(hb_mutex);
    }
    destroy_mutex(hb_mutex);

    return(0);
}

/************************************************************************/
/*                                                                      */
/* create_client                                                        */
/*                                                                      */
/* Checks for system overload, access permissions and creates a client  */
/* table entry.                                                         */
/*                                                                      */
/************************************************************************/
NEWSCLIENT *create_client(int sock, char *remotehost, long remoteaddr)
{
    NEWSCLIENT *pnc;
    NEWSHOST   *pnh = NULL;
    ACCESSPERMS perms;
    long sl;
    int i;

    if(debug_on)
        syslog(LOG_DEBUG, "Create client entry for %s", remotehost);

    /*
     * Check the total number of threads and clients, as well
     * as the system load and refuse connection if limits reached.
     */
    if(cfg.max_threads && *_threadid >= cfg.max_threads) {
        SoStreamPrintf(sock, "%d Too many processes. Try later.\r\n", ERR_GOODBYE);
        syslog(LOG_WARNING, "Too many threads (%d), reject %s", *_threadid, remotehost);
        return(NULL);
    }
    if(cfg.max_clients && npnc >= cfg.max_clients) {
        SoStreamPrintf(sock, "%d Too many connections. Try later.\r\n", ERR_GOODBYE);
        syslog(LOG_WARNING, "Too many connections (%d), reject %s", npnc, remotehost);
        return(NULL);
    }
    if(cfg.max_load && (sl = system_load()) >= cfg.max_load) {
        SoStreamPrintf(sock, "%d System overloaded. Try later.\r\n", ERR_GOODBYE);
        syslog(LOG_WARNING, "System overload (%ld), reject %s", sl, remotehost);
        return(NULL);
    }
    if(cfg.min_free && (sl = mbytes_avail(cfg.newsdir)) < cfg.min_free) {
        SoStreamPrintf(sock, "%d System overloaded. Try later.\r\n", ERR_GOODBYE);
        syslog(LOG_WARNING, "Disk space low (%ldM), reject %s", sl, remotehost);
        return(NULL);
    }
    if(cfg.min_core && (sl = mbytes_avail(NULL)) < cfg.min_core) {
        SoStreamPrintf(sock, "%d System overloaded. Try later.\r\n", ERR_GOODBYE);
        syslog(LOG_WARNING, "Memory low (%ldM), reject %s", sl, remotehost);
        return(NULL);
    }

    /*
     * Check access permissions.
     */
    if(host_access(cfg.hostaccessfile, remotehost, remoteaddr, &perms)) {
        SoStreamPrintf(sock, "%d NNTP server can't talk to you. Goodbye.\r\n", ERR_ACCESS);
        syslog(LOG_WARNING, "%s: *!*!* ILLEGAL ACCESS *!*!*", remotehost);
        return(NULL);
    }

    /*
     * Gain access to our list of clients.
     */
    if(request_mutex(&hb_mutex, MUTEXWAIT)) {
        SoStreamPrintf(sock, "%d System overloaded. Try later.\r\n", ERR_GOODBYE);
        syslog(LOG_WARNING, "Can't get mutex access, reject %s", remotehost);
        return(NULL);
    }

    /*
     * First try to find the host.
     */
    for(i = 0; i < npnh; i++)
        if(remoteaddr == apnh[i]->remoteaddr) {
            pnh = apnh[i];
            break;
        }

    /*
     * If we can't find the host, we must create a new entry.
     */
    if(pnh == NULL) {
        pnh = memset(malloc(sizeof(NEWSHOST)), 0, sizeof(NEWSHOST));
        if(npnh >= xpnh) {
            xpnh += 8;
            apnh = realloc(apnh, xpnh * sizeof(NEWSHOST *));
        }
        apnh[npnh++] = pnh;
        pnh->remoteaddr = remoteaddr;
        strcpy(pnh->remotehost, remotehost);
        if(isdigit(*remotehost))
            syslog(LOG_INFO, "%s connected", remotehost);
        else
            syslog(LOG_INFO, "%s (%lu.%lu.%lu.%lu) connected",
                   remotehost,
                   ((unsigned long)remoteaddr      ) & 0xff,
                   ((unsigned long)remoteaddr >> 8 ) & 0xff,
                   ((unsigned long)remoteaddr >> 16) & 0xff,
                   ((unsigned long)remoteaddr >> 24) & 0xff);
        time(&(pnh->conn_time));
    }
    pnh->npnc++;
    if(debug_on)
        syslog(LOG_DEBUG, "%d connection%s to %s", pnh->npnc,
               pnh->npnc == 1 ? "" : "s", remotehost);

    /*
     * Create a new client structure and add it to our list.
     */
    pnc = memset(malloc(sizeof(NEWSCLIENT)), 0, sizeof(NEWSCLIENT));
    if(npnc >= xpnc) {
        xpnc += 8;
        apnc = realloc(apnc, xpnc * sizeof(NEWSCLIENT *));
    }
    apnc[npnc++] = pnc;
    if(debug_on)
        syslog(LOG_DEBUG, "%d connection%s in total", npnc, npnc == 1 ? "" : "s");

    /*
     * Set the client structure.
     */
    pnc->tid = *_threadid;
    pnc->remoteaddr = remoteaddr;
    pnc->s = sock;
    pnc->rb = SoStreamBufCreate(sock);
    memcpy(&(pnc->perms), &perms, sizeof(perms));
    pnc->remotehost = pnh->remotehost;
    init_article(&(pnc->art), pnh->remotehost);

    /*
     * Release exclusive access to the client list.
     */
    release_mutex(hb_mutex);

    /*
     * Welcome the new client.
     */
    if(debug_on)
        syslog(LOG_DEBUG, "Welcoming %s", pnc->remotehost);
    SoStreamPrintf(sock, "%d %s (%s).\r\n",
                   (perms.canpost && servermode == SMOD_RUNNING) ? OK_CANPOST : OK_NOPOST,
                   cfg.welcome,
                   (perms.canpost && servermode == SMOD_RUNNING) ? "posting ok" : "no posting");

    return(pnc);
}

/************************************************************************/
/*                                                                      */
/* destroy_client                                                       */
/*                                                                      */
/* Removes a client table entry previously created by 'create_client'.  */
/*                                                                      */
/************************************************************************/
int destroy_client(NEWSCLIENT *pnc)
{
    int hx;
    int i;
    NEWSUSER *pnu = NULL;
    NEWSHOST *pnh = NULL;


    /*
     * Destroy the socket buffer and release all memory occupied
     * by the article structure.
     */
    if(debug_on)
        syslog(LOG_DEBUG, "Destroy entries of %s", pnc->remotehost);
    SoStreamBufDestroy(pnc->rb);
    free_article(&(pnc->art));

    /*
     * Gain access to our list of clients.
     */
    if(request_mutex(&hb_mutex, MUTEXWAIT)) {
        syslog(LOG_WARNING, "Can't get mutex access to remove %s", pnc->remotehost);
        return(-1);
    }

    /*
     * Remove the user entry
     */
    if(pnc->username[0]) {
        if(debug_on)
            syslog(LOG_DEBUG, "Destroy entries of %s", pnc->username);

        /*
         * Locate the user entry. If we can't find it, something is
         * miserably wrong. We should shutdown the server, or, even
         * better, restart it. <sigh>
         */
        for(hx = npnu; --hx >= 0; )
            if(pnc->remoteaddr == apnu[hx]->remoteaddr &&
               strcmp(pnc->username, apnu[hx]->user) == 0) {
                pnu = apnu[hx];
                break;
            }
        if(hx < 0) {
            syslog(LOG_CRIT, "Internal garbage, restart Changi");
            fprintf(stderr, "Internal garbage, restart Changi\n");
            release_mutex(hb_mutex);
            return(-1);
        }

        /*
         * If this was the user's only connection, we remove
         * him from the list.
         */
        if(--(pnu->npnc) <= 0) {
            if(--npnu)
                for(i = hx; i < npnu; i++)
                    apnu[i] = apnu[i + 1];
            free(pnu);
            syslog(LOG_INFO, "%s@%s logged out", pnc->username, pnc->remotehost);
        }
    }

    /*
     * Locate the host entry. If we can't find it, something is
     * miserably wrong. We should shutdown the server, or, even
     * better, restart it. <sigh>
     */
    for(hx = npnh; --hx >= 0; )
        if(pnc->remoteaddr == apnh[hx]->remoteaddr) {
            pnh = apnh[hx];
            break;
        }
    if(hx < 0) {
        syslog(LOG_CRIT, "Internal garbage, restart Changi");
        fprintf(stderr, "Internal garbage, restart Changi\n");
        release_mutex(hb_mutex);
        return(-1);
    }

    /*
     * Update the host statistics with values from this connection.
     */
    if(debug_on)
        syslog(LOG_DEBUG, "Statistic update for %s", pnc->remotehost);
    pnh->art_count += pnc->art_count;
    pnh->grp_count += pnc->grp_count;
    pnh->post_count += pnc->post_count;
    pnh->rejct_count += pnc->rejct_count;
    pnh->usr_clock += pnc->usr_clock;
    pnh->received += pnc->received;
    pnh->refused += pnc->refused;
    pnh->rejected += pnc->rejected;

    /*
     * If this is the last connection to this remote host, we do some
     * statistics output. It is delayed by a background thread, because
     * the same host may reconnect us shortly.
     */
    if(--(apnh[hx]->npnc) <= 0) {
        if(debug_on)
            syslog(LOG_DEBUG, "Starting delay thread");
        _beginthread(delay_hoststat, NULL, 32767, pnh);
    }

    /*
     * At least we remote the client's local storage from our table.
     */
    for(hx = npnc; --hx >= 0; )
        if(pnc->tid == apnc[hx]->tid)
            break;
    if(hx < 0) {
        syslog(LOG_CRIT, "Internal garbage, restart Changi");
        fprintf(stderr, "Internal garbage, restart Changi\n");
        release_mutex(hb_mutex);
        return(-1);
    }
    if(--npnc)
        for(i = hx; i < npnc; i++)
            apnc[i] = apnc[i + 1];
    free(pnc);

    /*
     * Release exclusive access to the client list.
     */
    release_mutex(hb_mutex);

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static void _Optlink delay_hoststat(void *pv)
{
    int i;
    int hx;
    long remoteaddr;
    NEWSHOST *pnh = pv;
    time_t now;

    /*
     * Gain access to our list of clients.
     */
    if(request_mutex(&hb_mutex, MUTEXWAIT)) {
        syslog(LOG_ERR, "Can't get mutex access to remove %s", pnh->remotehost);
        return;
    }

    /*
     * Save the remote address as a key into our table.
     */
    remoteaddr = pnh->remoteaddr;
    if(debug_on)
        syslog(LOG_DEBUG, "Delay stat output for %lu.%lu.%lu.%lu",
                   ((unsigned long)remoteaddr      ) & 0xff,
                   ((unsigned long)remoteaddr >> 8 ) & 0xff,
                   ((unsigned long)remoteaddr >> 16) & 0xff,
                   ((unsigned long)remoteaddr >> 24) & 0xff);

    /*
     * Release exclusive access to the client list.
     */
    release_mutex(hb_mutex);

    /*
     * Take a nap and see, if we got another connection from this host
     * in the meantime.
     */
    ms_sleep(15000);
    if(debug_on)
        syslog(LOG_DEBUG, "Delay time elapsed");

    /*
     * Gain access to our list of clients.
     */
    if(request_mutex(&hb_mutex, MUTEXWAIT)) {
        syslog(LOG_ERR, "Can't get mutex access to remove %s", pnh->remotehost);
        return;
    }

    /*
     * Locate our host in the table again to make sure it hadn't been
     * removed in the meantime.
     */
    for(hx = npnh; --hx >= 0; )
        if(remoteaddr == apnh[hx]->remoteaddr) {
            pnh = apnh[hx];
            break;
        }
    if(hx >= 0 && pnh->npnc <= 0) {
        /*
         * Print statistics.
         */
        pnh->conn_time = time(&now) - pnh->conn_time;
        if(pnh->art_count || pnh->grp_count)
            syslog(LOG_NOTICE, "%s exit articles %d groups %d",
                   pnh->remotehost, pnh->art_count, pnh->grp_count);
        if(pnh->post_count || pnh->rejct_count)
            syslog(LOG_NOTICE, "%s posts received %d rejected %d",
                   pnh->remotehost, pnh->post_count, pnh->rejct_count);
        if(pnh->received || pnh->refused || pnh->rejected)
            syslog(LOG_NOTICE, "%s closed seconds %d accepted %d refused %d rejected %d",
                   pnh->remotehost, pnh->conn_time, pnh->received, pnh->refused, pnh->rejected);
        syslog(LOG_NOTICE, "%s times user %d.000 system 0.000 elapsed %ld",
               pnh->remotehost,
               pnh->usr_clock / CLOCKS_PER_SEC,
               pnh->conn_time);

        /*
         * If this site send us any articles, flush internal buffers.
         */
        if(pnh->post_count || pnh->received) {
            if(servermode == SMOD_RUNNING) {
                if(debug_on)
                    syslog(LOG_DEBUG, "Flushing buffers");
                flush_overview();
                flush_dircache();
                /*
                 * 08/19/97 harald
                 * Do not force active update.
                 */
                save_active(cfg.activefile, 0);
            }
        }

        /*
         * Remove host from our list.
         */
        for(hx = npnh; --hx >= 0; )
            if(pnh->remoteaddr == apnh[hx]->remoteaddr)
                break;
        if(hx >= 0) {
            if(--npnh)
                for(i = hx; i < npnh; i++)
                    apnh[i] = apnh[i + 1];
            free(pnh);
        }
        if(debug_on)
            syslog(LOG_DEBUG, "Host removed from list");
    }

    /*
     * Release exclusive access to the client list.
     */
    release_mutex(hb_mutex);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
NEWSCLIENT *find_client(int tid)
{
    int i;
    NEWSCLIENT *pnc = NULL;

    if(request_mutex(&hb_mutex, MUTEXWAIT) == 0) {
        for(i = 0; i < npnc; i++) {
            if(tid == apnc[i]->tid) {
                pnc = apnc[i];
                break;
            }
        }
        release_mutex(hb_mutex);
    }
    return(pnc);
}


/************************************************************************/
/*                                                                      */
/************************************************************************/
int get_hostentry(int index, char *info)
{
    int result = -1;

    if(request_mutex(&hb_mutex, MUTEXWAIT))
        return(-1);

    if(index < npnh && index >= 0) {
        strcat(info, apnh[index]->remotehost);
        result = 0;
    }
    release_mutex(hb_mutex);

    return(result);
}


/************************************************************************/
/*                                                                      */
/************************************************************************/
int get_userentry(int index, char *info)
{
    int result = -1;

    if(request_mutex(&hb_mutex, MUTEXWAIT))
        return(-1);

    if(index < npnu && index >= 0) {
        sprintf(info + strlen(info), "%s@%s", apnu[index]->user, apnu[index]->remotehost);
        result = 0;
    }
    release_mutex(hb_mutex);

    return(result);
}


/************************************************************************/
/*                                                                      */
/************************************************************************/
void add_userentry(long remoteaddr, char *remotename, char *username)
{
    int i;
    NEWSUSER *pnu = NULL;

    /*
     * Gain access to our list of logged-in users.
     */
    if(request_mutex(&hb_mutex, MUTEXWAIT)) {
        syslog(LOG_WARNING, "Can't get mutex access to add %s", username);
        return;
    }

    /*
     * Look for an existing entry.
     */
    for(i = 0; i < npnu; i++)
        if(remoteaddr == apnu[i]->remoteaddr &&
           strcmp(username, apnu[i]->user) == 0) {
            pnu = apnu[i];
            break;
        }

    /*
     * If we can't find the user, we must create a new entry.
     */
    if(pnu == NULL) {
        pnu = memset(malloc(sizeof(NEWSUSER)), 0, sizeof(NEWSUSER));
        if(npnu >= xpnu) {
            xpnu += 8;
            apnu = realloc(apnu, xpnu * sizeof(NEWSUSER *));
        }
        apnu[npnu++] = pnu;
        pnu->remoteaddr = remoteaddr;
        strcpy(pnu->remotehost, remotename);
        strcpy(pnu->user, username);
        syslog(LOG_INFO, "%s@%s logged in", username, remotename);
    }
    pnu->npnc++;

    /*
     * Release exclusive access to the client list.
     */
    release_mutex(hb_mutex);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int kick_entries(char *name)
{
    int result = -1;
    int i;
    char *hostname = NULL;

    if((hostname = strchr(name, '@')) != NULL)
        *hostname++ = '\0';
    if(request_mutex(&hb_mutex, MUTEXWAIT))
        return(-1);

    for(i = 0; i < npnc; i++) {
        if(hostname) {
           if(strcmp(apnc[i]->remotehost, hostname) == 0 &&
              strcmp(apnc[i]->username, name) == 0) {
                apnc[i]->kickout = 1;
                result = 0;
            }
        }
        else if(strcmp(apnc[i]->remotehost, name) == 0) {
            apnc[i]->kickout = 1;
            result = 0;
        }
    }
    release_mutex(hb_mutex);

    return(result);
}

/*
 * $Log: clientab.c,v $
 * Revision 1.1  1997/02/13 14:03:35  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:38  oliver
 * initial revision
 *
 */
