/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char distrib_rcsid[] = "$Id: distrib.c,v 1.1 1997/02/13 14:03:38 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <time.h>

#include <arpatime.h>

#include <syslog.h>
#include <match.h>
#include <sysdep.h>
#include <mutex.h>
#include <xfile.h>
#include <pipe.h>

#include "globals.h"
#include "config.h"
#include "misc.h"
#include "spool.h"
#include <overview.h>
#include "newsfeeds.h"
#include "distrib.h"

static long mutex_wait = 42000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

static int send_site(SITE *sp, ARTICLE *artp);
static int groups_wanted(ARTICLE *artp, char **patlist, int wanted);

/************************************************************************/
/*                                                                      */
/* distrib_maxwait                                                      */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void distrib_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/*                                                                      */
/* distrib_art                                                          */
/*                                                                      */
/* Distribute an article, either got via post or transfer. Headers of   */
/* posted articles should have been properly checked and updated before */
/* calling this routine.                                                */
/*                                                                      */
/* Accepted articles will be filed in the local newsbase as well as     */
/* sent to our feeds.                                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* artp         Points to the article structure.                        */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on errors.                             */
/*                                                                      */
/************************************************************************/
int distrib_art(ARTICLE *artp, int posted)
{
    register SITE *sp;
    register char **ip;
    int i;
    int wanted;

    /*
     * Check if we excluded the newsgroup from distribution.
     */
    wanted = groups_wanted(artp, mysite.feedpatt, 0);

    /*
     * Check each site if they want this article. Start with
     * simple comparisions first.
     */
    for(i = 0, sp = sites; i < n_sites; i++, sp++) {

        /*
         * Only send local posts.
         */
        if(sp->local_only && !posted) {
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "Not a local post for %s", sp->sitename);
            continue;
        }

        /*
         * Send (un)moderated only.
         */
        if(sp->just_moderated) {
            if(sp->just_unmoderated) {
                if(artp->validgroups == 0) {
                    /*
                     * Site wants moderated and unmoderated. That means
                     * they want only those groups, which we keep in our
                     * active file.
                     */
                    if(dolog(LOG_FEED))
                        syslog(LOG_DEBUG, "No valid group for %s", sp->sitename);
                    continue;
                }
            }
            else if(artp->moderated == 0) {
                /*
                 * Site wants moderated only.
                 */
                if(dolog(LOG_FEED))
                    syslog(LOG_DEBUG, "No moderated group for %s", sp->sitename);
                continue;
            }
        }
        else if(sp->just_unmoderated && artp->moderated < artp->validgroups) {
            /*
             * Site wants unmoderated only.
             */
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "No unmoderated group for %s", sp->sitename);
            continue;
        }

        /*
         * Article is too big.
         */
        if(sp->max_artsize) {
            if(artp->artsize == 0)
                artp->artsize = article_size(artp, 0);
            if(sp->max_artsize < artp->artsize) {
                if(dolog(LOG_FEED))
                    syslog(LOG_DEBUG, "Article too large for %s", sp->sitename);
                continue;
            }
        }

        /*
         * Too much crossposting
         */
        if(sp->max_xposts && sp->max_xposts < artp->n_groups) {
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "Too much crossposting for %s", sp->sitename);
            continue;
        }

        /*
         * Too many hops
         */
        if(sp->max_hops && sp->max_hops < artp->hopcount) {
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "Too many hops for %s", sp->sitename);
            continue;
        }

        if(artp->pathlist) {
            /*
             * Do not send article if site name is in the path line.
             */
            if(sp->ignore_path == 0 && in_list(sp->sitename, artp->pathlist)) {
                if(dolog(LOG_FEED))
                    syslog(LOG_DEBUG, "%s in path", sp->sitename);
                continue;
            }

            /*
             * Do not send articles with any item of the feed's
             * exclusion list in our pathline.
             */
            if(sp->feedexcl) {
                for(ip = artp->pathlist; *ip; ip++)
                    if(in_list(*ip, sp->feedexcl))
                        break;
                if(*ip) {
                    if(dolog(LOG_FEED))
                        syslog(LOG_DEBUG, "%s in path", *ip);
                    continue;
                }
            }
        }

        /*
         * Do not send articles with distributions which
         * the site isn't interested in.
         */
        if(artp->distriblist) {
            if(sp->feeddist) {
                for(ip = artp->distriblist; *ip; ip++)
                    if(want_distrib(sp->feeddist, *ip))
                        break;
                if(*ip == NULL) {
                    if(dolog(LOG_FEED))
                        syslog(LOG_DEBUG, "Unwanted distribution for %s", sp->sitename);
                    continue;
                }
            }
            else if(dolog(LOG_DISTRIB))
                syslog(LOG_DEBUG, "Accept any distribution");
        }
        else if(dolog(LOG_DISTRIB))
            syslog(LOG_DEBUG, "No distribution header");


        /*
         * Site needs distribution header.
         */
        else if(sp->need_distrib) {
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "%s needs distribution", sp->sitename);
            continue;
        }

        /*
         * At least check if this site wants any of the newsgroups.
         */
        if(groups_wanted(artp, sp->feedpatt, wanted) == 0) {
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "No groups for %s", sp->sitename);
            continue;
        }
        send_site(sp, artp);
    }
    return(0);
}

/************************************************************************/
/* groups_wanted                                                        */
/*                                                                      */
/* This routine determines wether any of an article's newgroups fits    */
/* into a given list of wildmat patterns.                               */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* artp         Points to the article's info structure.                 */
/*                                                                      */
/* patlist      An array of string pointers containing wildmat          */
/*              patterns.                                               */
/*                                                                      */
/* wanted       This value defines the default result. If this value    */
/*              is 1, then at least one pattern with an exclamation     */
/*              mark in front must match to get a function result of 0. */
/* ------------ ------------------------------------------------------- */
/* Return value 1 if any newsgroup fits the pattern, 0 otherwise.       */
/*                                                                      */
/************************************************************************/
static int groups_wanted(ARTICLE *artp, char **patlist, int wanted)
{
    register char *pat;
    register char *cp;
    char **ngp;
    char **patp;
    int nn;

    /*
     * Don't process empty pattern lists.
     */
    if(patlist == NULL || *patlist == NULL)
        return(wanted);

    /*
     * See if any valid group in the newsgroups list matches the pattern.
     * We are done as soon as one group is wanted after passing the
     * complete pattern list.
     */
    for(cp = artp->groupsmode, ngp = artp->groupslist; *ngp; ngp++, cp++) {
        if(*cp != 'y')
            continue;
        for(patp = patlist; (pat = *patp++) != NULL; ) {
            if((nn = (*pat != '!')) == 0)
                pat++;
            if(wildmat(*ngp, pat))
                wanted = nn;
        }
        if(wanted)
            break;
    }
    if(wanted != 0 || artp->x_control[0] == '\0')
        return(wanted);

    /*
     * If no group is wanted, we get a second chance in case of control
     * or junked messages.
     */
    for(patp = patlist; (pat = *patp++) != NULL; ) {
        if((nn = (*pat != '!')) == 0)
            pat++;
        if(wildmat(artp->x_control, pat))
            wanted = nn;
    }
    if(wanted != 0 || (cp = artp->header[ID_Control].info) == NULL)
        return(wanted);

    /*
     * If neither the control group nor junk is wanted, there may be
     * even a third chance. In case this is a 'newgroup' or 'rmgroup'
     * control message, we check the group to be added or removed
     * against our pattern list.
     */
    if(strncmp(cp, "newgroup", 8) == 0)
        cp += 9;
    else if(strncmp(cp, "rmgroup", 7) == 0)
        cp += 8;
    else
        return(wanted);

    while(*cp == ' ' || *cp == '\t')
        cp++;

    for(patp = patlist; (pat = *patp++) != NULL; ) {
        if((nn = (*pat != '!')) == 0)
            pat++;
        if(wildmat(cp, pat))
            wanted = nn;
    }
    return(wanted);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int spew_artinfo(SITE *sp, ARTICLE *artp)
{
    char *cp;
    char *bp;
    char *ip;
    char *xp;
    char *item;
    char *buf;
    int first;
    int gotone;
    int goteol = 0;
    int i;

    /*
     * If site needs a filed article and we didn't store one,
     * then we skip this article.
     */
    if(sp->need_artfile && artp->firstgroup == -1) {
        if(dolog(LOG_FEED))
            syslog(LOG_DEBUG, "Not a locally stored article for %s", sp->sitename);
        return(-1);
    }
    first = artp->firstgroup;

    if(dolog(LOG_FEED))
        syslog(LOG_DEBUG, "Write '%s' for %s", sp->writeflags, sp->sitename);
    item = malloc(MAX_BUFSIZ);
    bp = buf = malloc(MAX_BUFSIZ);
    *buf = *item = '\0';
    for(cp = sp->writeflags; *cp; cp++) {
        switch(*cp) {
        case 'b':   /* Size of article in bytes */
            ltoa(artp->artsize, item, 10);
            break;
        case 'f':   /* Full pathname of the article */
            ip = strcpy(item, cfg.newsdir);
            ip += strlen(ip);
            *ip++ = '\\';
            for(xp = artp->xrefs[first]; *xp; xp++) {
                if(*xp == '.')
                    *ip++ = '\\';
                else
                    *ip++ = *xp;
            }
            *ip++ = '\\';
            ltoa(artp->xrefnums[first], ip, 10);
            break;
        case 'g':   /* Name of the first newsgroup */
            strcpy(item, artp->xrefs[first]);
            break;
        case 'G':   /* Name of all newsgroups separated by commas */
            for(gotone = 0, i = 0; i < artp->nxrefs; i++) {
                if(artp->xrefnums[i]) {
                    if(gotone)
                        strcat(item, ",");
                    else
                        gotone = 1;
                    strcat(item, artp->xrefs[i]);
                }
            }
            break;
        case 'm':   /* Message-ID of the article */
            strcpy(item, artp->header[ID_Message_ID].info);
            break;
        case 'n':   /* Relative pathname of the article */
            for(ip = item, xp = artp->xrefs[first]; *xp; xp++) {
                if(*xp == '.')
                    *ip++ = '\\';
                else
                    *ip++ = *xp;
            }
            *ip++ = '\\';
            ltoa(artp->xrefnums[first], ip, 10);
            break;
        case 's':   /* Name of the site which transmitted the article */
            strcpy(item, artp->fromhost);
            break;
        case 't':   /* The article's arrival time */
            ltoa(artp->arrived, item, 10);
            break;
        case 'D':   /* Contents of the distribution header */
            if(artp->header[ID_Distribution].info)
                strcpy(item, artp->header[ID_Distribution].info);
            else
                strcpy(item, "?");
            break;
        case 'H':   /* All headers of the article (xref always included) */
            if(*buf)
                *bp++ = '\n';
            for(i = 0; artp->header[i].key; i++) {
                if(artp->header[i].info) {
                    strcpy(bp, artp->header[i].key);
                    bp += strlen(bp);
                    *bp++ = ':';
                    *bp++ = ' ';
                    strcpy(bp, artp->header[i].info);
                    bp += strlen(bp);
                    *bp++ = '\n';
                }
            }
            *bp++ = '\n';
            goteol = 1;
            break;
        case 'N':   /* Contents of the newsgroup header */
            strcpy(item, artp->header[ID_Newsgroups].info);
            break;
#if 0
        case 'O':   /* The article's overview data */
            overviewline(item, 0, artp);
            break;
#endif
        }
        if(*item) {
            if(*buf != '\0' && !goteol)
                *bp++ = ' ';
            strcpy(bp, item);
            bp += strlen(bp);
            *item = '\0';
            goteol = 0;
        }
    }
    if(!goteol)
        *bp++ = '\n';
    *bp = '\0';
    fputs(buf, sp->fp);

    free(item);
    free(buf);

    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int send_art(SITE *sp, ARTICLE *artp)
{
    int result = -1;
    long l;

    if(sp->send_info) {
        if(spew_artinfo(sp, artp) == 0) {
            sp->queuesize += artp->artsize;
            result = 0;
        }
    }
    else {
        if((l = spew_article(sp->fp, artp, sp->batch ? "rnews" : NULL, 0, 0)) != 0) {
            sp->queuesize += l;
            result = 0;
        }
    }
    if(fflush(sp->fp) == EOF)
        syslog(LOG_ERR, "Flush %s", _strerror(sp->queuefile));

    return(result);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
static int send_site(SITE *sp, ARTICLE *artp)
{
    int result = -1;

    if(request_mutex(&(sp->mutex), MUTEXWAIT) == 0) {
        if(dolog(LOG_FEED))
            syslog(LOG_DEBUG, "Send %s to %s",
                   artp->header[ID_Message_ID].info, sp->sitename);

        if(sp->fp == NULL) {
            /*
             * Store articles of file feeds and batched program
             * feeds in a queue file.
             */
            if(sp->feed_type == CST_FILE || sp->batch) {
                if((sp->fp = xfopen(sp->queuefile, "ab")) == NULL) {
                    syslog(LOG_ERR, "Open %s", _strerror(sp->queuefile));
                    sp->dropped = 1;
                    release_mutex(sp->mutex);
                    return(-1);
                }
                if((sp->queuesize = filelength(fileno(sp->fp))) == -1)
                    sp->queuesize = 0;
            }
            /*
             * Feed articles of unbatched program feeds directly.
             */
            else {
#ifdef __OS2__
                if((sp->fp = popen(sp->feedparm, "wb")) == NULL) {
                    syslog(LOG_ERR, "%s failed", sp->feedparm);
                    sp->dropped = 1;
                    release_mutex(sp->mutex);
                    return(-1);
                }
#else
                sp->tname = tempnam(cfg.tempdir, "FED");
                if((sp->fp = xfopen(sp->tname, "wb")) == NULL) {
                    syslog(LOG_ERR, "%s", _strerror(sp->tname));
                    sp->dropped = 1;
                    release_mutex(sp->mutex);
                    return(-1);
                }
#endif
            }
        }
        if((result = send_art(sp, artp)) == 0) {
            if(sp->batch == 0 ||
              (sp->flush_size && sp->queuesize > sp->flush_size)) {
                result = flush_a_site(sp);
            }
        }
        release_mutex(sp->mutex);
    }
    return(result);
}

/*
 * $Log: distrib.c,v $
 * Revision 1.1  1997/02/13 14:03:38  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:40  oliver
 * initial revision
 *
 */
