/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * This module contains routines to handle dynamic line buffers.
 */

const char linbuf_rcsid[] = "$Id: linbuf.c,v 1.1 1997/02/13 14:03:43 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "linbuf.h"

#define LINBUF_HYSTERESIS   32768
#define LINBUF_ADMINSIZE    16
#define LINBUF_LINES        128

/************************************************************************/
/*                                                                      */
/* linbuf_create                                                        */
/*                                                                      */
/* Creates a dynamic line buffer. This routine must be called prior to  */
/* any other line buffer routine below.                                 */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* bp           Pointer to a line buffer structure. If this pointer     */
/*              is NULL then the buffer structure will be created in    */
/*              heap memory.                                            */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to the line buffer created.                     */
/*                                                                      */
/************************************************************************/
LINBUF *linbuf_create(LINBUF *bp)
{
    if(bp == NULL) {
        bp = memset(malloc(sizeof(LINBUF)), 0, sizeof(LINBUF));
        bp->inheap = 1;
    }
    else {
        memset(bp, 0, sizeof(LINBUF));
    }
    bp->data = malloc(bp->size = LINBUF_HYSTERESIS - LINBUF_ADMINSIZE);
    bp->offset = malloc((bp->alines = LINBUF_LINES) * sizeof(size_t));

    return(bp);
}

/************************************************************************/
/*                                                                      */
/* linbuf_destroy                                                       */
/*                                                                      */
/* Destroys a dynamic line buffer. Any allocated memory is released     */
/* and you must call 'linbuf_create' in order to use the buffer again.  */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* bp           Pointer to the line buffer structure to be destroyed.   */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to a cleared line buffer structure. If the      */
/*              structure has been created by 'linbuf_create' in heap   */
/*              memory, then NULL is returned.                          */
/*                                                                      */
/************************************************************************/
LINBUF *linbuf_destroy(LINBUF *bp)
{
    if(bp == NULL)
        return(NULL);
    if(bp->data)
        free(bp->data);
    if(bp->offset)
        free(bp->offset);
    if(bp->inheap) {
        free(bp);
        return(NULL);
    }
    memset(bp, 0, sizeof(LINBUF));
    return(bp);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void linbuf_clear(LINBUF *bp)
{
    if(bp == NULL)
        return;

    bp->used = 0;
    bp->nlines = 0;
}

/************************************************************************/
/*                                                                      */
/* linbuf_add                                                           */
/*                                                                      */
/* Adds a new line to the line buffer.                                  */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* bp           Pointer to the line buffer structure. This structure    */
/*              must have been created by 'linbuf_create'.              */
/*                                                                      */
/* line         Points to a '\0' terminated character array containing  */
/*              the line to be added.                                   */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to the added line                               */
/*                                                                      */
/************************************************************************/
char *linbuf_add(LINBUF *bp, const char *line)
{
    char *result;
    size_t len = strlen(line) + 1;

    /*
     * See if we got enough buffer space left.
     * If not, then allocate more.
     */
    if(bp->used + len > bp->size) {
        bp->size += LINBUF_HYSTERESIS;
        bp->data = realloc(bp->data, bp->size);
    }

    /*
     * Copy the line to the end of the buffer.
     */
    result = bp->data + bp->used;
    memcpy(result, line, len);

    /*
     * See if the offset array is large enough. Allocate
     * more elements, if needed.
     */
    if(bp->nlines >= bp->alines) {
        bp->alines += LINBUF_LINES;
        bp->offset = realloc(bp->offset, bp->alines * sizeof(size_t));
    }

    /*
     * Add the offset to the offset array
     */
    bp->offset[bp->nlines++] = bp->used;
    bp->used += len;

    return(result);
}

/************************************************************************/
/*                                                                      */
/* linbuf_append                                                        */
/*                                                                      */
/* Appends a line to the previously added line.                         */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* bp           Pointer to the line buffer structure. This structure    */
/*              must have been created by 'linbuf_create'.              */
/*                                                                      */
/* line         Points to a '\0' terminated string which should be      */
/*              appended.                                               */
/*                                                                      */
/* delim        This character will be used to overwrite the '\0'       */
/*              delimiter of the last added line.                       */
/* ------------ ------------------------------------------------------- */
/* Return value Pointer to the last added line.                         */
/*                                                                      */
/************************************************************************/
char *linbuf_append(LINBUF *bp, const char *line, char delim)
{
    char *cp;
    size_t len = strlen(line) + 1;

    /*
     * See if we got enough buffer space left.
     * If not, then allocate more.
     */
    if(bp->used + len > bp->size) {
        bp->size += LINBUF_HYSTERESIS;
        bp->data = realloc(bp->data, bp->size);
    }

    /*
     * If something is stored in our buffer, then overwrite
     * the last string delimiter with a new given one.
     */
    if(bp->used) {
        cp = bp->data + bp->used;
        *(cp - 1) = delim;
    }

    /*
     * If this is our first line to store, then we have to
     * update the offset array.
     */
    else {
        cp = bp->data;
        bp->offset[bp->nlines++] = 0;
    }

    /*
     * Copy the line to the end of the buffer.
     */
    memcpy(cp, line, len);
    bp->used += len;

    return(LBPTR(bp, bp->nlines - 1));
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int check_maxquote(LINBUF *bodybuf, int limit)
{
    register char *cp;
    int i;
    int n = 0;

    if(bodybuf->nlines <= limit)
        return(0);

    for(i = 0; i < bodybuf->nlines; i++) {
        cp = LBPTR(bodybuf, i);
        while(*cp) {
            if(*cp == '>' || *cp == ':' || *cp == '|') {
                if(strchr(cp, '>') == NULL)
                    n++;
                break;
            }
            cp++;
        }
    }
    if((n * 2) > bodybuf->nlines)
        return(-1);
    return(0);
}

/*
 * $Log: linbuf.c,v $
 * Revision 1.1  1997/02/13 14:03:43  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:44  oliver
 * initial revision
 *
 */
