/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 *
 * This module processes the NEWGROUPS command.
 *
 */

const char newgroups_rcsid[] = "$Id: newgroups.c,v 1.1 1997/02/13 14:03:47 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <datetime.h>
#include <sostream.h>

#include <syslog.h>
#include <xfile.h>
#include <match.h>

#include "config.h"
#include "globals.h"
#include "nntp.h"
#include "misc.h"
#include "ngmatch.h"
#include "client.h"


/************************************************************************/
/*                                                                      */
/* NEWGROUPS date time ["GMT"] [<distributions>]                        */
/*                                                                      */
/* Display new newsgroups since a given date and time, but only         */
/* for those in <distributions>.                                        */
/*                                                                      */
/************************************************************************/
void newgroups(PNEWSCLIENT pnc, int argc, char *argv[])
{
    char *line;
    register char *cp;
    int i;
    long date;
    register FILE *fp;
    int  distcount = 0;
    char **dist_list = NULL;

    if (strlen(argv[1]) != 6 || strlen(argv[2]) != 6) {
        SoStreamPrintf(pnc->s, "%d Date/time must be in form YYMMDD HHMMSS.\r\n",
		  ERR_CMDSYN);
        if(dolog(LOG_CMDFAIL))
            syslog(LOG_INFO, "%s: Date format error in newgroups command", pnc->remotehost);
	return;
    }

    /*
     * Convert yymmdd hhmmss to number of seconds since the epoch
     */
    line = malloc(MAX_BUFSIZ);
    strcpy(line, argv[1]);
    strcat(line, argv[2]);
    date = dtol(line);
    /*
     * 07/10/97 harald
     * Bad date crashes Changi. Added this verification.
     */
    if (date < 86400) {
        SoStreamPrintf(pnc->s, "%d Invalid date specification.\r\n", ERR_CMDSYN);
        syslog(LOG_INFO, "%s: Bad date in newgroups command", pnc->remotehost);
	return;
    }

    argc -= 3;
    argv += 3;

    /*
     * We store the time in GMT by default.
     */
    if (argc > 0 && !stricmp(*argv, "GMT")) {
        ++argv;
        --argc;
    }
    else
        date = local_to_gmt(date);

    if (date < 0L) {
        SoStreamPrintf(pnc->s, "%d Invalid date specification.\r\n", ERR_CMDSYN);
        if(dolog(LOG_CMDFAIL))
            syslog(LOG_INFO, "%s: Invalid date in newgroups command", pnc->remotehost);
        free(line);
	return;
    }

    /*
     * Handle distributions
     */
    if (argc > 0) {
        distcount = get_distlist(&dist_list, *argv);
        if (distcount < 0) {
            SoStreamPrintf(pnc->s, "%d Bad distribution list: %s\r\n",
                      ERR_CMDSYN, *argv);
            if(dolog(LOG_CMDFAIL))
                syslog(LOG_INFO, "%s: Bad distribution list in newgroups command", pnc->remotehost);
            free(line);
	    return;
	}
    }

    /*
     * Open active.times
     */
    if((fp = xfopen(cfg.active_times, "r")) == NULL) {
        syslog(LOG_ERR, "%s", _strerror(cfg.active_times));
        SoStreamPrintf(pnc->s, "%d Cannot open newsgroup date file.\r\n",
                  ERR_FAULT);
        arg_list(NULL, &dist_list);
        free(line);
        return;
    }

    SoStreamPrintf(pnc->s, "%d New newsgroups since %s follow.\r\n",
              OK_NEWGROUPS, line);

    while (fgets(line, MAX_BUFSIZ, fp) != NULL) {
        if ((cp = strchr(line, '\n')) == NULL) {
            syslog(LOG_INFO, "%s: Long line skipped", cfg.active_times);
            continue;
        }
        *cp = '\0';

        if ((cp = strchr(line, ' ')) != NULL)
	    *cp = '\0';
        if (atol(cp + 1) < date)
            continue;

        if (distcount == 0) {
            if(pnc->perms.ngaccess &&
               ngmatch(line, pnc->perms.ngaccess) == 0 &&
               team_access(cfg.teamaccessfile, pnc->remotehost, pnc->remoteaddr, line, 0))
                continue;

            if(SoStreamPuts(pnc->s, line) == -1)
                break;
            if(SoStreamPuts(pnc->s, "\r\n") == -1)
                break;
	}
        else {
            int fbreak = 0;

            if((cp = strchr(line, '.')) == NULL)
		continue;
	    *cp = '\0';
            for (i = 0; i < distcount; ++i)
                if (strcmp(line, dist_list[i]) == 0) {
		    *cp = '.';
                    if(SoStreamPuts(pnc->s, line) == -1) {
                        fbreak = 1;
                        break;
                    }
                    if(SoStreamPuts(pnc->s, "\r\n") == -1)
                        fbreak = 1;
		    break;
                }
            if(fbreak)
                break;
	}
    }

    SoStreamPuts(pnc->s, ".\r\n");
    fclose(fp);
    arg_list(NULL, &dist_list);
    free(line);
}

/*
 * $Log: newgroups.c,v $
 * Revision 1.1  1997/02/13 14:03:47  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:48  oliver
 * initial revision
 *
 */
