/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Additional code had been stolen from                                 */
/*                                                                      */
/*      INN written by Rich $alz                                        */
/*                                                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * Routines in this module deal with the newsfeeds file.
 */

const char newsfeeds_rcsid[] = "$Id: newsfeeds.c,v 1.1 1997/02/13 14:03:48 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>
#include <ctype.h>

#include <syslog.h>
#include <nactive.h>
#include <split.h>
#include <sysdep.h>
#include <mutex.h>
#include <xfile.h>
#ifdef __OS2__
#include <pipe.h>
#endif
#include <loadline.h>

#include "globals.h"
#include "config.h"
#include "misc.h"
#include "spool.h"
#include "newsfeeds.h"

static long mutex_wait = 43000L;
#define MUTEXWAIT   (mutex_wait + __LINE__)

SITE mysite;
SITE *sites;
int n_sites;
int NeedHeaders;
int NeedOverview;

static int nfeed_lines = 0;         /* size of 'feed_lines' array       */
static char **feed_lines = NULL;    /* array of lines in newsfeeds file */

/************************************************************************/
/*                                                                      */
/* newsfeeds_maxwait                                                    */
/*                                                                      */
/* Set max wait for mutex semaphores.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* maxwait      Time to wait for mutex access in milliseconds. -1 will  */
/*              wait forever.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void newsfeeds_maxwait(long maxwait)
{
    mutex_wait = maxwait;
}

/************************************************************************/
/* parse_feedentry                                                      */
/*                                                                      */
/* Parse an individual site entry.                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* line         String pointer to a single line of the newsfeeds file.  */
/*                                                                      */
/* sp           Pointer to a SITE structure, which will be filled by    */
/*              this routine.                                           */
/* ------------ ------------------------------------------------------- */
/* Return value 0 on success, -1 on errors.                             */
/*                                                                      */
/************************************************************************/
static int parse_feedentry(char *line, SITE *sp)
{
    char *cp;
    char *f2;
    char *f3;
    char *f4;
    char **save;
    char **argv;
    unsigned int i;
    int send_id = 0;
    int add_id = 0;

    if(dolog(LOG_FEEDINI))
        syslog(LOG_DEBUG, "Parse '%s'", line);

    memset(sp, 0, sizeof(SITE));

    /*
     * Nip off the first field, the site name.
     */
    if((f2 = strchr(line, ':')) == NULL) {
        syslog(LOG_ERR, "Missing subscriptions in '%s'", line);
        return(-1);
    }
    *f2++ = '\0';
    sp->sitename = strlwr(line);

    /*
     * Check for site exclusions. Articles are not send
     * to this site if any of the exclusions are found
     * in the header's pathline.
     */
    if((cp = strchr(sp->sitename, '/')) != NULL) {
        *cp++ = '\0';
        if(dolog(LOG_FEEDINI))
            syslog(LOG_DEBUG, "Exclusions: '%s'", cp);
        sp->feedexcl = split_list(strlwr(cp), ',');
    }

    /*
     * Moved down, so exclusions are removed first from
     * the sitename.
     */
    if(dolog(LOG_FEEDINI))
        syslog(LOG_DEBUG, "Sitename: '%s'", sp->sitename);
    sprintf(sp->queuefile, "%s\\queue\\%s", cfg.spooldir, sp->sitename);

    /*
     * Parse the second field, the subscriptions.
     */
    if((f3 = strchr(f2, ':')) == NULL) {
        syslog(LOG_ERR, "Missing flags in '%s'", line);
        return(-1);
    }
    *f3++ = '\0';

    /*
     * Check for distributions within the subscription field.
     */
    if((cp = strchr(f2, '/')) != NULL) {
        *cp++ = '\0';
        if(dolog(LOG_FEEDINI))
            syslog(LOG_DEBUG, "Distributions: '%s'", cp);
        sp->feeddist = split_list(cp, ',');
    }
    if(dolog(LOG_FEEDINI))
        syslog(LOG_DEBUG, "Newsgroups: '%s'", f2);
    sp->feedpatt = split_list(f2, ',');

    /*
     * Get the third field, the flags.
     */
    if((f4 = strchr(f3, ':')) == NULL) {
        syslog(LOG_ERR, "Missing parameters in '%s'", line);
        return(-1);
    }
    *f4++ = '\0';

    /*
     * Parse the flags, assume article process by default
     */
    if(dolog(LOG_FEEDINI))
        syslog(LOG_DEBUG, "Flags: '%s'", f3);
    for(save = argv = split_list(f3, ','); (cp = *argv++) != NULL; ) {
        switch (*cp) {
        case '\0':
	    break;

        /*
         * Do not send articles equal or above this size
         */
        case '<':
            cp++;
            sp->max_artsize = atol(cp);
            break;

        /*
         * Special header specific treatment.
         */
        case 'A':
            while (*++cp) {
                /*
                 * Site needs distribution header. Articles
                 * without this header line will not be sent
                 * to this site.
                 */
                if (*cp == 'd')
                    sp->need_distrib = 1;
                /*
                 * Do not check the pathline before propagating.
                 */
                else if (*cp == 'p')
                    sp->ignore_path = 1;
                else {
                    syslog(LOG_ERR, "Unknown flag A%c", *cp);
                    free(save);
                    return(-1);
                }
            }
            break;

        /*
         * Send articles to a file. We implemented this for sys-file
         * compatibility.
         */
        case 'f':
            sp->feed_type = CST_FILE;
            break;

        case 'F':
            sp->feed_type = CST_FILE;
            break;

        /*
         * Maximum number of crosspostings
         */
        case 'G':
            if(*++cp && isdigit(*cp))
                sp->max_xposts = atoi(cp);
            else
                sp->max_xposts = 1;
            break;

        /*
         * Maximum number of hops
         */
        case 'H':
            if(*++cp && isdigit(*cp))
                sp->max_hops = atoi(cp);
            else
                sp->max_hops = 3;
            break;

        /*
         * Write message-ids instead of filenames
         */
        case 'I':
            send_id = 1;
            sp->send_info = 1;
            break;

        /*
         * Only send locally posted articles
         */
        case 'L':
            sp->local_only = 1;
            break;

        /*
         * Transmit only moderated groups
         */
        case 'm':
            sp->just_moderated = 1;
            break;

        /*
         * Send (un)moderated only
         */
        case 'N':
            while(*++cp)
                if (*cp == 'm')
                    sp->just_moderated = 1;
                else if (*cp == 'u')
                    sp->just_unmoderated = 1;
                else {
                    syslog(LOG_ERR, "Unknown flag N%c", *cp);
                    free(save);
                    return(-1);
                }
            break;

        /*
         * Write message-ids after filenames
         */
        case 'n':
            add_id = 1;
            sp->send_info = 1;
            break;

        /*
         * Flush size
         */
        case 'S':
            if(*++cp && isdigit(*cp))
                sp->flush_size = atol(cp);
            else
                sp->flush_size = 131072;
            break;

        /*
         * Feed's type
         */
        case 'T':
            while(*++cp)

                /*
                 * Feed batch files
                 */
                if (*cp == 'b')
                    sp->batch = 1;

                /*
                 * Feed compressed batch files
                 */
                else if (*cp == 'c') {
                    sp->batch = 1;
                    sp->compress = 1;
                }

                /*
                 * Feed by file
                 */
                else if (*cp == 'f')
                    sp->feed_type = CST_FILE;

                 /*
                 * Send info about article, not the article itself
                 */
                else if (*cp == 'i')
                    sp->send_info = 1;

                /*
                 * Feed by program
                 */
                else if (*cp == 'p')
                    sp->feed_type = CST_PROC;

                /*
                 * Feed by UUPC compatible file
                 */
                else if (*cp == 'U') {
                    sp->feed_type = CST_FILE;
                    sp->uupc_mangle = 1;
                    if(sp->uupc_grade == '\0')
                        sp->uupc_grade = 'C';
                }

                else {
                    syslog(LOG_ERR, "Unknown flag T%c", *cp);
                    free(save);
                    return(-1);
                }
            break;

        /*
         * Transmit only unmoderated groups
         */
        case 'u':
            sp->just_unmoderated = 1;
            break;

        /*
         * Set UUPC mangle mode with given grade
         */
        case 'U':
            sp->feed_type = CST_FILE;
            sp->uupc_mangle = 1;
            sp->uupc_grade = *++cp;
            break;

        /*
         * Write flags
         */
        case 'W':
            for(i = 0; *++cp && i < (sizeof(sp->writeflags) - 1); i++) {
                if (*cp == 'n' ||   /* Relative pathname of the article */
                    *cp == 'f' ||   /* Full pathname of the article */
                    *cp == 'g' ||   /* Name of the first newsgroup */
                    *cp == 'G' )    /* Name of all newsgroups separated by commas */
                    sp->need_artfile = 1;
                else if (
                    *cp == 'b' ||   /* Size of article in bytes */
                    *cp == 'm' ||   /* Message-ID of the article */
                    *cp == 's' ||   /* Name of the site which transmitted the article */
                    *cp == 't' ||   /* The article's arrival time */
                    *cp == 'D' ||   /* Contents of the distribution header */
                    *cp == 'H' ||   /* All headers of the article */
                    *cp == 'N' ||   /* Contents of the newsgroup header */
                    *cp == 'O' )    /* The article's overview data */
                    sp->send_info = 1;
                else {
                    syslog(LOG_ERR, "Unknown flag W%c", *cp);
                    free(save);
                    return(-1);
                }
                sp->writeflags[i] = *cp;
            }
            if(*cp) {
                syslog(LOG_ERR, "Too many W flags");
                free(save);
                return(-1);
            }
            sp->writeflags[i] = '\0';
            break;

        default:
            syslog(LOG_ERR, "Unknown flag %c", *cp);
            free(save);
            return(-1);
        }
    }
    if(dolog(LOG_FEEDINI))
        syslog(LOG_DEBUG, "All feed flags done");
    free(save);

    /*
     * Get the fourth field, the param.
     */
    if(*f4 != '\0') {
        sp->feedparm = strdup(f4);
        if(dolog(LOG_FEEDINI))
            syslog(LOG_DEBUG, "Site parameter '%s'", sp->feedparm);
    }

    /*
     * Set defaults for remaining items not yet specified.
     */
    if(sp->feed_type == 0)
        sp->feed_type = CST_FILE;

    if(send_id) {
        sp->writeflags[0] = 'm';
        sp->writeflags[1] = '\0';
    }
    else if(add_id) {
        sp->writeflags[0] = 'f';
        sp->writeflags[1] = 'm';
        sp->writeflags[2] = '\0';
        sp->need_artfile = 1;
    }
    else if(sp->writeflags[0] == '\0') {
        sp->writeflags[0] = 'f';
        sp->writeflags[1] = '\0';
        sp->need_artfile = 1;
    }

    if(strcmp(sp->sitename, "me")) {
        if(sp->feed_type == CST_FILE) {
            if(sp->send_info) {
                if(sp->feedparm == NULL) {
                    sp->feedparm = malloc(_MAX_PATH);
                    sprintf(sp->feedparm, "%s\\outgoing\\%s", cfg.spooldir,
                                                              sp->sitename);
                }
            }
            if(sp->feedparm)
                sp->batch = 1;
        }
        else { /* CST_PROC */
            if(sp->feedparm == NULL) {
                syslog(LOG_ERR, "Missing parameter field");
                return(-1);
            }
            if(sp->batch) {
                if(strings_in_format(sp->feedparm) != 1) {
                    syslog(LOG_ERR, "Missing string flag in parameter field");
                    return(-1);
                }
                sp->batchname = malloc(_MAX_PATH);
                sprintf(sp->batchname, "%s\\outgoing\\%s", cfg.spooldir,
                                                           sp->sitename);
            }
        }
    }

    return(0);
}


/************************************************************************/
/*                                                                      */
/* read_newsfeeds                                                       */
/*                                                                      */
/* This routine reads the entire newsfeeds file and parses it to set    */
/* an array of 'SITE' structures.                                       */
/*                                                                      */
/*                                                                      */
/************************************************************************/
int read_newsfeeds(char *name)
{
    register int i;
    register char *cp;
    char **lp;
    SITE *sp;

    /*
     * Free old site info and initialize a new
     * array of site parameters.
     */
    if(sites) {
        if(dolog(LOG_FEEDINI))
            syslog(LOG_DEBUG, "Freeing old site info");
        for(sp = sites, i = 0; i < n_sites; i++, sp++) {
            if(sp->mutex != -1)
                destroy_mutex(sp->mutex);
            if(sp->feedexcl)
                free(sp->feedexcl);
            if(sp->feeddist)
                free(sp->feeddist);
            if(sp->feedpatt)
                free(sp->feedpatt);
            if(sp->feedparm)
                free(sp->feedparm);
        }
        free(sites);
    }

    if(access(name, 0)) {
        FILE *fp;

        name = "newsfeeds.default";
        if((fp = xfopen(name, "w")) == NULL) {
            syslog(LOG_ERR, "%s", _strerror(name));
            return(-1);
        }
        /*
         * 09/01/1996 harald
         * Removed '!local.*' from ME, added '/!local' to remote
         */
        fputs("ME:*,!control,!control.*,!junk/!local::\n", fp);
        if(cfg.newsserver[0])
            fprintf(fp, "%s:/!local:L:\n", cfg.newsserver);
        else
            syslog(LOG_INFO, "No posting to the world");
        if(fclose(fp)) {
            syslog(LOG_ERR, "%s", _strerror(name));
            return(-1);
        }
    }


    /*
     * Read site entries into a static array of line pointers
     */
    if(dolog(LOG_FEEDINI))
        syslog(LOG_DEBUG, "Reading %s", name);
    if((nfeed_lines = load_textfile(name, &feed_lines)) > 0) {
        sites = malloc((size_t)nfeed_lines * sizeof(SITE));
        if(dolog(LOG_FEEDINI))
            syslog(LOG_DEBUG, "Got room for %d sites", nfeed_lines);

        /*
         * Fill the site array with parameters
         */
        for(sp = sites, n_sites = 0, lp = feed_lines; (cp = *lp++) != NULL;) {

            /*
             * The 'ME' entry should be given once and as the first entry.
             * With this entry all other entries may be preset.
             */
            if(*cp == 'M' && *(cp + 1) == 'E' &&
               (*(cp + 2) == ':' || *(cp + 2) == '/')) {
                if(parse_feedentry(cp, &mysite))
                    memset(&mysite, 0, sizeof(mysite));
                continue;
            }

            /*
             * Other sides are processed here
             */
            if(parse_feedentry(cp, sp)) {
                continue;
            }

            if((sp->mutex = create_mutex()) == -1) {
                syslog(LOG_ERR, "Failed to create mutex %s", sp->sitename);
                continue;
            }

            sp++;
            n_sites++;
        }
    }

    return(0);
}


/************************************************************************/
/*                                                                      */
/************************************************************************/
SITE *find_site(char *name)
{
    int i;
    SITE *sp;
    SITE *result = NULL;

    if(n_sites > 0) {
        for(i = 0, sp = sites; i < n_sites; i++, sp++) {
            if(stricmp(sp->sitename, name) == 0) {
                result = sp;
                break;
            }
        }
    }
    return(result);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int flush_a_site(SITE *sp)
{
#ifndef __OS2__
    char *cmdline;
#endif

    /*
     * Check if anything in queue, return if not.
     */
    if(sp->queuesize == 0) {
        if(dolog(LOG_FEED))
            syslog(LOG_DEBUG, "Nothing to flush for %s", sp->sitename);
        return(0);
    }

    /*
     * We won't flush if the last one still exists.
     */
    if(sp->batch) {
        if(sp->feed_type == CST_FILE && sp->feedparm != NULL &&
           access(sp->feedparm, 0) == 0) {
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "%s not flushed, %s exists", sp->sitename,
                       sp->feedparm);
            return(0);
        }
        if(sp->feed_type == CST_PROC && access(sp->batchname, 0) == 0) {
            if(dolog(LOG_FEED))
                syslog(LOG_DEBUG, "%s not flushed, %s exists", sp->sitename,
                       sp->batchname);
            return(0);
        }
    }

    /*
     * Close queuefile.
     */
    sp->queuesize = 0;
    if(sp->fp) {
        if(sp->feed_type == CST_FILE || sp->batch) {
            if(fclose(sp->fp)) {
                syslog(LOG_ERR, "Close %s", _strerror(sp->sitename));
                sp->fp = NULL;
                sp->dropped = 1;
                return(-1);
            }
            sp->fp = NULL;
        }
        else if(sp->feed_type == CST_PROC) {
#ifdef __OS2__
            if(pclose(sp->fp)) {
                syslog(LOG_ERR, "%s: %s returns error", sp->sitename,
                       sp->feedparm);
                sp->fp = NULL;
                sp->dropped = 1;
                return(-1);
            }
#else
            if(fclose(sp->fp)) {
                syslog(LOG_ERR, "%s", _strerror(sp->tname));
                sp->fp = NULL;
                sp->dropped = 1;
                free(sp->tname);
                return(-1);
            }
            cmdline = malloc(_MAX_PATH);
            sprintf(cmdline, "%s < %s", sp->feedparm, sp->tname);
            if(system(cmdline)) {
                syslog(LOG_ERR, "'%s' returned error", cmdline);
                sp->fp = NULL;
                sp->dropped = 1;
                free(sp->tname);
                free(cmdline);
                return(-1);
            }
            if(xremove(sp->tname))
                syslog(LOG_WARNING, "%s", _strerror(sp->tname));
            free(sp->tname);
            free(cmdline);
#endif
            sp->fp = NULL;
        }
    }

    /*
     * Move or spool file and/or run program.
     */
    if(sp->feed_type == CST_FILE) {
        if(sp->feedparm)
            return(xrename(sp->queuefile, sp->feedparm));
        if(sp->uupc_grade == '\0')
            sp->uupc_grade = 'C';
        return(make_spool(sp->queuefile, sp->sitename, sp->uupc_mangle,
               "rnews", sp->uupc_grade, sp->compress));
    }
    else if(sp->batch && access(sp->batchname, 0)) {
        char *cmd;

        if(xrename(sp->queuefile, sp->batchname)) {
            syslog(LOG_ERR, "Rename %s %s", sp->queuefile, _strerror(sp->batchname));
            sp->dropped = 1;
            return(-1);
        }
        cmd = malloc(strlen(sp->feedparm) + strlen(sp->batchname));
        sprintf(cmd, sp->feedparm, sp->batchname);
        system(cmd);
        free(cmd);
    }
    return(0);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int flush_site(SITE *sp)
{
    int result = 0;
    int i;

    if(sp) {
        if(request_mutex(&(sp->mutex), MUTEXWAIT))
            return(-1);
        if((result = flush_a_site(sp)) == 0 && dolog(LOG_FEED) && sp->batch)
            syslog(LOG_NOTICE, "%s flushed", sp->sitename);
        release_mutex(sp->mutex);
    }
    else {
        syslog(LOG_NOTICE, "Flushing all sites...");
        for(i = 0, sp = sites; i < n_sites; i++, sp++) {
            if(request_mutex(&(sp->mutex), MUTEXWAIT) == 0) {
                flush_a_site(sp);
                release_mutex(sp->mutex);
            }
        }
    }
    return(result);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
int drop_site(SITE *sp)
{
    int result = 0;
    int i;


    if(sp) {
        if(request_mutex(&(sp->mutex), MUTEXWAIT) == 0) {
            flush_a_site(sp);
            release_mutex(sp->mutex);
        }
        sp->dropped = 1;
        syslog(LOG_NOTICE, "%s dropped", sp->sitename);
    }
    else {
        syslog(LOG_NOTICE, "Dropping all sites");
        for(i = 0, sp = sites; i < n_sites; i++, sp++) {
            if(request_mutex(&(sp->mutex), MUTEXWAIT) == 0) {
                flush_a_site(sp);
                release_mutex(sp->mutex);
            }
            sp->dropped = 1;
        }
    }
    return(result);
}

/*
 * $Log: newsfeeds.c,v $
 * Revision 1.1  1997/02/13 14:03:48  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:49  oliver
 * initial revision
 *
 */
