/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char ntmatch_rcsid[] = "$Id: ngmatch.c,v 1.1 1997/02/13 14:03:50 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <syslog.h>

#include "nntp.h"
#include "globals.h"
#include "ngmatch.h"

/************************************************************************/
/*                                                                      */
/* nglistmatch                                                          */
/*                                                                      */
/* Matches a list of newsgroup specifiers with a list of given          */
/* newsgroups.  A pointer to the routine which determines a match is    */
/* also given.  This allows us to do regular expression handling for    */
/* RFC 977's NEWNEWS, and more efficient "strncmps" for the access      */
/* file, which must be checked often.                                   */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* func         Pointer to a function for comparision.                  */
/*                                                                      */
/* dflt         Default result. This result is returned if 'matchlist'  */
/*              is empty.                                               */
/*                                                                      */
/* ngspec       Array of a list of group specifiers and may contain     */
/*              limited regular expressions.                            */
/*                                                                      */
/* ngspeccount  Number of entries in 'ngspec'.                          */
/*                                                                      */
/* matchlist    Array of a list of newsgroups to match against 'ngspec'.*/
/*                                                                      */
/* matchcount   Number of entries in 'matchlist'.                       */
/* ------------ ------------------------------------------------------- */
/* Return value Returns 1 if the named newsgroup is in the list,        */
/*              0 if it is not or dflt if 'matchlist' is empty.         */
/*                                                                      */
/************************************************************************/
int nglistmatch(int (*func) (char *, char *), int dflt, char **ngspec,
                int ngspeccount, char **matchlist, int matchcount)
{
    int i, j;
    int match;
    char *cp;

    if (ngspeccount == 0)
	return (1);

    match = dflt;

    for (i = 0; i < matchcount; ++i) {
        if ((cp = strchr(matchlist[i], ':')) != NULL)
	    *cp = '\0';
        for (j = 0; j < ngspeccount; ++j) {
            if (ngspec[j][0] == '!') {  /* Handle negation */
                if ((*func) (ngspec[j] + 1, matchlist[i])) {
		    match = 0;
                }
            }
            else {
                if ((*func) (ngspec[j], matchlist[i])) {
                    match = 1;
                }
	    }
	}
	if (cp)
	    *cp = ':';
    }
    return (match);
}


/************************************************************************/
/*                                                                      */
/* restreql                                                             */
/*                                                                      */
/* A small regular expression string equivalence routine.               */
/* Thanks to Nick Lai for this time saving device.                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* w            Pointer to an asterisk-broadened, regular expression.   */
/*                                                                      */
/* s            A non-regular-expression string.                        */
/* ------------ ------------------------------------------------------- */
/* Return value Returns 1 if match, 0 otherwise.                        */
/*                                                                      */
/************************************************************************/
int restreql(char *w, char *s)
{
    while (*s && *w) {
	switch (*w) {
        case '*':
	    for (w++; *s; s++)
		if (restreql(w, s))
		    return 1;
	    break;
	default:
	    if (*w != *s)
		return 0;
	    w++, s++;
	    break;
	}
    }
    if (*s)
	return 0;
    while (*w)
	if (*w++ != '*')
	    return 0;
    return 1;
}


/************************************************************************/
/*                                                                      */
/* restreql                                                             */
/*                                                                      */
/* See if s1 is equivalent to s2 up to the length of s1.                */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* s1           Pointer to the first string.                            */
/*                                                                      */
/* s1           Pointer to the second string.                           */
/* ------------ ------------------------------------------------------- */
/* Return value Returns 1 if match, 0 otherwise.                        */
/*                                                                      */
/************************************************************************/
int s1strneql(char *s1, char *s2)
{
    size_t slen;

    if (!strcmp(s1, "all"))
	return (1);
    if ((slen = strlen(s1)) > 4) {
	if (!strcmp(s1 + slen - 4, ".all"))
	    return (!strncmp(s1, s2, slen - 3));
    }
    return (!strncmp(s1, s2, slen));
}

/*
 * $Log: ngmatch.c,v $
 * Revision 1.1  1997/02/13 14:03:50  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:50  oliver
 * initial revision
 *
 */
