/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char oversync_rcsid[] = "$Id: oversync.c,v 1.1 1997/02/13 14:03:54 oliver Exp $";

#include <stdlib.h>
#include <process.h>

#include <syslog.h>
#include <chlimits.h>
#include <sysdep.h>
#include <overview.h>
#include <nactive.h>
#include <dircache.h>

#include "serve.h"
#include "oversync.h"

static int go_on;
static int busy = 0;
static int priority = 0;
static int debug_on = 0;

#include <stdio.h>

/************************************************************************/
/*                                                                      */
/* debug_oversync                                                       */
/*                                                                      */
/* Switch overview sync debugging on/off.                               */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* enable       A value of 0 will switch off debugging, otherwise it    */
/*              is switched on.                                         */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void debug_oversync(int enable)
{
    debug_on = enable;
}

/************************************************************************/
/*                                                                      */
/* oversync_thread                                                      */
/*                                                                      */
/* This routine loops through all newsgroups, queries its directory     */
/* and requests the overview line of each article. If an overview       */
/* line doesn't exist, 'get_overline' will dynamically create one. So   */
/* all our overview lines get updated. We leave it to the LRU buffers   */
/* to write them back to disk.                                          */
/*                                                                      */
/* The thread may be interrupted by clearing the global flag 'go_on'.   */
/* If stopped, the thread clears the global flag 'busy'.                */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* pv           Not used                                                */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
static void _Optlink oversync_thread(void *pv)
{
    char *group;
    char *overline;
    int num_arts;
    int i;
    long *art_array = NULL;

    busy = 1;
    syslog(LOG_INFO, "Updating overview with priority %d", priority);
    set_priority(priority);
    overline = malloc(MAX_OVERLINE);
    group = malloc(MAX_GROUPNAME);
    *group = '\0';
    while(go_on && next_group(group)) {
        if(debug_on)
            syslog(LOG_DEBUG, "Oversync %s", group);
        while(go_on && servermode != SMOD_RUNNING) {
            if(debug_on)
                syslog(LOG_DEBUG, "Server not running, oversync delayed");
            ms_sleep(1000);
            continue;
        }
        if((num_arts = query_groupdir(group, &art_array)) < 1) {
            if(debug_on)
                syslog(LOG_DEBUG, "No articles, oversync skipped");
            continue;
        }
        if(debug_on)
            syslog(LOG_DEBUG, "%d articles to oversync", num_arts);
        for(i = num_arts; go_on && --i >= 0; )
            if(get_overline(group, art_array[i], overline, 0) != 1)
                break;
        free(art_array);
    }
    free(group);
    free(overline);
    if(go_on)
        syslog(LOG_INFO, "Overview updated");
    else
        syslog(LOG_INFO, "Overview update interrupted");
    busy = 0;
}

/************************************************************************/
/*                                                                      */
/* exit_oversync                                                        */
/*                                                                      */
/* Interrupts our overview sync thread, clearing global flag 'go_on'.   */
/*                                                                      */
/************************************************************************/
void exit_oversync(void)
{
    int retries = 10;

    go_on = 0;
    while(busy && retries--)
        ms_sleep(200);
}

/************************************************************************/
/*                                                                      */
/* init_oversync                                                        */
/*                                                                      */
/* Starts a thread to update all overview files. If a thread is         */
/* already running, it is interrupted first.                            */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* speed        Set to 0 for normal, 1 for high and -1 for low priority */
/* ------------ ------------------------------------------------------- */
/* Return value None.                                                   */
/*                                                                      */
/************************************************************************/
void init_oversync(int speed)
{
    if(busy)
        exit_oversync();
    if(!busy) {
        go_on = 1;
        priority = speed;
        _beginthread(oversync_thread, NULL, 32767, NULL);
    }
}

/*
 * $Log: oversync.c,v $
 * Revision 1.1  1997/02/13 14:03:54  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:53  oliver
 * initial revision
 *
 */
