/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char post_rcsid[] = "$Id: post.c,v 1.1 1997/02/13 14:03:56 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <direct.h>

#include <syslog.h>

#include <chlimits.h>
#include <sostream.h>
#include <nhistory.h>
#include <match.h>

#include "config.h"
#include "article.h"
#include "globals.h"
#include "nntp.h"
#include "ngmatch.h"
#include "recvart.h"
#include "moderate.h"
#include "control.h"
#include "chartran.h"
#include "newsfeeds.h"
#include "distrib.h"
#include "artqueue.h"
#include "client.h"
#include "serve.h"
#include "changi.h"

extern int dirty_cache;

/************************************************************************/
/*                                                                      */
/* POST                                                                 */
/*                                                                      */
/* Post an article to a set of newsgroups.                              */
/*                                                                      */
/************************************************************************/
void post(PNEWSCLIENT pnc, int argc, char *argv[])
{
    int i;

    SoStreamPrintf(pnc->s, "%d Continue posting. Period on a line by itself to end.\r\n", CONT_POST);

    /*
     * Receive article's header into an array of line pointers.
     * This will concat continuation lines.
     */
    if(receive_header(pnc->rb, pnc->art.headbuf) <= 0) {
        SoStreamPrintf(pnc->s, "%d Post failed receiving header.\r\n", ERR_POSTFAIL);
        pnc->rejct_count++;
        return;
    }

    /*
     * Receive article's body lines into an array of string pointers.
     * At this point we accept empty bodies because it may be a control
     * message.
     */
    if(receive_block(pnc->rb, pnc->art.bodybuf, cfg.signfile) < 0) {
        SoStreamPrintf(pnc->s, "%d Post failed receiving body.\r\n", ERR_POSTFAIL);
        pnc->rejct_count++;
        return;
    }

    if(cfg.max_quotelines &&
       check_maxquote(pnc->art.bodybuf, cfg.max_quotelines)) {
        SoStreamPrintf(pnc->s, "%d Post rejected, quoted too much.\r\n", ERR_POSTFAIL);
        syslog(LOG_NOTICE, "%s: Post rejected, quoted too much", pnc->remotehost);
        pnc->rejct_count++;
        return;
    }

    /*
     * Parse the header and set up the header line array of the
     * article structure. Headers containing duplicate lines are
     * rejected.
     */
    if(!parse_header(&(pnc->art))) {
        SoStreamPrintf(pnc->s, "%d Post failed, bad header.\r\n", ERR_POSTFAIL);
        pnc->rejct_count++;
        return;
    }

    /*
     * 12/02/1996 harald
     * Do not allow posts while flushed. The main problem is that
     * we are not able to determine the sending host during de-queue.
     */
    if(servermode != SMOD_RUNNING) {
        SoStreamPrintf(pnc->s, "%d Posts currently not accepted. Try later.\r\n", ERR_POSTFAIL);
        syslog(LOG_NOTICE, "%d Post rejected, server flushed.");
        pnc->rejct_count++;
        return;
    }

    /*
     * Inspect header lines.
     */
    if(inspect_article(&(pnc->art), pnc->username, mysite.feedexcl, NULL) == 0) {
        SoStreamPrintf(pnc->s, "%d Post failed, illegal header.\r\n", ERR_POSTFAIL);
        pnc->rejct_count++;
        return;
    }
    if(cfg.force_followup &&
       pnc->art.n_groups >= cfg.force_followup &&
       (pnc->art.n_followups == 0 || pnc->art.n_followups > 2)) {
        SoStreamPrintf(pnc->s, "%d Post rejected, need 1 or 2 follow-ups.\r\n", ERR_POSTFAIL);
        syslog(LOG_NOTICE, "%s: Post rejected, %d crossposts, %d follow-ups",
               pnc->remotehost, pnc->art.n_groups, pnc->art.n_followups);
        pnc->rejct_count++;
        return;
    }

    /*
     * Check access and posting permissions.
     */
    if(pnc->perms.ngaccess || pnc->perms.ngreadonly) {
        for(i = 0; i < pnc->art.n_groups; i++) {
            if((pnc->perms.ngaccess &&
                ngmatch(pnc->art.groupslist[i], pnc->perms.ngaccess) == 0) ||
               (pnc->perms.ngreadonly &&
                ngmatch(pnc->art.groupslist[i], pnc->perms.ngreadonly) != 0)) {
                if(team_access(cfg.teamaccessfile, pnc->remotehost, pnc->remoteaddr, pnc->art.groupslist[i], 1)) {
                    SoStreamPrintf(pnc->s, "%d Post failed, no permission.\r\n", ERR_POSTFAIL);
                    syslog(LOG_DEBUG, "%s: No posting permission", pnc->remotehost);
                    pnc->rejct_count++;
                    return;
                }
            }
        }
    }

    /*
     * Process grep groups.
     */
    if(strcmp(pnc->art.groupslist[0], "local.grep") == 0) {
        if(start_grep(pnc->art.header[ID_Subject].info, pnc->art.header[ID_From].info)) {
            SoStreamPrintf(pnc->s, "%d Post failed, illegal header.\r\n", ERR_POSTFAIL);
            pnc->rejct_count++;
        }
        else {
            SoStreamPrintf(pnc->s, "%d Grep command recognized.\r\n", OK_POSTED);
            pnc->post_count++;
        }
        return;
    }


    /*
     * Process control messages.
     */
    process_control(&(pnc->art));

    /*
     * Perform character translation.
     */
    if(cfg.char_translate) {
        if(pnc->art.header[ID_Subject].info)
            tran2bse(pnc->art.header[ID_Subject].info);
        for(i = 0; i < pnc->art.bodybuf->nlines; i++)
            tran2bse(LBPTR(pnc->art.bodybuf, i));
    }

    /*
     * Unapproved articles to moderated groups will be mailed to
     * the moderator. We store all other articles in our local
     * newsbase and/or distribute them to external feeds.
     */
    if(mail_moderator(&(pnc->art), cfg.smailcall)) {
        SoStreamPrintf(pnc->s, "%d Can't send article to moderator.\r\n", ERR_POSTFAIL);
        pnc->rejct_count++;
        return;
    }
    switch(file_article(&(pnc->art))) {
    default:
        if(addhistart(&(pnc->art)) == 0)
            break;
    case -1:
        SoStreamPrintf(pnc->s, "%d Failed to store article.\r\n", ERR_POSTFAIL);
        pnc->rejct_count++;
        return;
    case 0:
        break;
    }
    if(distrib_art(&(pnc->art), 1)) {
        SoStreamPrintf(pnc->s, "%d Failed distributing article.\r\n", ERR_POSTFAIL);
        pnc->rejct_count++;
        return;
    }
    dirty_cache = atoi(cfg.autoflush);
    SoStreamPrintf(pnc->s, "%d Article posted successfully.\r\n", OK_POSTED);
    pnc->post_count++;
}

/*
 * $Log: post.c,v $
 * Revision 1.1  1997/02/13 14:03:56  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:55  oliver
 * initial revision
 *
 */
