/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char recvart_rcsid[] = "$Id: recvart.c,v 1.1 1997/02/13 14:03:57 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <syslog.h>

#include <chlimits.h>
#include <sostream.h>
#include <xfile.h>

#include "globals.h"
#include "config.h"
#include "recvart.h"

/************************************************************************/
/* receive_header                                                       */
/*                                                                      */
/* Collects all header lines from a tcp stream into one large buffer    */
/* and fills an array of character pointers with the starting position  */
/* of these lines.                                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* rbv          Points to a buffered tcp stream.                        */
/*                                                                      */
/* headlines    Dynamic line buffer storage.                            */
/* ------------ ------------------------------------------------------- */
/* Return value The number of lines or -1 on errors.                    */
/*                                                                      */
/************************************************************************/
int receive_header(int *rbv, LINBUF *headlines)
{
    char *linebuf = malloc(MAX_BUFSIZ);
    char *line;
    int  nlines = 0;

    /*
     * Read all lines upto the first empty line and
     * interpret them as header lines.
     */
    linbuf_clear(headlines);
    for(;;) {
        if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.sockwait) <= 0) {
            nlines = -1;
            break;
        }
        line = linebuf;
        if(linebuf[0] == '.') {
            if(linebuf[1] == '\0')
                break;
            line++;
        }

        if(*line == ' ' || *line == '\t') {
            /*
             * Continuation lines start with space or tab. Ignore
             * them until we found our first real header line.
             */
            if(!nlines)
                continue;

#if 0
            /*
             * Flag this continuation line and skip leading
             * spaces.
             */
            while(*line == ' ' || *line == '\t')
                line++;
#endif
            linbuf_append(headlines, line, '\n');
        }

        /*
         * The header ends with an empty line
         */
        else if(*line == '\0')
            break;

        /*
         * Ignore 'From ' and 'Received: ' lines which may have been
         * left over by some mailers.
         */
        else if(strnicmp(line, "From ", 5) == 0)
            continue;
        else if(strnicmp(line, ">From ", 6) == 0)
            continue;
        else if(!strnicmp(line, "Received: ", 10))
            continue;
        else {
            linbuf_add(headlines, line);
            nlines++;
        }
    }
    free(linebuf);

    return(nlines);
}

/************************************************************************/
/*  receive_block                                                       */
/*                                                                      */
/*                                                                      */
/* Parameter    Description                                             */
/* ------------ ------------------------------------------------------- */
/* rbv          Points to a buffered tcp stream.                        */
/*                                                                      */
/* bodylines    Dynamic line buffer storage.                            */
/* ------------ ------------------------------------------------------- */
/* Return value The number of lines or -1 on errors.                    */
/*                                                                      */
/************************************************************************/
int receive_block(int *rbv, LINBUF *bodylines, char *signfile)
{
    char *linebuf = malloc(MAX_BUFSIZ);
    char *line;
    int  nlines = 0;
    FILE *fp;

    linbuf_clear(bodylines);
    for(;;) {
        if(SoStreamBufRecv(linebuf, MAX_BUFSIZ, rbv, cfg.sockwait) <= 0) {
            nlines = -1;
            break;
        }
        if(linebuf[0] == '.') {
            if(linebuf[1] == '\0')
                break;
            line = linebuf + 1;
        }
        else
            line = linebuf;
        linbuf_add(bodylines, line);
        nlines++;
    }

    /*
     * Add an optional signature file.
     */
    if(signfile && *signfile) {
        if((fp = xfopen(signfile, "r")) != NULL) {
            while(fgets(linebuf, MAX_BUFSIZ, fp)) {
                linbuf_add(bodylines, linebuf);
                nlines++;
            }
            fclose(fp);
        }
        else
            syslog(LOG_ERR, "%s", _strerror(signfile));
    }

    free(linebuf);

    return(nlines);
}

/*
 * $Log: recvart.c,v $
 * Revision 1.1  1997/02/13 14:03:57  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:05:56  oliver
 * initial revision
 *
 */
