/*----------------------------------------------------------------------*/
/* Copyright (c) 1996 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/* This software was originally based on the                            */
/*                                                                      */
/*      Network News Transfer Protocol server                           */
/*                                                                      */
/*      Phil Lapsley                                                    */
/*      University of California, Berkeley                              */
/*      Stan Barber                                                     */
/*      Baylor College of Medicine                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@os2point.ping.de                                         */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char xauth_rcsid[] = "$Id: xauth.c,v 1.1 1997/02/13 14:04:04 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>

#include <sys/stat.h>


#include <syslog.h>

#include <chlimits.h>
#include <sostream.h>

#ifdef USE_CRYPT
#include "pwd.h"
#else
#include "config.h"
#include "access.h"
#endif

#include "nntp.h"
#include "globals.h"
#include "client.h"

#ifdef USE_CRYPT
extern char *crypt(char *key, char *salt);
static int getpass(char *u, char *p);
#endif

/************************************************************************/
/*                                                                      */
/* Simple user/password authentication                                  */
/*                                                                      */
/* Client must supply                                                   */
/*                                                                      */
/*      AUTHINFO USER <userid>                                          */
/*                                                                      */
/* followed by                                                          */
/*                                                                      */
/*      AUTHINFO PASS <password>                                        */
/*                                                                      */
/* which will be looked up in the local password file.                  */
/*                                                                      */
/* Note that passwords travel over the network in plaintext. This can   */
/* be a problem but isn't easy to remedy. At least it's as safe as      */
/* logging in over the network would be.                                */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/************************************************************************/
void doxauthinfo(PNEWSCLIENT pnc, int argc, char *argv[])
{
    if(!pnc->perms.needreadauth && !pnc->perms.needpostauth) {
        SoStreamPrintf(pnc->s, "%d Authorization already completed\r\n", ERR_AUTHREJ);
        return;
    }

    strupr(argv[1]);
    if(strcmp(argv[1], "USER") == 0) {
        if(pnc->username[0])
            SoStreamPrintf(pnc->s, "%d USER already specified\r\n", ERR_AUTHREJ);
        else {
            /*
             * Get user id and prompt for password
             */
            strncpy(pnc->username, argv[2], 8);
            pnc->username[8] = '\0';
            SoStreamPrintf(pnc->s, "%d PASS required\r\n", NEED_AUTHDATA);
        }
    }
    else if(strcmp(argv[1], "PASS") == 0) {
        if(pnc->username[0] == '\0')
            SoStreamPrintf(pnc->s, "%d USER required first\r\n", ERR_AUTHREJ);
#ifdef USE_CRYPT
        else if(getpass(pnc->username, argv[2])) {
#else
        else if(user_access(cfg.useraccessfile, pnc->username, argv[2], &(pnc->perms))) {
#endif
            syslog(LOG_ERR, "AUTHENTICATION ERROR %s@%s",
                   pnc->username, pnc->remotehost);
            SoStreamPrintf(pnc->s, "%d Authentication error\r\n", ERR_ACCESS);
            pnc->username[0] = '\0';
        }
        else {
            SoStreamPrintf(pnc->s, "%d Authentication accepted\r\n",OK_AUTH);
            add_userentry(pnc->remoteaddr, pnc->remotehost, pnc->username);
        }
    }
    else
        SoStreamPrintf(pnc->s, "%d Usage: AUTHINFO USER|PASS user|password.\r\n", ERR_CMDSYN);
}

#ifdef USE_CRYPT
/************************************************************************/
/*                                                                      */
/************************************************************************/
static int getpass(char *u, char *p)
{
    static char pass[10];
    char *namep = NULL;
    struct passwd *pwd;

    strncpy(pass, p, 8);
    pass[8] = 0;
    if((pwd = getpwnam(u)) != NULL)
        namep = crypt(pass, pwd->pw_passwd);

    if(pwd == NULL || namep == NULL || strcmp(namep, pwd->pw_passwd))
        return(-1);

    return(0);
}
#endif

/*
 * $Log: xauth.c,v $
 * Revision 1.1  1997/02/13 14:04:04  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/13  10:06:00  oliver
 * initial revision
 *
 */
