/*----------------------------------------------------------------------*/
/* Copyright (c) 1995-98 by Harald Kipp. All rights reserved.           */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/*                                                                      */
/* Send bug reports to                                                  */
/*                                                                      */
/*      harald@egnite.com                                               */
/*                                                                      */
/*----------------------------------------------------------------------*/

/*
 * This module contains routines for stream socket connections.
 */

/*
 * $Log$
 */

const char sostream_rcsid[] = "$Id$";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef __OS2__

#include <types.h>
#include <netinet\in.h>
#include <sys\socket.h>

#ifdef min
#undef min
#endif

#else

#include <winsock.h>
#define sock_errno    WSAGetLastError

#endif

#include <syslog.h>
#include <sostream.h>

/************************************************************************\
*
* SoStreamSetServer
*
* Called by servers prior to 'SoStreamAccept' to bind a server socket
* to a specified port and put it in listening state.
*
* Parameter    Description
* ------------ ---------------------------------------------------------
* service      Pointer to a string, either containing the name of the
*              service port or the ASCII representation of the port
*              number. May be NULL, in which case defport should
*              contain a valid number.
*
* defport      An unsigned 16 bit number to be used as a port number
*              in case 'service' is NULL or doesn't contain a valid
*              service name or port number.
* ------------ ---------------------------------------------------------
* Return value Handle of the server socket if successful, -1 otherwise.
*
\************************************************************************/
int _Export _System SoStreamSetServer(const char *service, unsigned short defport)
{
    int sock;
    int on = 1;
    unsigned short port;
    struct sockaddr_in server;

    /*
     * Open a socket. Make it reusable, so we can bind it to a
     * port already occupied by another process. Otherwise
     * we weren't able to restart an application, of which
     * another instance is already hanging around as a zombie.
     */
    if((sock = SoStreamOpen()) == -1)
        return -1;
    if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0)
        syslog(LOG_NOTICE, "setsockopt(SO_REUSEADDR) error: %s", SoErrorString(sock_errno()));

    /*
     * Set the address structure.
     */
    if(service == NULL
      || *service == '\0'
      || (port = SoGetServPort(service, "tcp")) == 0)
        port = defport;
    memset(&server, 0, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_port = htons(port);
    server.sin_addr.s_addr = INADDR_ANY;

    /*
     * Bind the socket to a port.
     */
    if (bind(sock,  (void *)&server, sizeof(server)) < 0) {
        syslog(LOG_ERR, "bind(%d) error: %s", port, SoErrorString(sock_errno()));
        SoStreamClose(sock);
        return -1;
    }

    /*
     * Put the socket in listening state.
     */
    if (listen(sock, 5)) {
        syslog(LOG_ERR, "listen(%d) error: %s", port, SoErrorString(sock_errno()));
        SoStreamClose(sock);
        return -1;
    }
    return sock;
}

/************************************************************************\
*
* SoStreamConnect
*
* This routine will be called by clients to setup a connection
* with a server.
*
* Parameter    Description
* ------------ -------------------------------------------------------
* hostname     Pointer to a string containing either the name of
*              the host or the dotted decimal ASCII representation
*              of the host to connect. This pointer may be NULL,
*              in which case the routine will connect to a server
*              on the local host.
*
* service      Pointer to a string, either containing the name of the
*              service port or the ASCII representation of the port
*              number. May be NULL, in which case defport should
*              contain a valid number.
*
* defport      An unsigned 16 bit number to be used as a port number
*              in case 'service' is NULL or doesn't contain a valid
*              service name or port number.
* ------------ -------------------------------------------------------
* Return value Handle of the socket or -1 in case of any error.
*
\************************************************************************/
int _Export _System SoStreamConnect(const char *hostname, const char *service, unsigned short defport)
{
    int sock;
    unsigned short port;
    struct sockaddr_in server;

    /*
     * Set socket address structure.
     */
    memset(&server, 0, sizeof(server));
    server.sin_family = AF_INET;
    if((server.sin_addr.s_addr = SoGetHostAddr(hostname)) == -1)
        return -1;
    if(service == NULL
      || *service == '\0'
      || (port = SoGetServPort(service, "tcp")) == 0)
        port = defport;
    server.sin_port = htons(port);

    /*
     * Open a stream socket and connect the server.
     */
    if((sock = SoStreamOpen()) != -1) {
        if (connect(sock, (void *) & server, sizeof(server)) < 0) {
            syslog(LOG_ERR, "Connect(%s, %d) error: %s",
                   inet_ntoa(server.sin_addr), port, SoErrorString(sock_errno()));
            SoStreamClose(sock);
            sock = -1;
        }
    }
    return sock;
}
