/*
**  Modified by Jochen Friedrich and Harald Kipp
*/
/*
 * Copyright (c) 1983, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * SYSLOG -- print message on log file
 *
 * This routine looks a lot like printf, except that it outputs to the
 * log file instead of the standard output.  Also:
 *      adds a timestamp,
 *      prints the module name in front of the message,
 *      has some other formatting types (or will sometime),
 *      adds a newline on the end of the message.
 *
 * The output of this routine is intended to be read by syslogd(8).
 *
 * Author: Eric Allman
 * Modified to use UNIX domain IPC by Ralph Campbell
 */

const char syslog_rcsid[] = "$Id: syslog.c,v 1.1 1997/02/13 13:48:16 oliver Exp $";

#include <syslog.h>

#ifdef __OS2__

#define OS2
#include <types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#else

#include <stdio.h>
#include <winsock.h>
#define WINSOCK_VERSION (1 < (1 << 8))
#define psock_errno perror
#define soclose(x) closesocket(x)
int sock_init(void)
   {
   WSADATA wsadata;
   return WSAStartup(WINSOCK_VERSION, &wsadata);
   }

#endif



#ifdef min
#undef min
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <process.h>
#include <time.h>
#include <fcntl.h>
#include <io.h>
#include <ctype.h>


static char *LogTag = "syslog";     /* string to tag the entry with */
static int  LogFacility = LOG_USER; /* default facility code */
static int  LogStat = 0;            /* status bits, set by openlog() */
static int  LogFile = -1;           /* fd for log */
static struct sockaddr_in SyslogAddr;   /* AF_INET address of local logger */
static int  connected = 0;          /* have done connect */

/************************************************************************/
/*                                                                      */
/************************************************************************/
void _System _Export vsyslog(int pri, char *fmt, va_list ap)
{
    register int cnt;
    register char *p;
    time_t now;
    int saved_errno;
    char tbuf[2048], fmt_cpy[1024];

    saved_errno = errno;

    /*
     * see if we should just throw out this message
     */
    if (!LOG_MASK(LOG_PRI(pri)) || (pri &~ (LOG_PRIMASK|LOG_FACMASK)))
        return;
    if (LogFile < 0 || !connected)
        openlog(LogTag, LogStat | LOG_NDELAY, 0);

    /*
     * set default facility if none specified
     */
    if ((pri & LOG_FACMASK) == 0)
        pri |= LogFacility;

    /*
     * build the message
     */
    time(&now);
    sprintf(tbuf, "<%d>%.15s ", pri, ctime(&now) + 4);
    for (p = tbuf; *p; ++p)
        continue;
    if (LogTag) {
        strcpy(p, LogTag);
        for (; *p; ++p)
            continue;
    }
    if (LogStat & LOG_PID) {
        sprintf(p, "[%d:%d]", getpid(), *_threadid);
        for (; *p; ++p)
            continue;
    }
    if (LogTag) {
        *p++ = ':';
        *p++ = ' ';
    }

    /*
     * substitute error message for %m
     */
    {
        register char ch, *t1, *t2;

        for (t1 = fmt_cpy; (ch = *fmt) != '\0'; ++fmt)
            if (ch == '%' && fmt[1] == 'm') {
                ++fmt;
                for(t2 = strerror(saved_errno); (*t1 = *t2++) != '\0'; ++t1)
                    continue;
            }
            else
                *t1++ = ch;
        *t1 = '\0';
    }

    vsprintf(p, fmt_cpy, ap);

    cnt = strlen(tbuf);

    /*
     * output to stderr if requested
     */
    if (LogStat & LOG_PERROR) {
        write(2, tbuf,cnt);
        write(2,"\r\n",2);
    }

    /*
     * output the message to the local logger
     */
    send(LogFile, tbuf, cnt, 0);
    return;
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
/*
 * OPENLOG -- open system log
 */
void _System _Export openlog(char *ident, int logstat, int logfac)
{
    if (ident != NULL)
        LogTag = ident;
    LogStat = logstat;
    if (logfac != 0 && (logfac &~ LOG_FACMASK) == 0)
        LogFacility = logfac;

    if (LogFile == -1) {
        SyslogAddr.sin_family = AF_INET;
        SyslogAddr.sin_port = htons(514);
#ifdef __OS2__
        SyslogAddr.sin_addr.s_addr = INADDR_ANY;
#else
        SyslogAddr.sin_addr.s_addr = 0x0100007f;
#endif
        if (LogStat & LOG_NDELAY) {
            if((LogFile = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
                perror("socket");
        }
    }
    if (LogFile != -1 && !connected) {
        if(connect(LogFile, (struct sockaddr *)&SyslogAddr, sizeof(SyslogAddr)) != -1)
            connected = 1;
        else
            fprintf(stderr, "Syslog connect failed\n");
    }
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
/*
 * CLOSELOG -- close the system log
 */
void _System _Export closelog(void)
{
    soclose(LogFile);
    LogFile = -1;
    connected = 0;
}

/*
 * $Log: syslog.c,v $
 * Revision 1.1  1997/02/13 13:48:16  oliver
 * initial revision
 *
 * Revision 1.1  1997/02/07  18:25:51  oliver
 * initial release
 *
 */
