/*----------------------------------------------------------------------*/
/* Copyright (c) 1997 by Harald Kipp. All rights reserved.              */
/*                                                                      */
/* Redistribution and use in source and binary forms, with or without   */
/* modification, are permitted provided that the following conditions   */
/* are met:                                                             */
/*                                                                      */
/* 1. Redistributions of source code must retain the above copyright    */
/*    notice, this list of conditions and the following disclaimer.     */
/*                                                                      */
/* 2. Redistributions in binary form must reproduce the above copyright */
/*    notice, this list of conditions and the following disclaimer in   */
/*    the documentation and/or other materials provided with the        */
/*    distribution.                                                     */
/*                                                                      */
/* 3. All advertising materials mentioning features or use of this      */
/*    software must display the following acknowledgment:               */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* 4. The name "Changi Server" must not be used to endorse or promote   */
/*    products derived from this software without prior written         */
/*    permission.                                                       */
/*                                                                      */
/* 5. Redistributions of any form whatsoever must retain the following  */
/*    acknowledgment:                                                   */
/*    "This product includes software developed by Harald Kipp          */
/*    for use in the Changi news server project."                       */
/*                                                                      */
/* ==================================================================== */
/* THIS SOFTWARE IS PROVIDED BY HARALD KIPP ``AS IS'' AND ANY EXPRESSED */
/* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED    */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   */
/* ARE DISCLAIMED. IN NO EVENT SHALL HARALD KIPP BE LIABLE FOR ANY      */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL   */
/* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE    */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS        */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER */
/* IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN  */
/* IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                        */
/* ==================================================================== */
/*                                                                      */
/*----------------------------------------------------------------------*/

const char xbatch_rcsid[] = "$Id: rnews.c,v 1.1 1997/02/13 14:00:59 oliver Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include <getopt.h>

#include <version.h>
#include <chlimits.h>
#include <sostream.h>
#include <syslog.h>

#include "config.h"
#include "xbatch.h"

static int syslog_opt = LOG_CONS | LOG_NDELAY | LOG_NOWAIT;

static void parsedbgopt(char *opt);
static void usage(void);

/************************************************************************/
/*                                                                      */
/*  XBATCH                                                              */
/*                                                                      */
/************************************************************************/
int main(int argc, char **argv)
{
    time_t t_proc;
    int option;
    int result = 0;
    FILE *fp;
    void *rbv;
    int sock;

    tzset();
    t_proc = time(NULL);

    /*
     * Initialize socket interface.
     */
    if(SoInitialize()) {
        fprintf(stderr, "FATAL ERROR: TCP/IP initialization failed\n");
        return(128);
    }

    /*
     * Set up initial configuration
     */
    init_cfg();

    /*
     * Process command line options
     */
    while((option = getopt(argc, argv,
            "?a:d:F:f:H:p:W:")) != EOF) {
        switch(option) {
        case 'a':
            strcpy(cfg.authentication, optarg);
            break;
        case 'd':
            parsedbgopt(optarg);
            break;
        case 'F':
            cfg.del_input = 1;
            strcpy(cfg.newsfile, optarg);
            break;
        case 'f':
            strcpy(cfg.newsfile, optarg);
            break;
        case 'H':
            strcpy(cfg.remoteaddr, optarg);
            break;
        case 'p':
            strcpy(cfg.remoteport, optarg);
            break;
        case 'W':
            cfg.sockwait = atol(optarg);
            break;
        default:
            syslog(LOG_INFO, "Unknown option -%c", option);
        case '?':
            usage();
            return(1);
        }
    }
    openlog("xbatch", syslog_opt, LOG_NEWS);
    syslog(LOG_INFO, "Version %s-%s up and running...", version, __DATE__);

    if((fp = freopen(cfg.newsfile, "rb", stdin)) == NULL) {
        syslog(LOG_ERR, "Input %s: %m", cfg.newsfile);
        result = 3;
    }

    if(fp) {
        if(validate_cfg() == 0) {
            char *intro;

            if ((sock = SoStreamConnect(cfg.remoteaddr, cfg.remoteport, 119)) < 0)
                return(3);

            rbv = SoStreamBufCreate(sock);
            intro = malloc(MAX_BUFSIZ);
            if(SoStreamBufRecv(intro, MAX_BUFSIZ, rbv, cfg.sockwait) <= 0) {
                syslog(LOG_ERR, "Failed to get server intro");
                free(intro);
                SoStreamBufDestroy(rbv);
                SoStreamClose(sock);
                return(3);
            }

            if (intro[0] != '2') {
                syslog(LOG_ERR, "Error %s", intro);
                free(intro);
                SoStreamBufDestroy(rbv);
                SoStreamClose(sock);
                return(3);
            }
            free(intro);
#if 0
            /*
             * HERE WE'RE GOING TO SEND THE BATCH
             */
            if(artfileproc(rbv, fp))
                result = 2;
#endif
            SoStreamBufDestroy(rbv);
            SoStreamClose(sock);
        }
        else
            result = 3;
        fclose(fp);
    }

    if(result == 0 && cfg.del_input && unlink(cfg.newsfile)) {
        syslog(LOG_ERR, "%s %m", cfg.newsfile);
        result = 5;
    }
    if(result)
        syslog(LOG_ERR, "Return with error %d", result);
    else {
        t_proc = time(NULL) - t_proc;
        syslog(LOG_NOTICE, "Processed batch in %lu second%s",
                           t_proc, t_proc == 1 ? "" : "s");
    }

    return(result);
}

/************************************************************************/
/*                                                                      */
/* Parse debug options                                                  */
/*                                                                      */
/* This has been moved to an own procedure in order to overcome trouble */
/* with VisualAge for Windows.                                          */
/*                                                                      */
/************************************************************************/
static void parsedbgopt(char *opt)
{
    while(*opt) {
        switch(*opt) {
        case 'C':
            cfg.logflg |= LOG_CONFIG;
            break;
        case 'd':
            cfg.logflg |= LOG_DUPES;
            break;
        case 'h':
            cfg.logflg |= LOG_HISTORY;
            break;
        case 'j':
            cfg.logflg |= LOG_JUNK;
            break;
        case 'm':
            cfg.logflg |= LOG_HEAP;
            break;
        /*
         * 08/30/1996
         * Added option -do and -dp
         */
        case 'o':
            syslog_opt |= LOG_PERROR;
            break;
        case 'p':
            syslog_opt |= LOG_PID;
            break;
        case 'r':
            cfg.logflg |= LOG_READ;
            break;
        case 't':
            cfg.logflg |= LOG_TEST;
            break;
        case 'w':
            SoStreamDebug(1);
            break;
        default:
            syslog(LOG_NOTICE, "Unknown option -d%c ignored", *opt);
            break;
        }
        opt++;
    }
}

/************************************************************************/
/*                                                                      */
/* Usage display                                                        */
/*                                                                      */
/************************************************************************/
static void usage(void)
{
    fprintf(stderr, "\nxbatch %s - %s %s\n\n", version, __DATE__, __TIME__);
    fprintf(stderr, "usage: xbatch [options]\n");
    fputs("options:\n"
         "  -a<user:passwd>  authentication\n"
         "  -d<log-flags>    logfile flags\n"
         "  -f<filename>     input file\n"
         "  -F<filename>     like -f, but delete\n"
         "  -H<remote>       remote host\n"
         "  -p<port>         server port\n"
         "  -W<time>         socket timeout\n", stderr);
}
